@@ -2,7 +2,7 @@
  *
  *  \brief Tandy Colour Computer 3 machine.
  *
- *  \copyright Copyright 2003-2023 Ciaran Anscomb
+ *  \copyright Copyright 2003-2025 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -47,14 +47,18 @@
 #include "mc6809/hd6309.h"
 #include "mc6809/mc6809.h"
 #include "mc6821.h"
+#include "messenger.h"
 #include "ntsc.h"
 #include "part.h"
 #include "printer.h"
+#include "ram.h"
+#include "rombank.h"
 #include "romlist.h"
 #include "serialise.h"
 #include "sound.h"
 #include "tape.h"
 #include "tcc1014/tcc1014.h"
+#include "ui.h"
 #include "vo.h"
 #include "xroar.h"
 
@@ -62,14 +66,68 @@
 # define M_PI 3.14159265358979323846
 #endif
 
-// Guesses, based on not much:
-static float rgb_intensity_map[4] = { 0.00, 0.47, 0.75, 0.92 };
+// Measured peak-to-peak voltages: 0V, 1V, 1.6V, 2V.  Scaled here:
+static float rgb_intensity_map[4] = { 0.000, 0.460, 0.736, 0.920 };
 
-// Same:
-static float hue_intensity_map[4] = { 0.36, 0.52, 0.66, 0.83 };
+// The GIME appears to generate its composite output (used in NTSC machines
+// only) by switching between a set of 7 voltages at 3.58Mhz (presumably with
+// R/C net to smooth to almost-but-not-quite sine waves).  AFAICT there is no
+// black level separate to blank (which would be usual for an NTSC signal).
+//
+// The grey and colour luminances are different at each intensity because there
+// is no output voltage corresponding to the luminance of colour output: it is
+// simply the average of the high and low voltages used to form the colour
+// signal.
+//
+// The colour amplitude is similar for intensity levels 0-2, but reduced for
+// intensity 3, leading to less saturated colour.
+//
+// Monochrome output (bit 4 of VMODE register at $FF98) only affects composite,
+// and as well as removing the colourburst, will only emit the grey level for
+// each intensity; TVs don't immediately (or ever) switch to mono just because
+// the colourburst is missing.
+//
+// Observed on a scope, there is a larger jump in phase between hues 11 and 12
+// than between any other adjacent hues (including between 1 and 15). This
+// supports the theory that the colour phase is a simple offset counted in GIME
+// clock edges, with a gap as there are 16 edges in one colour cycle, and only
+// 15 hues.
 
-// These are the approximate values measured by John Kowalski:
-static float grey_intensity_map[4] = { 0.03, 0.23, 0.50, 1.00 };
+// Approximate measured voltages (at composite video port, relative to blank):
+
+#define CMP_V_SYNC (-0.350)
+#define CMP_V_BURST_LOW (-0.210)
+#define CMP_V_0 (0.000)
+#define CMP_V_1 (0.170)
+#define CMP_V_2 (0.380)
+#define CMP_V_3 (0.580)
+#define CMP_V_4 (0.750)
+
+// Aliases:
+#define CMP_V_BLANK (CMP_V_0)
+#define CMP_V_BURST_HIGH (CMP_V_1)
+#define CMP_V_GREY0 (CMP_V_0)
+#define CMP_V_GREY1 (CMP_V_1)
+#define CMP_V_GREY2 (CMP_V_2)
+#define CMP_V_GREY3 (CMP_V_4)
+#define CMP_V_PEAK (CMP_V_4)
+
+// Map selected intensity level to the grey and and colour peak voltages:
+
+static struct {
+	float grey;
+	float clr_low;
+	float clr_high;
+} const cmp_intensity[4] = {
+	{ CMP_V_GREY0, CMP_V_BURST_LOW, CMP_V_2 },
+	{ CMP_V_GREY1, CMP_V_0, CMP_V_3 },
+	{ CMP_V_GREY2, CMP_V_1, CMP_V_4 },
+	{ CMP_V_GREY3, CMP_V_2, CMP_V_4 },
+};
+
+// Note that the rest of XRoar's video system as been based on the idea that
+// measured voltages will be Y'PbPr, but these measurements are at the
+// composite port, so are already in Y'UV so will need some massaging.
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
@@ -79,18 +137,16 @@ struct machine_coco3 {
 	struct MC6809 *CPU;
 	struct TCC1014 *GIME;
 	struct MC6821 *PIA0, *PIA1;
+	struct rombank *ROM0;
+	struct ram *RAM;
 
 	struct vo_interface *vo;
 	int frame;  // track frameskip
 	struct sound_interface *snd;
 
-	unsigned ram_size;
-	unsigned ram_mask;
-	uint8_t *ram;
-	uint8_t rom0[0x8000];
-
 	_Bool inverted_text;
 	struct cart *cart;
+	unsigned configured_frameskip;
 	unsigned frameskip;
 
 	int cycles;
@@ -122,27 +178,33 @@ struct machine_coco3 {
 		uint32_t vram_bank;
 	} dat;
 
+	// UI message receipt
+	int msgr_client_id;
+
 	// Useful configuration side-effect tracking
 	_Bool has_secb;
 	uint32_t crc_secb;
 };
 
-#define COCO3_SER_RAM          (2)
-#define COCO3_SER_DAT_MMU_BANK (11)
+#define COCO3_SER_RAM           (2)
+#define COCO3_SER_RAM_SIZE      (3)
+#define COCO3_SER_RAM_MASK      (4)
+#define COCO3_SER_DAT_MMU_BANK  (11)
+#define COCO3_SER_DAT_VRAM_BANK (12)
 
 static const struct ser_struct ser_struct_coco3[] = {
 	SER_ID_STRUCT_NEST(1,  &machine_ser_struct_data),
 	SER_ID_STRUCT_UNHANDLED(COCO3_SER_RAM),
-	SER_ID_STRUCT_ELEM(3,  ser_type_unsigned, struct machine_coco3, ram_size),
-	SER_ID_STRUCT_ELEM(4,  ser_type_unsigned, struct machine_coco3, ram_mask),
-	SER_ID_STRUCT_ELEM(5,  ser_type_bool,     struct machine_coco3, inverted_text),
-	SER_ID_STRUCT_ELEM(6,  ser_type_bool,     struct machine_coco3, dat.enabled),
-	SER_ID_STRUCT_ELEM(7,  ser_type_bool,     struct machine_coco3, dat.readable),
-	SER_ID_STRUCT_ELEM(8,  ser_type_bool,     struct machine_coco3, dat.MMUEN),
-	SER_ID_STRUCT_ELEM(9,  ser_type_bool,     struct machine_coco3, dat.MC3),
-	SER_ID_STRUCT_ELEM(10, ser_type_unsigned, struct machine_coco3, dat.task),
+	SER_ID_STRUCT_UNHANDLED(COCO3_SER_RAM_SIZE),
+	SER_ID_STRUCT_UNHANDLED(COCO3_SER_RAM_MASK),
+	SER_ID_STRUCT_ELEM(5,  struct machine_coco3, inverted_text),
+	SER_ID_STRUCT_ELEM(6,  struct machine_coco3, dat.enabled),
+	SER_ID_STRUCT_ELEM(7,  struct machine_coco3, dat.readable),
+	SER_ID_STRUCT_ELEM(8,  struct machine_coco3, dat.MMUEN),
+	SER_ID_STRUCT_ELEM(9,  struct machine_coco3, dat.MC3),
+	SER_ID_STRUCT_ELEM(10, struct machine_coco3, dat.task),
 	SER_ID_STRUCT_UNHANDLED(COCO3_SER_DAT_MMU_BANK),
-	SER_ID_STRUCT_ELEM(12, ser_type_uint32,   struct machine_coco3, dat.vram_bank),
+	SER_ID_STRUCT_UNHANDLED(COCO3_SER_DAT_VRAM_BANK),
 };
 
 static _Bool coco3_read_elem(void *sptr, struct ser_handle *sh, int tag);
@@ -161,13 +223,16 @@ static void coco3_config_complete(struct machine_config *mc) {
 	if (!mc->description) {
 		mc->description = xstrdup(mc->name);
 	}
+	if (mc->cpu != CPU_MC6809 && mc->cpu != CPU_HD6309) {
+		mc->cpu = CPU_MC6809;
+	}
 	if (mc->tv_standard == ANY_AUTO)
 		mc->tv_standard = TV_PAL;
 	if (mc->tv_input == ANY_AUTO) {
 		switch (mc->tv_standard) {
 		default:
 		case TV_PAL:
-			mc->tv_input = TV_INPUT_RGB;
+			mc->tv_input = TV_INPUT_SVIDEO;
 			break;
 		case TV_NTSC:
 		case TV_PAL_M:
@@ -179,8 +244,11 @@ static void coco3_config_complete(struct machine_config *mc) {
 		mc->vdg_type = VDG_GIME_1986;
 	if (mc->vdg_type != VDG_GIME_1986 && mc->vdg_type != VDG_GIME_1987)
 		mc->vdg_type = VDG_GIME_1986;
-	if (mc->ram != 128 && mc->ram != 512 && mc->ram != 1024 && mc->ram != 2048)
-		mc->ram = 128;
+
+	if (mc->ram_init == ANY_AUTO) {
+		mc->ram_init = ram_init_pattern;
+	}
+
 	mc->keymap = dkbd_layout_coco3;
 	/* Now find which ROMs we're actually going to use */
 	if (!mc->extbas_dfn && !mc->extbas_rom) {
@@ -197,9 +265,8 @@ static void coco3_config_complete(struct machine_config *mc) {
 static _Bool coco3_is_working_config(struct machine_config *mc) {
 	if (!mc)
 		return 0;
-	sds tmp;
 	if (mc->extbas_rom) {
-		tmp = romlist_find(mc->extbas_rom);
+		sds tmp = romlist_find(mc->extbas_rom);
 		if (!tmp)
 			return 0;
 		sdsfree(tmp);
@@ -211,6 +278,9 @@ static _Bool coco3_is_working_config(struct machine_config *mc) {
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
+static _Bool coco3_has_interface(struct part *p, const char *ifname);
+static void coco3_attach_interface(struct part *p, const char *ifname, void *intf);
+
 static void coco3_connect_cart(struct part *p);
 static void coco3_insert_cart(struct machine *m, struct cart *c);
 static void coco3_remove_cart(struct machine *m);
@@ -223,12 +293,14 @@ static void coco3_trap(void *sptr);
 static void coco3_bp_add_n(struct machine *m, struct machine_bp *list, int n, void *sptr);
 static void coco3_bp_remove_n(struct machine *m, struct machine_bp *list, int n);
 
-static int coco3_set_keyboard_type(struct machine *m, int action);
+static void coco3_ui_set_keymap(void *, int tag, void *smsg);
 static _Bool coco3_set_pause(struct machine *m, int state);
-static _Bool coco3_set_inverted_text(struct machine *m, int state);
+static void coco3_ui_set_picture(void *, int tag, void *smsg);
+static void coco3_ui_set_tv_input(void *, int tag, void *smsg);
+static void coco3_ui_set_text_invert(void *, int tag, void *smsg);
 static void *coco3_get_interface(struct machine *m, const char *ifname);
-static void coco3_set_frameskip(struct machine *m, unsigned fskip);
-static void coco3_set_ratelimit(struct machine *m, _Bool ratelimit);
+static void coco3_ui_set_frameskip(void *, int tag, void *smsg);
+static void coco3_ui_set_ratelimit(void *, int tag, void *smsg);
 
 static uint8_t coco3_read_byte(struct machine *m, unsigned A, uint8_t D);
 static void coco3_write_byte(struct machine *m, unsigned A, uint8_t D);
@@ -248,6 +320,11 @@ static void gime_hs(void *sptr, _Bool level);
 // static void gime_hs_pal_coco(void *sptr, _Bool level);
 static void gime_fs(void *sptr, _Bool level);
 static void gime_render_line(void *sptr, unsigned burst, unsigned npixels, uint8_t const *data);
+static void coco3_print_byte(void *);
+
+static struct machine_bp coco3_print_breakpoint[] = {
+	BP_COCO3_ROM(.address = 0xa2c1, .handler = DELEGATE_INIT(coco3_print_byte, NULL) ),
+};
 
 static void cpu_cycle(void *sptr, int ncycles, _Bool RnW, uint16_t A);
 static void cpu_cycle_noclock(void *sptr, int ncycles, _Bool RnW, uint16_t A);
@@ -288,13 +365,7 @@ static const struct partdb_entry_funcs coco3_funcs = {
 	.is_a = machine_is_a,
 };
 
-const struct machine_partdb_extra coco3_machine_extra = {
-	.config_complete = coco3_config_complete,
-	.is_working_config = coco3_is_working_config,
-	.cart_arch = "dragon-cart",
-};
-
-const struct partdb_entry coco3_part = { .name = "coco3", .funcs = &coco3_funcs, .extra = { &coco3_machine_extra } };
+const struct machine_partdb_entry coco3_part = { .partdb_entry = { .name = "coco3", .description = "Tandy | Colour Computer 3", .funcs = &coco3_funcs }, .config_complete = coco3_config_complete, .is_working_config = coco3_is_working_config, .cart_arch = "dragon-cart" };
 
 static struct part *coco3_allocate(void) {
         struct machine_coco3 *mcc3 = part_new(sizeof(*mcc3));
@@ -303,6 +374,9 @@ static struct part *coco3_allocate(void) {
 
 	*mcc3 = (struct machine_coco3){0};
 
+	m->has_interface = coco3_has_interface;
+	m->attach_interface = coco3_attach_interface;
+
 	m->insert_cart = coco3_insert_cart;
 	m->remove_cart = coco3_remove_cart;
 	m->reset = coco3_reset;
@@ -312,12 +386,8 @@ static struct part *coco3_allocate(void) {
 	m->bp_add_n = coco3_bp_add_n;
 	m->bp_remove_n = coco3_bp_remove_n;
 
-	m->set_keyboard_type = coco3_set_keyboard_type;
 	m->set_pause = coco3_set_pause;
-	m->set_inverted_text = coco3_set_inverted_text;
 	m->get_interface = coco3_get_interface;
-	m->set_frameskip = coco3_set_frameskip;
-	m->set_ratelimit = coco3_set_ratelimit;
 
 	m->read_byte = coco3_read_byte;
 	m->write_byte = coco3_write_byte;
@@ -329,6 +399,51 @@ static struct part *coco3_allocate(void) {
 	return p;
 }
 
+static void create_ram(struct machine_coco3 *mcc3) {
+	struct machine *m = &mcc3->public;
+	struct part *p = &m->part;
+	struct machine_config *mc = m->config;
+
+	if (mc->ram < 512) {
+		mc->ram = 128;
+	} else if (mc->ram < 1024) {
+		mc->ram = 512;
+	} else if (mc->ram < 2048) {
+		mc->ram = 1024;
+	} else {
+		mc->ram = 2048;
+	}
+
+	// TODO: rejig everything to use 16-bit RAM!
+	struct ram_config ram_config = {
+		.d_width = 8,
+		.organisation = RAM_ORG(19, 9, 0),
+	};
+	if (mc->ram == 128) {
+		ram_config.organisation = RAM_ORG(17, 9, 0);
+	}
+	struct ram *ram = (struct ram *)part_create("ram", &ram_config);
+
+	unsigned nbanks = mc->ram / 512;
+	if (nbanks < 1)
+		nbanks = 1;
+	if (nbanks > 4)
+		nbanks = 4;
+
+	if (nbanks > 1) {
+		mcc3->dat.enabled = 1;
+		if (nbanks > 2)
+			mcc3->dat.mask = 0xc0;  // 2MB
+		else
+			mcc3->dat.mask = 0x40;  // 1MB
+	}
+
+	for (unsigned i = 0; i < nbanks; i++)
+		ram_add_bank(ram, i);
+
+	part_add_component(p, (struct part *)ram, "RAM");
+}
+
 static void coco3_initialise(struct part *p, void *options) {
         struct machine_config *mc = options;
         assert(mc != NULL);
@@ -349,6 +464,9 @@ static void coco3_initialise(struct part *p, void *options) {
 	part_add_component(&m->part, part_create("MC6821", NULL), "PIA0");
 	part_add_component(&m->part, part_create("MC6821", NULL), "PIA1");
 
+	// RAM
+	create_ram(mcc3);
+
 	// Keyboard
 	m->keyboard.type = mc->keymap;
 }
@@ -359,9 +477,9 @@ static _Bool coco3_finish(struct part *p) {
 	struct machine_config *mc = m->config;
 
 	// Interfaces
-	mcc3->vo = xroar_vo_interface;
-	mcc3->snd = xroar_ao_interface->sound_interface;
-	mcc3->tape_interface = xroar_tape_interface;
+	mcc3->vo = xroar.vo_interface;
+	mcc3->snd = xroar.ao_interface->sound_interface;
+	mcc3->tape_interface = xroar.tape_interface;
 
 	mcc3->tape_interface->default_paused = 0;
 
@@ -370,13 +488,46 @@ static _Bool coco3_finish(struct part *p) {
 	mcc3->CPU = (struct MC6809 *)part_component_by_id_is_a(p, "CPU", "MC6809");
 	mcc3->PIA0 = (struct MC6821 *)part_component_by_id_is_a(p, "PIA0", "MC6821");
 	mcc3->PIA1 = (struct MC6821 *)part_component_by_id_is_a(p, "PIA1", "MC6821");
+	mcc3->RAM = (struct ram *)part_component_by_id_is_a(p, "RAM", "ram");
 
 	// Check all required parts are attached
 	if (!mcc3->GIME || !mcc3->CPU || !mcc3->PIA0 || !mcc3->PIA1 ||
+	    !mcc3->RAM ||
 	    !mcc3->vo || !mcc3->snd || !mcc3->tape_interface) {
 		return 0;
 	}
 
+	// Register as a messenger client
+	mcc3->msgr_client_id = messenger_client_register();
+
+	// Join the ui messenger groups we're interested in
+	ui_messenger_preempt_group(mcc3->msgr_client_id, ui_tag_picture, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_picture, mcc3));
+	ui_messenger_preempt_group(mcc3->msgr_client_id, ui_tag_tv_input, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_tv_input, mcc3));
+	ui_messenger_preempt_group(mcc3->msgr_client_id, ui_tag_vdg_inverse, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_text_invert, mcc3));
+	ui_messenger_preempt_group(mcc3->msgr_client_id, ui_tag_keymap, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_keymap, mcc3));
+	ui_messenger_join_group(mcc3->msgr_client_id, ui_tag_frameskip, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_frameskip, mcc3));
+	ui_messenger_join_group(mcc3->msgr_client_id, ui_tag_ratelimit, MESSENGER_NOTIFY_DELEGATE(coco3_ui_set_ratelimit, mcc3));
+
+	// ROM
+	mcc3->ROM0 = rombank_new(8, 32768, 1);
+
+	// Super Extended Colour BASIC
+        if (mc->extbas_rom) {
+                sds tmp = romlist_find(mc->extbas_rom);
+                if (tmp) {
+                        rombank_load_image(mcc3->ROM0, 0, tmp, 0);
+                        sdsfree(tmp);
+                }
+	}
+
+	// Report and check CRC (Super Extended Colour BASIC)
+	rombank_report(mcc3->ROM0, "coco3", "Super Extended Colour BASIC");
+	mcc3->crc_secb = 0xb4c88d6c;  // Super Extended Colour BASIC (NTSC)
+	mcc3->has_secb = rombank_verify_crc(mcc3->ROM0, "Super Extended Colour BASIC", -1, "@coco3", xroar.cfg.force_crc_match, &mcc3->crc_secb);
+
+	// RAM configuration
+	ram_report(mcc3->RAM, "coco3", "total RAM");
+
 	// Connect any cartridge part
 	coco3_connect_cart(p);
 
@@ -388,40 +539,75 @@ static _Bool coco3_finish(struct part *p) {
 	// GIME reports changes in active area
 	mcc3->GIME->set_active_area = mcc3->vo->set_active_area;
 
+	// PAL CoCo 3s only emit the RGB palette, even over composite.  At the
+	// moment, there's no special handling for PAL composite.  If we decide
+	// to implement it properly, it's worth noting that the GIME crystal in
+	// PAL machines is 28.475MHz (2 * 14.2373MHz).
+
 	// Actual GIME clock is 2× this, but we treat it the same as a SAM
-	vo_set_cmp_fs(mcc3->vo, 1, VO_RENDER_FS_14_31818);
-	// PAL CoCo 3s only emit RGB, so don't bother tuning the composite
-	// for anything other than NTSC
-	vo_set_cmp_fsc(mcc3->vo, 1, VO_RENDER_FSC_3_579545);
-	vo_set_cmp_system(mcc3->vo, 1, VO_RENDER_SYSTEM_NTSC);
+	ui_update_state(-1, ui_tag_cmp_fs, VO_RENDER_FS_14_31818, NULL);
+	ui_update_state(-1, ui_tag_cmp_fsc, VO_RENDER_FSC_3_579545, NULL);
+	ui_update_state(-1, ui_tag_cmp_system, VO_RENDER_SYSTEM_NTSC, NULL);
 
 	// Bodge factor to bring centred active area in line with chroma
-	DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_phase_offset, 90);
+	DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_phase_offset, 0);
 
 	DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_lead_lag, 0., 100.);
-	for (int j = 0; j < 64; j++) {
-		int intensity = (j >> 4) & 3;
-		int phase = j & 15;
-		double y, b_y, r_y;
-		if (phase == 0 || j == 63) {
-			y = grey_intensity_map[intensity];
-			b_y = 0.0;
-			r_y = 0.0;
-		} else {
-			// Conversion gain of 0.6 for chroma through MC1372.
-			double hue = (2.0 * M_PI * (double)(phase+7.072)) / 14.0;
-			y = hue_intensity_map[intensity];
-			b_y = 0.5 * sin(hue) * 0.6;
-			r_y = 0.5 * cos(hue) * 0.6;
+
+	if (mc->tv_standard != TV_PAL) {
+		// Very slight tweak to the phase
+		double hue_offset = (2. * M_PI * 15.) / 1600.;
+		for (int intensity = 0; intensity < 4; intensity++) {
+			float grey = cmp_intensity[intensity].grey;
+			float clr_low = cmp_intensity[intensity].clr_low;
+			float clr_high = cmp_intensity[intensity].clr_high;
+
+			// Scale signal and add a little brightness.
+			grey     = grey     * (1.00 / CMP_V_PEAK) + 0.20;
+			clr_low  = clr_low  * (1.00 / CMP_V_PEAK) + 0.20;
+			clr_high = clr_high * (1.00 / CMP_V_PEAK) + 0.20;
+
+			for (int phase = 0; phase < 16; phase++) {
+				int c = (intensity * 16) + phase;
+
+				double y, b_y, r_y;
+				if (phase == 0 || c == 63) {
+					y = grey;
+					b_y = 0.0;
+					r_y = 0.0;
+				} else {
+					int ph = ((phase + (phase >= 12)) + 9) % 16;
+					double hue = ((2.0 * M_PI * (double)ph) / 16.0) + hue_offset;
+					b_y = ((clr_high - clr_low) / 2.0) * sin(hue) / 1.414;
+					r_y = ((clr_high - clr_low) / 2.0) * cos(hue) / 1.414;
+					y = (clr_high + clr_low) / 2.0;
+				}
+				// These values were measured at the composite port,
+				// already in U/V, so we need to scale to Pb/Pr before
+				// adding them to the palette.
+				b_y /= 0.504;
+				r_y /= 0.711;
+				DELEGATE_SAFE_CALL(mcc3->vo->palette_set_ybr, c, y, b_y, r_y);
+			}
 		}
-		DELEGATE_CALL(mcc3->vo->palette_set_ybr, j, y, b_y, r_y);
 	}
 
 	for (int j = 0; j < 64; j++) {
 		float r = rgb_intensity_map[((j>>4)&2)|((j>>2)&1)];
 		float g = rgb_intensity_map[((j>>3)&2)|((j>>1)&1)];
 		float b = rgb_intensity_map[((j>>2)&2)|((j>>0)&1)];
-		DELEGATE_CALL(mcc3->vo->palette_set_rgb, j, r, g, b);
+		DELEGATE_SAFE_CALL(mcc3->vo->palette_set_rgb, j, r, g, b);
+		if (mc->tv_standard == TV_PAL) {
+			// XXX need to check the actual PAL CoCo 3 RGB to composite
+			// circuit.  For now, bodge composite colours:
+			r *= 1.08;
+			g *= 1.08;
+			b *= 1.08;
+			float y = 0.299 * r + 0.587 * g + 0.114 * b;
+			float b_y = b - y;
+			float r_y = r - y;
+			DELEGATE_SAFE_CALL(mcc3->vo->palette_set_ybr, j, y, b_y, r_y);
+		}
 	}
 
 	DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_burst, 1, 0);    // Normal burst
@@ -459,103 +645,34 @@ static _Bool coco3_finish(struct part *p) {
 	// Tape
 	mcc3->tape_interface->update_audio = DELEGATE_AS1(void, float, update_audio_from_tape, mcc3);
 
-	mcc3->GIME->signal_hs = DELEGATE_AS1(void, bool, gime_hs, mcc3);
-	mcc3->GIME->signal_fs = DELEGATE_AS1(void, bool, gime_fs, mcc3);
-	mcc3->GIME->render_line = DELEGATE_AS3(void, unsigned, unsigned, uint8cp, gime_render_line, mcc3);
-	tcc1014_set_inverted_text(mcc3->GIME, mcc3->inverted_text);
-
-	// Load appropriate ROMs.  The CoCo 3 ROM is a single 32K image: Super
-	// Extended Colour BASIC.  There are NTSC and PAL variants though.
-
-	memset(mcc3->rom0, 0, sizeof(mcc3->rom0));
-
-	mcc3->has_secb = 0;
-	mcc3->crc_secb = 0;
-
-	/* ... Super Extended BASIC */
-	if (mc->extbas_rom) {
-		sds tmp = romlist_find(mc->extbas_rom);
-		if (tmp) {
-			int size = machine_load_rom(tmp, mcc3->rom0, sizeof(mcc3->rom0));
-			if (size > 0) {
-				mcc3->has_secb = 1;
-			}
-			sdsfree(tmp);
-		}
-	}
-
-	switch (mc->ram) {
-	case 2048:
-		mcc3->ram_size = 2048 * 1024;
-		mcc3->ram_mask = 0x7ffff;
-		mcc3->dat.enabled = 1;
-		mcc3->dat.mask = 0xc0 << 13;
-		//mcc3->dat.readable = 1;  // XXX make configurable
-		break;
-	case 1024:
-		mcc3->ram_size = 1024 * 1024;
-		mcc3->ram_mask = 0x7ffff;
-		mcc3->dat.enabled = 1;
-		mcc3->dat.mask = 0x40 << 13;
-		//mcc3->dat.readable = 1;  // XXX make configurable
-		break;
-	case 512:
-		mcc3->ram_size = 512 * 1024;
-		mcc3->ram_mask = 0x7ffff;
-		break;
-	default:
-	case 128:
-		mcc3->ram_size = 128 * 1024;
-		mcc3->ram_mask = 0x1ffff;
-		break;
-	}
-	if (!mcc3->ram) {
-		mcc3->ram = xmalloc(mcc3->ram_size);
-	}
-
-	// Check CRCs
-
-	if (mcc3->has_secb) {
-		_Bool forced = 0, valid_crc = 0;
-
-		mcc3->crc_secb = crc32_block(CRC32_RESET, mcc3->rom0, 0x8000);
-
-		valid_crc = crclist_match("@coco3", mcc3->crc_secb);
-
-		if (xroar_cfg.force_crc_match) {
-			mcc3->crc_secb = 0xb4c88d6c;  // CoCo 3 Super Extended BASIC
-			forced = 1;
-		}
-
-		(void)forced;  // avoid warning if no logging
-		LOG_DEBUG(1, "\tSuper Extended BASIC CRC = 0x%08x%s\n", mcc3->crc_secb, forced ? " (forced)" : "");
-		if (!valid_crc) {
-			LOG_WARN("Invalid CRC for Super Extended BASIC ROM\n");
-		}
-	}
-
 	// Default all PIA connections to unconnected (no source, no sink)
 	mcc3->PIA0->b.in_source = 0;
 	mcc3->PIA1->b.in_source = 0;
 	mcc3->PIA0->a.in_sink = mcc3->PIA0->b.in_sink = 0xff;
 	mcc3->PIA1->a.in_sink = mcc3->PIA1->b.in_sink = 0xff;
 
+	mcc3->GIME->signal_hs = DELEGATE_AS1(void, bool, gime_hs, mcc3);
+	mcc3->GIME->signal_fs = DELEGATE_AS1(void, bool, gime_fs, mcc3);
+	mcc3->GIME->render_line = DELEGATE_AS3(void, unsigned, unsigned, uint8cp, gime_render_line, mcc3);
+	ui_update_state(-1, ui_tag_tv_input, mc->tv_input, NULL);
+	ui_update_state(-1, ui_tag_vdg_inverse, mcc3->inverted_text, NULL);
+
 	// Until I implement serial, this appear to pull low by default
 	mcc3->PIA1->b.in_sink &= ~(1<<0);
 
 	// Keyboard interface
-	mcc3->keyboard.interface = keyboard_interface_new(m);
+	mcc3->keyboard.interface = keyboard_interface_new();
 	mcc3->keyboard.interface->update = DELEGATE_AS0(void, keyboard_update, mcc3);
 	keyboard_set_chord_mode(mcc3->keyboard.interface, keyboard_chord_mode_coco_basic);
-	keyboard_set_keymap(mcc3->keyboard.interface, m->keyboard.type);
+	ui_update_state(-1, ui_tag_keymap, m->keyboard.type, NULL);
 
 	// Printer interface
-	mcc3->printer_interface = printer_interface_new(m);
+	mcc3->printer_interface = printer_interface_new();
 
 #ifdef WANT_GDB_TARGET
 	// GDB
-	if (xroar_cfg.debug.gdb) {
-		mcc3->gdb_interface = gdb_interface_new(xroar_cfg.debug.gdb_ip, xroar_cfg.debug.gdb_port, m, mcc3->bp_session);
+	if (xroar.cfg.debug.gdb) {
+		mcc3->gdb_interface = gdb_interface_new(xroar.cfg.debug.gdb_ip, xroar.cfg.debug.gdb_port, m, mcc3->bp_session);
 	}
 #endif
 
@@ -563,11 +680,17 @@ static _Bool coco3_finish(struct part *p) {
 	update_sound_mux_source(mcc3);
 	sound_set_mux_enabled(mcc3->snd, PIA_VALUE_CB2(mcc3->PIA1));
 
+	mcc3->dat.readable = 1;
+
+	tcc1014_notify_mode(mcc3->GIME);
+
 	return 1;
 }
 
 static void coco3_free(struct part *p) {
 	struct machine_coco3 *mcc3 = (struct machine_coco3 *)p;
+	// Stop receiving any UI state updates
+	messenger_client_unregister(mcc3->msgr_client_id);
 #ifdef WANT_GDB_TARGET
 	if (mcc3->gdb_interface) {
 		gdb_interface_free(mcc3->gdb_interface);
@@ -576,36 +699,53 @@ static void coco3_free(struct part *p) {
 	if (mcc3->keyboard.interface) {
 		keyboard_interface_free(mcc3->keyboard.interface);
 	}
+	machine_bp_remove_list(&mcc3->public, coco3_print_breakpoint);
 	if (mcc3->printer_interface) {
 		printer_interface_free(mcc3->printer_interface);
 	}
 	if (mcc3->bp_session) {
 		bp_session_free(mcc3->bp_session);
 	}
-	free(mcc3->ram);
+	rombank_free(mcc3->ROM0);
 }
 
 static _Bool coco3_read_elem(void *sptr, struct ser_handle *sh, int tag) {
 	struct machine_coco3 *mcc3 = sptr;
+	struct machine *m = &mcc3->public;
+	struct part *p = &m->part;
 	size_t length = ser_data_length(sh);
 	switch (tag) {
 	case COCO3_SER_RAM:
-		if (!mcc3->public.config) {
-			return 0;
+		{
+			if (!mcc3->public.config) {
+				return 0;
+			}
+			if (length != ((unsigned)mcc3->public.config->ram * 1024)) {
+				LOG_MOD_WARN("coco3", "deserialise: RAM size mismatch %zu != %d\n", length, mcc3->public.config->ram * 1024);
+				return 0;
+			}
+			part_free(part_component_by_id_is_a(p, "RAM", "ram"));
+			create_ram(mcc3);
+			struct ram *ram = (struct ram *)part_component_by_id_is_a(p, "RAM", "ram");
+			ram_ser_read(ram, sh);
 		}
-		if (length != ((unsigned)mcc3->public.config->ram * 1024)) {
-			LOG_WARN("COCO3/DESERIALISE: RAM size mismatch %zd != %d\n", length, mcc3->public.config->ram * 1024);
-			return 0;
-		}
-		if (mcc3->ram) {
-			free(mcc3->ram);
-		}
-		mcc3->ram = ser_read_new(sh, length);
+		break;
+
+	case COCO3_SER_RAM_SIZE:
+	case COCO3_SER_RAM_MASK:
+		// no-op: RAM is now a sub-component
 		break;
 
 	case COCO3_SER_DAT_MMU_BANK:
 		for (int i = 0; i < 16; i++) {
-			mcc3->dat.mmu_bank[i] = ser_read_uint8(sh) << 13;
+			mcc3->dat.mmu_bank[i] = ser_read_uint8(sh);
+		}
+		break;
+
+	case COCO3_SER_DAT_VRAM_BANK:
+		{
+			uint32_t vbank = ser_read_vuint32(sh);
+			mcc3->dat.vram_bank = vbank >> 13;
 		}
 		break;
 
@@ -619,17 +759,24 @@ static _Bool coco3_write_elem(void *sptr, struct ser_handle *sh, int tag) {
 	struct machine_coco3 *mcc3 = sptr;
 	switch (tag) {
 	case COCO3_SER_RAM:
-		ser_write(sh, tag, mcc3->ram, mcc3->ram_size);
+	case COCO3_SER_RAM_SIZE:
+	case COCO3_SER_RAM_MASK:
+		// no-op: RAM is now a sub-component
 		break;
 
 	case COCO3_SER_DAT_MMU_BANK:
 		ser_write_tag(sh, tag, 16);
 		for (int i = 0; i < 16; i++) {
-			ser_write_uint8_untagged(sh, mcc3->dat.mmu_bank[i] >> 13);
+			ser_write_uint8_untagged(sh, mcc3->dat.mmu_bank[i]);
 		}
 		ser_write_close_tag(sh);
 		break;
 
+	case COCO3_SER_DAT_VRAM_BANK:
+		// compatibility
+		ser_write_vuint32(sh, tag, mcc3->dat.vram_bank << 13);
+		break;
+
 	default:
 		return 0;
 	}
@@ -638,6 +785,32 @@ static _Bool coco3_write_elem(void *sptr, struct ser_handle *sh, int tag) {
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
+static _Bool coco3_has_interface(struct part *p, const char *ifname) {
+	struct machine_coco3 *mp = (struct machine_coco3 *)p;
+
+	struct cart *c = mp->cart;
+	if (c) {
+		if (c->has_interface) {
+			return c->has_interface(c, ifname);
+		}
+	}
+
+	return 0;
+}
+
+static void coco3_attach_interface(struct part *p, const char *ifname, void *intf) {
+	struct machine_coco3 *mp = (struct machine_coco3 *)p;
+
+	struct cart *c = mp->cart;
+	if (c) {
+		if (c->attach_interface) {
+			return c->attach_interface(c, ifname, intf);
+		}
+	}
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
 static void coco3_connect_cart(struct part *p) {
 	struct machine_coco3 *mcc3 = (struct machine_coco3 *)p;
 	struct cart *c = (struct cart *)part_component_by_id_is_a(p, "cart", "dragon-cart");
@@ -665,20 +838,9 @@ static void coco3_remove_cart(struct machine *m) {
 
 static void coco3_reset(struct machine *m, _Bool hard) {
 	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
-	xroar_set_keyboard_type(1, m->keyboard.type);
+	struct machine_config *mc = m->config;
 	if (hard) {
-		// This initial RAM state is carried over from Dragon/CoCo1/2
-		// code, but might as well keep it.  As noted there, some
-		// non-uniform state is required for Wildcatting to work.
-		unsigned loc = 0, val = 0xff;
-		while (loc < (mcc3->ram_size-3)) {
-			mcc3->ram[loc++] = val;
-			mcc3->ram[loc++] = val;
-			mcc3->ram[loc++] = val;
-			mcc3->ram[loc++] = val;
-			if ((loc & 0xff) != 0)
-				val ^= 0xff;
-		}
+		ram_clear(mcc3->RAM, mc->ram_init);
 	}
 	mc6821_reset(mcc3->PIA0);
 	mc6821_reset(mcc3->PIA1);
@@ -689,6 +851,8 @@ static void coco3_reset(struct machine *m, _Bool hard) {
 	mcc3->CPU->reset(mcc3->CPU);
 	tape_reset(mcc3->tape_interface);
 	printer_reset(mcc3->printer_interface);
+	machine_bp_remove_list(m, coco3_print_breakpoint);
+	machine_bp_add_list(m, coco3_print_breakpoint, mcc3);
 }
 
 static enum machine_run_state coco3_run(struct machine *m, int ncycles) {
@@ -775,13 +939,15 @@ static void coco3_bp_remove_n(struct machine *m, struct machine_bp *list, int n)
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-static int coco3_set_keyboard_type(struct machine *m, int action) {
-	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
+static void coco3_ui_set_keymap(void *sptr, int tag, void *smsg) {
+	struct machine_coco3 *mcc3 = sptr;
+	struct machine *m = &mcc3->public;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_keymap);
+
 	int type = m->keyboard.type;
-	switch (action) {
-	case XROAR_QUERY:
-		break;
-	case XROAR_NEXT:
+	switch (uimsg->value) {
+	case UI_NEXT:
 		if (type == m->config->keymap) {
 			switch (m->config->keymap) {
 			case dkbd_layout_coco3:
@@ -796,16 +962,16 @@ static int coco3_set_keyboard_type(struct machine *m, int action) {
 			type = m->config->keymap;
 		}
 		break;
-	case XROAR_AUTO:
+	case UI_AUTO:
 		type = m->config->keymap;
 		break;
 	default:
-		type = action;
+		type = uimsg->value;
 		break;
 	}
 	m->keyboard.type = type;
 	keyboard_set_keymap(mcc3->keyboard.interface, type);
-	return type;
+	uimsg->value = type;
 }
 
 static _Bool coco3_set_pause(struct machine *m, int state) {
@@ -823,20 +989,66 @@ static _Bool coco3_set_pause(struct machine *m, int state) {
 	return mcc3->CPU->halt;
 }
 
-static _Bool coco3_set_inverted_text(struct machine *m, int action) {
-	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
-	switch (action) {
-	case 0: case 1:
-		mcc3->inverted_text = action;
-		break;
-	case 2:
-		mcc3->inverted_text = !mcc3->inverted_text;
-		break;
+static void coco3_ui_set_picture(void *sptr, int tag, void *smsg) {
+	struct machine_coco3 *mp = sptr;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_picture);
+	int picture = ui_msg_adjust_value_range(uimsg, mp->vo->picture, VO_PICTURE_ACTION,
+						VO_PICTURE_ZOOMED, VO_PICTURE_UNDERSCAN,
+						UI_ADJUST_FLAG_KEEP_AUTO);
+	vo_set_viewport(mp->vo, picture);
+}
+
+// TV input selection.  CoCo 3 allows RGB.  RGB monitors pull PIA1 PB3 low.
+
+static void coco3_ui_set_tv_input(void *sptr, int tag, void *smsg) {
+	struct machine_coco3 *mcc3 = sptr;
+	struct machine *m = &mcc3->public;
+	struct machine_config *mc = m->config;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_tv_input);
+
+	mc->tv_input = ui_msg_adjust_value_range(uimsg, mc->tv_input, TV_INPUT_SVIDEO,
+						 TV_INPUT_SVIDEO, TV_INPUT_RGB,
+						 UI_ADJUST_FLAG_CYCLE);
+	switch (mc->tv_input) {
 	default:
+	case TV_INPUT_SVIDEO:
+		vo_set_signal(mcc3->vo, VO_SIGNAL_SVIDEO);
+		tcc1014_set_composite(mcc3->GIME, 1);
+		mcc3->PIA1->b.in_sink |= (1<<3);
+		break;
+
+	case TV_INPUT_CMP_KBRW:
+		vo_set_signal(mcc3->vo, VO_SIGNAL_CMP);
+		DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_phase, 180);
+		tcc1014_set_composite(mcc3->GIME, 1);
+		mcc3->PIA1->b.in_sink |= (1<<3);
+		break;
+
+	case TV_INPUT_CMP_KRBW:
+		vo_set_signal(mcc3->vo, VO_SIGNAL_CMP);
+		DELEGATE_SAFE_CALL(mcc3->vo->set_cmp_phase, 0);
+		tcc1014_set_composite(mcc3->GIME, 1);
+		mcc3->PIA1->b.in_sink |= (1<<3);
+		break;
+
+	case TV_INPUT_RGB:
+		vo_set_signal(mcc3->vo, VO_SIGNAL_RGB);
+		tcc1014_set_composite(mcc3->GIME, 0);
+		mcc3->PIA1->b.in_sink &= ~(1<<3);
 		break;
 	}
+}
+
+static void coco3_ui_set_text_invert(void *sptr, int tag, void *smsg) {
+	struct machine_coco3 *mcc3 = sptr;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_vdg_inverse);
+
+	mcc3->inverted_text = ui_msg_adjust_value_range(uimsg, mcc3->inverted_text, 0,
+							0, 1, UI_ADJUST_FLAG_CYCLE);
 	tcc1014_set_inverted_text(mcc3->GIME, mcc3->inverted_text);
-	return mcc3->inverted_text;
 }
 
 /*
@@ -855,18 +1067,29 @@ static void *coco3_get_interface(struct machine *m, const char *ifname) {
 		return mcc3->printer_interface;
 	} else if (0 == strcmp(ifname, "tape-update-audio")) {
 		return update_audio_from_tape;
+	} else if (0 == strcmp(ifname, "bp-session")) {
+		return mcc3->bp_session;
 	}
 	return NULL;
 }
 
-static void coco3_set_frameskip(struct machine *m, unsigned fskip) {
-	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
-	mcc3->frameskip = fskip;
+static void coco3_ui_set_frameskip(void *sptr, int tag, void *smsg) {
+	(void)tag;
+	struct machine_coco3 *mp = sptr;
+	struct ui_state_message *uimsg = smsg;
+	mp->configured_frameskip = mp->frameskip = uimsg->value;
 }
 
-static void coco3_set_ratelimit(struct machine *m, _Bool ratelimit) {
-	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
-	sound_set_ratelimit(mcc3->snd, ratelimit);
+static void coco3_ui_set_ratelimit(void *sptr, int tag, void *smsg) {
+	(void)tag;
+	struct machine_coco3 *mp = sptr;
+	struct ui_state_message *uimsg = smsg;
+	sound_set_ratelimit(mp->snd, uimsg->value);
+	if (uimsg->value) {
+		mp->frameskip = mp->configured_frameskip;
+	} else {
+		mp->frameskip = 10;
+	}
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -888,14 +1111,16 @@ static void read_byte(struct machine_coco3 *mcc3, unsigned A) {
 	switch (mcc3->GIME->S) {
 	case 0:
 		// ROM
-		mcc3->CPU->D = mcc3->rom0[A & 0x7fff];
+		rombank_d8(mcc3->ROM0, A, &mcc3->CPU->D);
 		break;
+
 	case 1:
 		// CTS (cartridge ROM)
 		if (mcc3->cart) {
 			mcc3->CPU->D = mcc3->cart->read(mcc3->cart, A ^ 0x4000, 0, 1, mcc3->CPU->D);
 		}
 		break;
+
 	case 2:
 		// IO
 		if ((A & 32) == 0) {
@@ -904,34 +1129,40 @@ static void read_byte(struct machine_coco3 *mcc3, unsigned A) {
 			mcc3->CPU->D = mc6821_read(mcc3->PIA1, A);
 		}
 		break;
+
 	case 6:
 		// SCS (cartridge IO)
 		if (mcc3->cart)
 			mcc3->CPU->D = mcc3->cart->read(mcc3->cart, A, 1, 0, mcc3->CPU->D);
 		break;
+
 	case 7:
 		if (!mcc3->dat.enabled || !mcc3->dat.readable) {
 			break;
 		}
 		// Optional DAT board can optionally be read from
 		if (A == 0xff9b) {
-			mcc3->CPU->D = (mcc3->CPU->D & ~0x03) | (mcc3->dat.vram_bank >> 19);
+			mcc3->CPU->D = (mcc3->CPU->D & ~0x03) | (mcc3->dat.vram_bank >> 6);
 		} else if (A >= 0xffa0 && A < 0xffb0) {
-			mcc3->CPU->D = (mcc3->CPU->D & ~0xc0) | (mcc3->dat.mmu_bank[A & 15] >> 13);
+			mcc3->CPU->D = (mcc3->CPU->D & ~0xc0) | mcc3->dat.mmu_bank[A & 15];
 		}
+		break;
+
 	default:
 		// All the rest are N/C
 		break;
 	}
 	if (mcc3->GIME->RAS) {
-		uint32_t Z = mcc3->GIME->Z;
+		unsigned nWE = 1;
+		unsigned Zrow = mcc3->GIME->Z;
+		unsigned Zcol = mcc3->GIME->Z >> 9;
 		if (!mcc3->dat.MMUEN || (mcc3->dat.MC3 && A >= 0xfe00 && A < 0xff00)) {
 			// MMU not enabled, or CRM enabled and CRM region
-			mcc3->CPU->D = mcc3->ram[Z & mcc3->ram_mask];
+			ram_d8(mcc3->RAM, nWE, 0, Zrow, Zcol, &mcc3->CPU->D);
 		} else {
 			// Otherwise, translate
-			uint32_t bank = (A >> 13) | mcc3->dat.task;
-			mcc3->CPU->D = mcc3->ram[mcc3->dat.mmu_bank[bank] | (Z & mcc3->ram_mask)];
+			unsigned bank = mcc3->dat.mmu_bank[(A >> 13) | mcc3->dat.task] >> 6;
+			ram_d8(mcc3->RAM, nWE, bank, Zrow, Zcol, &mcc3->CPU->D);
 		}
 	}
 }
@@ -944,13 +1175,15 @@ static void write_byte(struct machine_coco3 *mcc3, unsigned A) {
 		switch (mcc3->GIME->S) {
 		case 0:
 			// ROM
-			mcc3->CPU->D = mcc3->rom0[A & 0x7fff];
+			rombank_d8(mcc3->ROM0, A, &mcc3->CPU->D);
 			break;
+
 		case 1:
 			// CTS (cartridge ROM)
 			if (mcc3->cart)
 				mcc3->cart->write(mcc3->cart, A ^ 0x4000, 0, 1, mcc3->CPU->D);
 			break;
+
 		case 2:
 			// IO
 			if ((A & 32) == 0) {
@@ -959,11 +1192,13 @@ static void write_byte(struct machine_coco3 *mcc3, unsigned A) {
 				mc6821_write(mcc3->PIA1, A, mcc3->CPU->D);
 			}
 			break;
+
 		case 6:
 			// SCS (cartridge IO)
 			if (mcc3->cart)
 				mcc3->cart->write(mcc3->cart, A, 1, 0, mcc3->CPU->D);
 			break;
+
 		case 7:
 			if (!mcc3->dat.enabled) {
 				break;
@@ -977,25 +1212,29 @@ static void write_byte(struct machine_coco3 *mcc3, unsigned A) {
 				mcc3->dat.task = (mcc3->CPU->D & 0x01) ? 8 : 0;
 			} else if (A == 0xff9b) {
 				// Video RAM limited to one of four 512K banks
-				mcc3->dat.vram_bank = ((mcc3->CPU->D & 0x03) << 19) & mcc3->dat.mask;
+				mcc3->dat.vram_bank = ((mcc3->CPU->D & 0x03) << 6) & mcc3->dat.mask;
 			} else if (A >= 0xffa0 && A < 0xffb0) {
 				// MMU banking extended by 2 bits
-				mcc3->dat.mmu_bank[A & 15] = ((mcc3->CPU->D & 0xc0) << 13) & mcc3->dat.mask;
+				mcc3->dat.mmu_bank[A & 15] = (mcc3->CPU->D & 0xc0) & mcc3->dat.mask;
 			}
+			break;
+
 		default:
 			// All the rest are N/C
 			break;
 		}
 	}
 	if (mcc3->GIME->RAS) {
-		uint32_t Z = mcc3->GIME->Z;
+		unsigned nWE = 0;
+		unsigned Zrow = mcc3->GIME->Z;
+		unsigned Zcol = mcc3->GIME->Z >> 9;
 		if (!mcc3->dat.MMUEN || (mcc3->dat.MC3 && A >= 0xfe00 && A < 0xff00)) {
 			// MMU not enabled, or CRM enabled and CRM region
-			mcc3->ram[Z & mcc3->ram_mask] = mcc3->CPU->D;
+			ram_d8(mcc3->RAM, nWE, 0, Zrow, Zcol, &mcc3->CPU->D);
 		} else {
 			// Otherwise, translate
-			uint32_t bank = (A >> 13) | mcc3->dat.task;
-			mcc3->ram[mcc3->dat.mmu_bank[bank] | (Z & mcc3->ram_mask)] = mcc3->CPU->D;
+			unsigned bank = mcc3->dat.mmu_bank[(A >> 13) | mcc3->dat.task] >> 6;
+			ram_d8(mcc3->RAM, nWE, bank, Zrow, Zcol, &mcc3->CPU->D);
 		}
 	}
 }
@@ -1079,13 +1318,25 @@ static void coco3_op_rts(struct machine *m) {
 
 static void coco3_dump_ram(struct machine *m, FILE *fd) {
 	struct machine_coco3 *mcc3 = (struct machine_coco3 *)m;
-	fwrite(mcc3->ram, mcc3->ram_size, 1, fd);
+	struct ram *ram = mcc3->RAM;
+	for (unsigned bank = 0; bank < ram->nbanks; bank++) {
+		if (ram->d && ram->d[bank]) {
+			fwrite(ram->d[bank], ram->bank_nelems, 1, fd);
+		}
+	}
 }
 
 static uint16_t fetch_vram(void *sptr, uint32_t A) {
 	struct machine_coco3 *mcc3 = sptr;
-	A = (mcc3->dat.vram_bank | (A & mcc3->ram_mask)) & ~1;
-	return (mcc3->ram[A] << 8) | mcc3->ram[A+1];
+	unsigned bank = mcc3->dat.vram_bank >> 6;
+	unsigned Zrow = A & ~1;
+	unsigned Zcol = A >> 9;
+	uint8_t *Vp = ram_a8(mcc3->RAM, bank, Zrow, Zcol);
+	static uint16_t D = 0;
+	if (Vp) {
+		D = (*Vp << 8) | *(Vp+1);
+	}
+	return D;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1192,6 +1443,23 @@ static void gime_render_line(void *sptr, unsigned burst, unsigned npixels, uint8
 	DELEGATE_CALL(mcc3->vo->render_line, burst, npixels, data);
 }
 
+// CoCo serial printing ROM hook.
+
+static void coco3_print_byte(void *sptr) {
+	struct machine_coco3 *mcc3 = sptr;
+	if (!mcc3->printer_interface) {
+		return;
+	}
+	// Not ROM?
+	if (tcc1014_decode(mcc3->GIME, mcc3->CPU->reg_pc) != 0) {
+		return;
+	}
+	int byte = MC6809_REG_A(mcc3->CPU);
+	printer_strobe(mcc3->printer_interface, 0, byte);
+	printer_strobe(mcc3->printer_interface, 1, byte);
+	mcc3->CPU->reg_pc = 0xa2df;
+}
+
 /* Sound output can feed back into the single bit sound pin when it's
  * configured as an input. */
 
