@@ -2,7 +2,7 @@
  *
  *  \brief GDB protocol support.
  *
- *  \copyright Copyright 2013-2022 Ciaran Anscomb
+ *  \copyright Copyright 2013-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -71,6 +71,7 @@
 
 #include <ctype.h>
 #include <errno.h>
+#include <limits.h>
 #include <pthread.h>
 #include <stdint.h>
 #include <stdlib.h>
@@ -126,6 +127,9 @@ struct gdb_interface_private {
 	pthread_t sock_thread;
 	int sockfd;
 
+	// Session state
+	_Bool no_ack_mode;
+
 	// Run state
 	enum gdb_run_state run_state;
 	pthread_cond_t run_state_cv;
@@ -148,9 +152,9 @@ enum gdb_error {
 static char in_packet[1025];
 static char packet[1025];
 
-static int read_packet(struct gdb_interface_private *gip, char *buffer, unsigned count);
-static int send_packet(struct gdb_interface_private *gip, const char *buffer, unsigned count);
-static int send_packet_string(struct gdb_interface_private *gip, const char *string);
+static ssize_t read_packet(struct gdb_interface_private *gip, void *buf, size_t count);
+static ssize_t send_packet(struct gdb_interface_private *gip, const void *buf, size_t count);
+static ssize_t send_packet_string(struct gdb_interface_private *gip, const char *string);
 static int send_char(struct gdb_interface_private *gip, char c);
 
 static void send_last_signal(struct gdb_interface_private *gip);  // ?
@@ -181,7 +185,7 @@ struct gdb_interface *gdb_interface_new(const char *hostname, const char *portna
 
 	struct MC6809 *cpu = (struct MC6809 *)part_component_by_id_is_a(&m->part, "CPU", "MC6809");
 	if (!cpu) {
-		LOG_WARN("GDB: MC6809 CPU not found - not enabling GDB support\n");
+		LOG_MOD_WARN("gdb", "MC6809 CPU not found - not enabling GDB support\n");
 		return NULL;
 	}
 
@@ -209,30 +213,30 @@ struct gdb_interface *gdb_interface_new(const char *hostname, const char *portna
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_family = AF_UNSPEC;
 	if (getaddrinfo(hostname, portname, &hints, &gip->info) < 0) {
-		LOG_WARN("gdb: getaddrinfo %s:%s failed\n", hostname, portname);
+		LOG_MOD_WARN("gdb", "getaddrinfo %s:%s failed\n", hostname, portname);
 		goto failed;
 	}
 	if (!gip->info) {
-		LOG_WARN("gdb: failed lookup %s:%s\n", hostname, portname);
+		LOG_MOD_WARN("gdb", "failed lookup %s:%s\n", hostname, portname);
 		goto failed;
 	}
 
 	// Create a socket...
 	gip->listenfd = socket(gip->info->ai_family, gip->info->ai_socktype, gip->info->ai_protocol);
 	if (gip->listenfd < 0) {
-		LOG_WARN("gdb: socket not created\n");
+		LOG_MOD_WARN("gdb", "socket not created\n");
 		goto failed;
 	}
 
 	// bind
 	if (bind(gip->listenfd, gip->info->ai_addr, gip->info->ai_addrlen) < 0) {
-		LOG_WARN("gdb: bind %s:%s failed\n", hostname, portname);
+		LOG_MOD_WARN("gdb", "bind %s:%s failed\n", hostname, portname);
 		goto failed;
 	}
 
 	// ... and listen
 	if (listen(gip->listenfd, 1) < 0) {
-		LOG_WARN("gdb: failed to listen to socket\n");
+		LOG_MOD_WARN("gdb", "failed to listen to socket\n");
 		goto failed;
 	}
 
@@ -240,7 +244,7 @@ struct gdb_interface *gdb_interface_new(const char *hostname, const char *portna
 	pthread_cond_init(&gip->run_state_cv, NULL);
 	pthread_create(&gip->sock_thread, NULL, handle_tcp_sock, gip);
 
-	LOG_DEBUG(1, "gdb: target listening on %s:%s\n", hostname, portname);
+	LOG_MOD_DEBUG(1, "gdb", "target listening on %s:%s\n", hostname, portname);
 
 	return (struct gdb_interface *)gip;
 
@@ -294,21 +298,22 @@ void gdb_run_unlock(struct gdb_interface *gi) {
 	pthread_mutex_unlock(&gip->run_state_mt);
 }
 
-static void gdb_handle_signal(struct gdb_interface_private *gip, int sig) {
+static void gdb_handle_signal(struct gdb_interface_private *gip, int sig, _Bool ack) {
 	gip->last_signal = sig;
-	send_last_signal(gip);
+	if (ack)
+		send_last_signal(gip);
 }
 
 void gdb_stop(struct gdb_interface *gi, int sig) {
 	struct gdb_interface_private *gip = (struct gdb_interface_private *)gi;
 	gip->run_state = gdb_run_state_stopped;
-	gdb_handle_signal(gip, sig);
+	gdb_handle_signal(gip, sig, 1);
 }
 
 void gdb_single_step(struct gdb_interface *gi) {
 	struct gdb_interface_private *gip = (struct gdb_interface_private *)gi;
 	gip->run_state = gdb_run_state_stopped;
-	gdb_handle_signal(gip, MACHINE_SIGTRAP);
+	gdb_handle_signal(gip, MACHINE_SIGTRAP, 1);
 	pthread_cond_signal(&gip->run_state_cv);
 }
 
@@ -321,18 +326,17 @@ static void gdb_machine_single_step(struct gdb_interface_private *gip) {
 	pthread_mutex_unlock(&gip->run_state_mt);
 }
 
-static void gdb_machine_signal(struct gdb_interface_private *gip, int sig) {
+static void gdb_machine_signal(struct gdb_interface_private *gip, int sig, _Bool ack) {
 	pthread_mutex_lock(&gip->run_state_mt);
 	if (gip->run_state == gdb_run_state_running) {
 		gip->machine->signal(gip->machine, sig);
 		gip->run_state = gdb_run_state_stopped;
-		gdb_handle_signal(gip, sig);
+		gdb_handle_signal(gip, sig, ack);
 	}
 	pthread_mutex_unlock(&gip->run_state_mt);
 }
 
 static void gdb_continue(struct gdb_interface_private *gip) {
-	//struct gdb_interface_private *gip = (struct gdb_interface_private *)gi;
 	pthread_mutex_lock(&gip->run_state_mt);
 	if (gip->run_state == gdb_run_state_stopped) {
 		gip->run_state = gdb_run_state_running;
@@ -372,37 +376,39 @@ static void *handle_tcp_sock(void *sptr) {
 		}
 
 		if (gip->sockfd < 0) {
-			LOG_WARN("gdb: accept() failed\n");
+			LOG_MOD_WARN("gdb", "accept() failed\n");
 			continue;
 		}
 		{
 			int flag = 1;
 			setsockopt(gip->sockfd, IPPROTO_TCP, TCP_NODELAY, (void const *)&flag, sizeof(flag));
 		}
-		LOG_DEBUG_GDB(LOG_GDB_CONNECT, "gdb: connection accepted\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_CONNECT, "gdb", "connection accepted\n");
 
-		gdb_machine_signal(gip, MACHINE_SIGINT);
+		gdb_machine_signal(gip, MACHINE_SIGINT, 0);
 		_Bool attached = 1;
 		while (attached) {
-			int l = read_packet(gip, in_packet, sizeof(in_packet));
+			ssize_t l = read_packet(gip, in_packet, sizeof(in_packet));
 			if (l == -GDBE_BREAK) {
-				LOG_DEBUG_GDB(LOG_GDB_PACKET, "gdb: BREAK\n");
-				gdb_machine_signal(gip, MACHINE_SIGINT);
+				LOG_MOD_DEBUG_GDB(LOG_GDB_PACKET, "gdb", "BREAK\n");
+				gdb_machine_signal(gip, MACHINE_SIGINT, 1);
 				continue;
 			} else if (l == -GDBE_BAD_CHECKSUM) {
-				if (send_char(gip, '-') < 0)
-					break;
-				continue;
+				if (!gip->no_ack_mode) {
+					if (send_char(gip, '-') < 0)
+						break;
+					continue;
+				}
 			} else if (l < 0) {
 				break;
 			}
 			if (logging.debug_gdb & LOG_GDB_PACKET) {
 				if (gip->run_state == gdb_run_state_stopped) {
-					LOG_PRINT("gdb: packet received: ");
+					LOG_MOD_PRINT("gdb", "packet received: ");
 				} else {
-					LOG_PRINT("gdb: packet ignored (send ^C first): ");
+					LOG_MOD_PRINT("gdb", "packet ignored (send ^C first): ");
 				}
-				for (unsigned i = 0; i < (unsigned)l; i++) {
+				for (ssize_t i = 0; i < l; ++i) {
 					if (isprint(in_packet[i])) {
 						LOG_PRINT("%c", in_packet[i]);
 					} else {
@@ -412,11 +418,11 @@ static void *handle_tcp_sock(void *sptr) {
 				LOG_PRINT("\n");
 			}
 			if (gip->run_state != gdb_run_state_stopped) {
-				if (send_char(gip, '-') < 0)
+				if (!gip->no_ack_mode && send_char(gip, '-') < 0)
 					break;
 				continue;
 			}
-			if (send_char(gip, '+') < 0)
+			if (!gip->no_ack_mode && send_char(gip, '+') < 0)
 				break;
 
 			char *args = &in_packet[1];
@@ -487,7 +493,7 @@ static void *handle_tcp_sock(void *sptr) {
 		}
 		close(gip->sockfd);
 		gdb_continue(gip);
-		LOG_DEBUG_GDB(LOG_GDB_CONNECT, "gdb: connection closed\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_CONNECT, "gdb", "connection closed\n");
 	}
 	return NULL;
 }
@@ -501,13 +507,17 @@ enum packet_state {
 	packet_csum1,
 };
 
-static int read_packet(struct gdb_interface_private *gip, char *buffer, unsigned count) {
+static ssize_t read_packet(struct gdb_interface_private *gip, void *buf, size_t count) {
+	char *cbuf = buf;
 	enum packet_state state = packet_wait;
-	unsigned length = 0;
+	size_t length = 0;
 	uint8_t packet_sum = 0;
 	uint8_t csum = 0;
-	char in_byte;
-	int tmp;
+
+	// Apply Linux read() limit
+	if (count > 0x7ffff000) {
+		count = 0x7ffff000;
+	}
 
 	while (1) {
 
@@ -526,11 +536,14 @@ static int read_packet(struct gdb_interface_private *gip, char *buffer, unsigned
 			}
 		}
 
-		int r = recv(gip->sockfd, &in_byte, 1, 0);
-		if (r < 0)
+		char in_byte;
+		ssize_t r = recv(gip->sockfd, &in_byte, 1, 0);
+		if (r < 0) {
 			return -GDBE_READ_ERROR;
-		if (r == 0)
+		}
+		if (r == 0) {
 			continue;
+		}
 
 		switch (state) {
 		case packet_wait:
@@ -546,40 +559,44 @@ static int read_packet(struct gdb_interface_private *gip, char *buffer, unsigned
 				state = packet_csum0;
 			} else {
 				if (length < (count - 1)) {
-					buffer[length++] = in_byte;
+					cbuf[length++] = in_byte;
 					packet_sum += (uint8_t)in_byte;
 				}
 			}
 			break;
 		case packet_csum0:
-			tmp = hexdigit(in_byte);
-			if (tmp < 0) {
-				state = packet_wait;
-			} else {
-				csum = tmp << 4;
-				state = packet_csum1;
+			{
+				int tmp = hexdigit(in_byte);
+				if (tmp < 0) {
+					state = packet_wait;
+				} else {
+					csum = tmp << 4;
+					state = packet_csum1;
+				}
 			}
 			break;
 		case packet_csum1:
-			tmp = hexdigit(in_byte);
-			if (tmp < 0) {
-				state = packet_wait;
-				break;
+			{
+				int tmp = hexdigit(in_byte);
+				if (tmp < 0) {
+					state = packet_wait;
+					break;
+				}
+				csum |= tmp;
 			}
-			csum |= tmp;
 			if (csum != packet_sum) {
 				if (logging.debug_gdb & LOG_GDB_CHECKSUM) {
-					LOG_PRINT("gdb: bad checksum in '");
-					if (isprint(buffer[0]))
-						LOG_PRINT("%c", buffer[0]);
+					LOG_MOD_PRINT("gdb", "bad checksum in '");
+					if (isprint(cbuf[0]))
+						LOG_PRINT("%c", cbuf[0]);
 					else
-						LOG_PRINT("0x%02x", buffer[0]);
+						LOG_PRINT("0x%02x", cbuf[0]);
 					LOG_PRINT("' packet.  Expected 0x%02x, got 0x%02x.\n",
 						  packet_sum, csum);
 				}
 				return -GDBE_BAD_CHECKSUM;
 			}
-			buffer[length] = 0;
+			cbuf[length] = 0;
 			return length;
 		}
 
@@ -588,42 +605,54 @@ static int read_packet(struct gdb_interface_private *gip, char *buffer, unsigned
 	return -GDBE_READ_ERROR;
 }
 
-static int send_packet(struct gdb_interface_private *gip, const char *buffer, unsigned count) {
+static ssize_t send_packet(struct gdb_interface_private *gip, const void *buf, size_t count) {
+	const char *cbuf = buf;
 	char tmpbuf[4];
 	uint8_t csum = 0;
+
+	// Apply Linux write() limit
+	if (count > 0x7ffff000) {
+		count = 0x7ffff000;
+	}
+
 	tmpbuf[0] = '$';
-	if (send(gip->sockfd, tmpbuf, 1, 0) < 0)
+	if (send(gip->sockfd, tmpbuf, 1, 0) < 0) {
 		return -GDBE_WRITE_ERROR;
-	for (unsigned i = 0; i < count; i++) {
-		csum += buffer[i];
-		switch (buffer[i]) {
+	}
+
+	for (size_t i = 0; i < count; ++i) {
+		csum += cbuf[i];
+		switch (cbuf[i]) {
 		case '#':
 		case '$':
 		case 0x7d:
 		case '*':
 			tmpbuf[0] = 0x7d;
-			tmpbuf[1] = buffer[i] ^ 0x20;
-			if (send(gip->sockfd, tmpbuf, 2, 0) < 0)
+			tmpbuf[1] = cbuf[i] ^ 0x20;
+			if (send(gip->sockfd, tmpbuf, 2, 0) < 0) {
 				return -GDBE_WRITE_ERROR;
+			}
 			break;
 		default:
-			if (send(gip->sockfd, &buffer[i], 1, 0) < 0)
+			if (send(gip->sockfd, &cbuf[i], 1, 0) < 0) {
 				return -GDBE_WRITE_ERROR;
+			}
 			break;
 		}
 	}
 	snprintf(tmpbuf, sizeof(tmpbuf), "#%02x", (unsigned)csum);
-	if (send(gip->sockfd, tmpbuf, 3, 0) < 0)
+	if (send(gip->sockfd, tmpbuf, 3, 0) < 0) {
 		return -GDBE_WRITE_ERROR;
+	}
 	// the reply ("+" or "-") will be discarded by the next read_packet
 
 	if (logging.debug_gdb & LOG_GDB_PACKET) {
-		LOG_PRINT("gdb: packet sent: ");
-		for (unsigned i = 0; i < (unsigned)count; i++) {
-			if (isprint(buffer[i])) {
-				LOG_PRINT("%c", buffer[i]);
+		LOG_MOD_PRINT("gdb", "packet sent: ");
+		for (size_t i = 0; i < count; ++i) {
+			if (isprint(cbuf[i])) {
+				LOG_PRINT("%c", cbuf[i]);
 			} else {
-				LOG_PRINT("\\%o", buffer[i] & 0xff);
+				LOG_PRINT("\\%o", cbuf[i] & 0xff);
 			}
 		}
 		LOG_PRINT("\n");
@@ -632,25 +661,32 @@ static int send_packet(struct gdb_interface_private *gip, const char *buffer, un
 	return count;
 }
 
-static int send_packet_string(struct gdb_interface_private *gip, const char *string) {
-	unsigned count = strlen(string);
-	return send_packet(gip, string, count);
+static ssize_t send_packet_string(struct gdb_interface_private *gip, const char *string) {
+	return send_packet(gip, string, strlen(string));
 }
 
-static int send_packet_hexstring(struct gdb_interface_private *gip, const char *string) {
-	unsigned count = strlen(string);
+static ssize_t send_packet_hexstring(struct gdb_interface_private *gip, const char *string) {
+	size_t count = strlen(string);
+
+	// Apply Linux write() limit
+	if (count > 0X3ffff800) {
+		count = 0X3ffff800;
+	}
+
 	char *hs = xmalloc(count * 2);
 	char *hsp = hs;
-	for (unsigned i = 0; i < count; i++)
+	for (size_t i = 0; i < count; ++i) {
 		hsp += sprintf(hsp, "%02x", string[i]);
-	int ret = send_packet(gip, hs, count * 2);
+	}
+	ssize_t ret = send_packet(gip, hs, count * 2);
 	free(hs);
 	return ret;
 }
 
 static int send_char(struct gdb_interface_private *gip, char c) {
-	if (send(gip->sockfd, &c, 1, 0) < 0)
+	if (send(gip->sockfd, &c, 1, 0) < 0) {
 		return -GDBE_WRITE_ERROR;
+	}
 	return GDBE_OK;
 }
 
@@ -744,7 +780,7 @@ static void send_memory(struct gdb_interface_private *gip, char *args) {
 	if (send(gip->sockfd, packet, 3, 0) < 0)
 		return;
 	// the ACK ("+") or NAK ("-") will be discarded by the next read_packet
-	LOG_DEBUG_GDB(LOG_GDB_PACKET, "gdb: packet sent (binary): %u bytes\n", length);
+	LOG_MOD_DEBUG_GDB(LOG_GDB_PACKET, "gdb", "packet sent (binary): %u bytes\n", length);
 	return;
 error:
 	send_packet(gip, NULL, 0);
@@ -870,10 +906,13 @@ static int qRcmd(struct gdb_interface_private *gip, char *args) {
 	}
 	*np = '\0';
 
-	/* parse our own gdb "monitor" command */
-	char *cmd = strsep(&args, " ");
-	if (!args)		/* no arguments to our cmd */
-		args = "";	/* always printable below */
+	// Parse our own gdb "monitor" command.  If no args found, use an empty
+	// string for printing later.  Note: The temporary assignment to argptr
+	// here works around the GCC -fanalyzer false positive
+	// -Wanalyzer-deref-before-check.
+	char *argptr = args;
+	char *cmd = strsep(&argptr, " ");
+	args = argptr ? argptr : "";
 
 	char reply[255];
 	*reply = '\0';
@@ -894,10 +933,14 @@ static int qRcmd(struct gdb_interface_private *gip, char *args) {
 
 static void general_query(struct gdb_interface_private *gip, char *args) {
 	if (0 == strncmp(args, "Rcmd", 4)) {
-		/* this query uses comma instead of colon as separator */
-		strsep(&args, ",");
-		if (!args || qRcmd(gip, args))
+		// This query uses comma instead of colon as separator.
+		// Copying args to argptr works around the GCC -fanalyzer false
+		// positive -Wanalyzer-deref-before-check.
+		char *argptr = args;
+		strsep(&argptr, ",");
+		if (!argptr || qRcmd(gip, argptr)) {
 			send_packet_string(gip, "E00");
+		}
 		return;
 	}
 	char *query = strsep(&args, ":");
@@ -905,21 +948,21 @@ static void general_query(struct gdb_interface_private *gip, char *args) {
 		query += 6;
 #ifdef WANT_MACHINE_ARCH_DRAGON
 		if (0 == strcmp(query, "sam") && gip->sam) {
-			LOG_DEBUG_GDB(LOG_GDB_QUERY, "gdb: query: xroar.sam\n");
+			LOG_MOD_DEBUG_GDB(LOG_GDB_QUERY, "gdb", "query: xroar.sam\n");
 			sprintf(packet, "%04x", mc6883_get_register(gip->sam));
 			send_packet(gip, packet, 4);
 		}
 		return;
 #endif
-		LOG_DEBUG_GDB(LOG_GDB_QUERY, "gdb: query: unknown xroar vendor query\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_QUERY, "gdb", "query: unknown xroar vendor query\n");
 	} else if (0 == strcmp(query, "Supported")) {
-		LOG_DEBUG_GDB(LOG_GDB_QUERY, "gdb: query: Supported\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_QUERY, "gdb", "query: Supported\n");
 		send_supported(gip, args);
 	} else if (0 == strcmp(query, "Attached")) {
-		LOG_DEBUG_GDB(LOG_GDB_QUERY, "gdb: query: Attached\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_QUERY, "gdb", "query: Attached\n");
 		send_packet_string(gip, "1");
 	} else {
-		LOG_DEBUG_GDB(LOG_GDB_QUERY, "gdb: query: unknown query\n");
+		LOG_MOD_DEBUG_GDB(LOG_GDB_QUERY, "gdb", "query: unknown query\n");
 		send_packet(gip, NULL, 0);
 	}
 }
@@ -935,6 +978,10 @@ static void general_set(struct gdb_interface_private *gip, char *args) {
 			return;
 		}
 #endif
+	} else if (0 == strcmp(set, "StartNoAckMode")) {
+		gip->no_ack_mode = 1;
+		send_packet_string(gip, "OK");
+		return;
 	}
 	send_packet(gip, NULL, 0);
 	return;
@@ -955,10 +1002,10 @@ static void add_breakpoint(struct gdb_interface_private *gip, char *args) {
 		goto error;
 	unsigned addr = strtoul(addr_str, NULL, 16);
 	if (type <= 1) {
-		bp_hbreak_add(gip->bp_session, addr, 0, 0);
+		bp_hbreak_add(gip->bp_session, addr);
 	} else {
 		unsigned nbytes = strtoul(kind_str, NULL, 16);
-		bp_wp_add(gip->bp_session, type, addr, nbytes, 0, 0);
+		bp_wp_add(gip->bp_session, type, addr, nbytes);
 	}
 	send_packet_string(gip, "OK");
 	return;
@@ -981,10 +1028,10 @@ static void remove_breakpoint(struct gdb_interface_private *gip, char *args) {
 		goto error;
 	unsigned addr = strtoul(addr_str, NULL, 16);
 	if (type <= 1) {
-		bp_hbreak_remove(gip->bp_session, addr, 0, 0);
+		bp_hbreak_remove(gip->bp_session, addr);
 	} else {
 		unsigned nbytes = strtoul(kind_str, NULL, 16);
-		bp_wp_remove(gip->bp_session, type, addr, nbytes, 0, 0);
+		bp_wp_remove(gip->bp_session, type, addr, nbytes);
 	}
 	send_packet_string(gip, "OK");
 	return;
