@@ -2,7 +2,7 @@
  *
  *  \brief Motorola MC6801/6803 CPUs.
  *
- *  \copyright Copyright 2021-2022 Ciaran Anscomb
+ *  \copyright Copyright 2021-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -63,10 +63,10 @@
 #endif
 
 static const struct ser_struct ser_struct_mc6801_port[] = {
-	SER_ID_STRUCT_ELEM(1, ser_type_uint8, struct MC6801_port, out_source),
-	SER_ID_STRUCT_ELEM(2, ser_type_uint8, struct MC6801_port, out_sink),
-	SER_ID_STRUCT_ELEM(3, ser_type_uint8, struct MC6801_port, in_source),
-	SER_ID_STRUCT_ELEM(4, ser_type_uint8, struct MC6801_port, in_sink),
+	SER_ID_STRUCT_ELEM(1, struct MC6801_port, out_source),
+	SER_ID_STRUCT_ELEM(2, struct MC6801_port, out_sink),
+	SER_ID_STRUCT_ELEM(3, struct MC6801_port, in_source),
+	SER_ID_STRUCT_ELEM(4, struct MC6801_port, in_sink),
 };
 
 static const struct ser_struct_data mc6801_port_ser_struct_data = {
@@ -78,46 +78,46 @@ static const struct ser_struct_data mc6801_port_ser_struct_data = {
 #define MC6801_SER_RAM (14)
 
 static const struct ser_struct ser_struct_mc6801[] = {
-	SER_ID_STRUCT_ELEM(1, ser_type_bool, struct MC6801, nmi),
-	SER_ID_STRUCT_ELEM(2, ser_type_bool, struct MC6801, irq1),
-	SER_ID_STRUCT_ELEM(3, ser_type_uint8, struct MC6801, D),
+	SER_ID_STRUCT_ELEM(1, struct MC6801, nmi),
+	SER_ID_STRUCT_ELEM(2, struct MC6801, irq1),
+	SER_ID_STRUCT_ELEM(3, struct MC6801, D),
 
 	SER_ID_STRUCT_SUBSTRUCT(4, struct MC6801, port1, &mc6801_port_ser_struct_data),
 	SER_ID_STRUCT_SUBSTRUCT(5, struct MC6801, port2, &mc6801_port_ser_struct_data),
 
-	SER_ID_STRUCT_ELEM(6, ser_type_unsigned, struct MC6801, state),
-	SER_ID_STRUCT_ELEM(7, ser_type_bool, struct MC6801, running),
+	SER_ID_STRUCT_ELEM(6, struct MC6801, state),
+	SER_ID_STRUCT_ELEM(7, struct MC6801, running),
 
-	SER_ID_STRUCT_ELEM(8, ser_type_uint8, struct MC6801, reg_cc),
-	SER_ID_STRUCT_ELEM(9, ser_type_uint16, struct MC6801, reg_d),
-	SER_ID_STRUCT_ELEM(10, ser_type_uint16, struct MC6801, reg_x),
-	SER_ID_STRUCT_ELEM(11, ser_type_uint16, struct MC6801, reg_sp),
-	SER_ID_STRUCT_ELEM(12, ser_type_uint16, struct MC6801, reg_pc),
+	SER_ID_STRUCT_ELEM(8, struct MC6801, reg_cc),
+	SER_ID_STRUCT_ELEM(9, struct MC6801, reg_d),
+	SER_ID_STRUCT_ELEM(10, struct MC6801, reg_x),
+	SER_ID_STRUCT_ELEM(11, struct MC6801, reg_sp),
+	SER_ID_STRUCT_ELEM(12, struct MC6801, reg_pc),
 
 	SER_ID_STRUCT_UNHANDLED(MC6801_SER_REG),
 	SER_ID_STRUCT_UNHANDLED(MC6801_SER_RAM),
 
-	SER_ID_STRUCT_ELEM(15, ser_type_uint8, struct MC6801, itmp),
-	SER_ID_STRUCT_ELEM(16, ser_type_bool, struct MC6801, nmi_latch),
-	SER_ID_STRUCT_ELEM(17, ser_type_bool, struct MC6801, nmi_active),
-	SER_ID_STRUCT_ELEM(18, ser_type_bool, struct MC6801, irq1_latch),
-	SER_ID_STRUCT_ELEM(19, ser_type_bool, struct MC6801, irq1_active),
-	SER_ID_STRUCT_ELEM(20, ser_type_bool, struct MC6801, irq2_latch),
-	SER_ID_STRUCT_ELEM(21, ser_type_bool, struct MC6801, irq2_active),
-	SER_ID_STRUCT_ELEM(22, ser_type_uint8, struct MC6801, ICF),
-	SER_ID_STRUCT_ELEM(23, ser_type_uint8, struct MC6801, OCF),
-	SER_ID_STRUCT_ELEM(24, ser_type_uint8, struct MC6801, TOF),
+	SER_ID_STRUCT_ELEM(15, struct MC6801, itmp),
+	SER_ID_STRUCT_ELEM(16, struct MC6801, nmi_latch),
+	SER_ID_STRUCT_ELEM(17, struct MC6801, nmi_active),
+	SER_ID_STRUCT_ELEM(18, struct MC6801, irq1_latch),
+	SER_ID_STRUCT_ELEM(19, struct MC6801, irq1_active),
+	SER_ID_STRUCT_ELEM(20, struct MC6801, irq2_latch),
+	SER_ID_STRUCT_ELEM(21, struct MC6801, irq2_active),
+	SER_ID_STRUCT_ELEM(22, struct MC6801, ICF),
+	SER_ID_STRUCT_ELEM(23, struct MC6801, OCF),
+	SER_ID_STRUCT_ELEM(24, struct MC6801, TOF),
 
-	SER_ID_STRUCT_ELEM(25, ser_type_uint16, struct MC6801, counter),
-	SER_ID_STRUCT_ELEM(26, ser_type_uint16, struct MC6801, output_compare),
-	SER_ID_STRUCT_ELEM(27, ser_type_bool, struct MC6801, output_compare_inhibit),
+	SER_ID_STRUCT_ELEM(25, struct MC6801, counter),
+	SER_ID_STRUCT_ELEM(26, struct MC6801, output_compare),
+	SER_ID_STRUCT_ELEM(27, struct MC6801, output_compare_inhibit),
 
-	SER_ID_STRUCT_ELEM(28, ser_type_uint8, struct MC6801, ICF_read),
-	SER_ID_STRUCT_ELEM(29, ser_type_uint8, struct MC6801, OCF_read),
-	SER_ID_STRUCT_ELEM(30, ser_type_uint8, struct MC6801, TOF_read),
-	SER_ID_STRUCT_ELEM(31, ser_type_uint8, struct MC6801, counter_lsb_buf),
+	SER_ID_STRUCT_ELEM(28, struct MC6801, ICF_read),
+	SER_ID_STRUCT_ELEM(29, struct MC6801, OCF_read),
+	SER_ID_STRUCT_ELEM(30, struct MC6801, TOF_read),
+	SER_ID_STRUCT_ELEM(31, struct MC6801, counter_lsb_buf),
 
-	SER_ID_STRUCT_ELEM(32, ser_type_bool, struct MC6801, is_6801),
+	SER_ID_STRUCT_ELEM(32, struct MC6801, is_6801),
 };
 
 static _Bool mc6801_read_elem(void *sptr, struct ser_handle *sh, int tag);
@@ -232,8 +232,8 @@ static const struct partdb_entry_funcs mc6801_funcs = {
 	.is_a = mc6801_is_a,
 };
 
-const struct partdb_entry mc6801_part = { .name = "MC6801", .funcs = &mc6801_funcs };
-const struct partdb_entry mc6803_part = { .name = "MC6803", .funcs = &mc6801_funcs };
+const struct partdb_entry mc6801_part = { .name = "MC6801", .description = "Motorola | MC6801 CPU", .funcs = &mc6801_funcs };
+const struct partdb_entry mc6803_part = { .name = "MC6803", .description = "Motorola | MC6803 CPU", .funcs = &mc6801_funcs };
 
 static struct part *mc6801_allocate(void) {
 	struct MC6801 *cpu = part_new(sizeof(*cpu));
@@ -358,7 +358,7 @@ static void mc6801_run(struct MC6801 *cpu) {
 			cpu->counter_lsb_buf = 0;
 			cpu->output_compare = 0xffff;
 #ifdef TRACE
-			if (logging.trace_cpu) {
+			if (UNLIKELY(logging.trace_cpu)) {
 				mc6801_trace_irq(cpu->tracer, MC6801_INT_VEC_RESET);
 			}
 #endif
@@ -368,7 +368,7 @@ static void mc6801_run(struct MC6801 *cpu) {
 			continue;
 
 		case mc6801_state_label_a:
-			if (cpu->nmi_active) {
+			if (UNLIKELY(cpu->nmi_active)) {
 				REG_CC = (REG_CC & ~CC_I) | cpu->itmp;
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
@@ -377,7 +377,7 @@ static void mc6801_run(struct MC6801 *cpu) {
 				cpu->state = mc6801_state_dispatch_irq;
 				continue;
 			}
-			if (!(REG_CC & CC_I) && (cpu->irq1_active || cpu->irq2_active)) {
+			if (UNLIKELY(!(REG_CC & CC_I) && (cpu->irq1_active || cpu->irq2_active))) {
 				REG_CC = (REG_CC & ~CC_I) | cpu->itmp;
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
@@ -862,8 +862,8 @@ static void mc6801_run(struct MC6801 *cpu) {
 				case 0x7: tmp1 = op_asr_v(cpu, tmp1); break; // ASR, ASRA, ASRB
 				case 0x8: tmp1 = op_asl(cpu, tmp1); break; // ASL, ASLA, ASLB
 				case 0x9: tmp1 = op_rol(cpu, tmp1); break; // ROL, ROLA, ROLB
-				case 0xb: // DEC illegal  XXX wrong flags
 				case 0xa: tmp1 = op_dec(cpu, tmp1); break; // DEC, DECA, DECB
+				case 0xb: tmp1 = op_dec_nc(cpu, tmp1); break; // DC*, DCA*, DCB*
 				case 0xc: tmp1 = op_inc(cpu, tmp1); break; // INC, INCA, INCB
 				case 0xd: tmp1 = op_tst_c(cpu, tmp1); break; // TST, TSTA, TSTB
 				case 0xf: tmp1 = op_clr(cpu, tmp1); break; // CLR, CLRA, CLRB
@@ -959,7 +959,7 @@ static void mc6801_run(struct MC6801 *cpu) {
 				case 0x4: tmp1 = op_and(cpu, tmp1, tmp2); break; // ANDA, ANDB
 				case 0x5: (void)op_and(cpu, tmp1, tmp2); break; // BITA, BITB
 				case 0x6: tmp1 = op_ld(cpu, 0, tmp2); break; // LDAA, LDAB
-				case 0x7: break;  // XXX illegal op 0x87, 0xc7
+				case 0x7: tmp1 = op_ld(cpu, 0xff, 0xff); break;  // I87*, IC7*
 				case 0x8: tmp1 = op_eor(cpu, tmp1, tmp2); break; // EORA, EORB
 				case 0x9: tmp1 = op_adc(cpu, tmp1, tmp2); break; // ADCA, ADCB
 				case 0xa: tmp1 = op_or(cpu, tmp1, tmp2); break; // ORA, ORB
@@ -1280,7 +1280,7 @@ static uint16_t fetch_word_notrace(struct MC6801 *cpu, uint16_t a) {
 static uint8_t fetch_byte(struct MC6801 *cpu, uint16_t a) {
 	uint8_t v = fetch_byte_notrace(cpu, a);
 #ifdef TRACE
-	if (logging.trace_cpu) {
+	if (UNLIKELY(logging.trace_cpu)) {
 		mc6801_trace_byte(cpu->tracer, v, a);
 	}
 #endif
@@ -1291,7 +1291,7 @@ static uint16_t fetch_word(struct MC6801 *cpu, uint16_t a) {
 #ifndef TRACE
 	return fetch_word_notrace(cpu, a);
 #else
-	if (!logging.trace_cpu) {
+	if (LIKELY(!logging.trace_cpu)) {
 		return fetch_word_notrace(cpu, a);
 	}
 	unsigned v0 = fetch_byte_notrace(cpu, a);
@@ -1353,7 +1353,7 @@ static void stack_irq_registers(struct MC6801 *cpu) {
 
 static void take_interrupt(struct MC6801 *cpu, uint16_t vec) {
 #ifdef TRACE
-	if (logging.trace_cpu) {
+	if (UNLIKELY(logging.trace_cpu)) {
 		mc6801_trace_irq(cpu->tracer, vec);
 	}
 #endif
@@ -1365,7 +1365,7 @@ static void take_interrupt(struct MC6801 *cpu, uint16_t vec) {
 
 static void instruction_posthook(struct MC6801 *cpu) {
 #ifdef TRACE
-	if (logging.trace_cpu) {
+	if (UNLIKELY(logging.trace_cpu)) {
 		mc6801_trace_print(cpu->tracer);
 	}
 #endif
