@@ -2,7 +2,7 @@
  *
  *  \brief Hitach HD6309 CPU.
  *
- *  \copyright Copyright 2012-2023 Ciaran Anscomb
+ *  \copyright Copyright 2012-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -32,6 +32,9 @@
  *
  *  - Undocumented 6309 Behaviours, David Banks [hoglet67]
  *    https://github.com/hoglet67/6809Decoder/wiki/Undocumented-6309-Behaviours
+ *
+ *  - Tim Lindner's Hitachi 6309 Fuzzing Project
+ *    https://github.com/tlindner/Fuzz6309
  */
 
 #include "top-config.h"
@@ -50,7 +53,14 @@
 #include "serialise.h"
 
 #ifdef TRACE
+// Tracing supported:
 #include "hd6309_trace.h"
+#define HD6309_TRACE_VECTOR(c) if (UNLIKELY(logging.trace_cpu)) { hd6309_trace_vector((c)->tracer); }
+#define HD6309_TRACE_INSTRUCTION(c) if (UNLIKELY(logging.trace_cpu)) { hd6309_trace_instruction((c)->tracer); }
+#else
+// Tracing not supported - no-op macros:
+#define HD6309_TRACE_VECTOR(c)
+#define HD6309_TRACE_INSTRUCTION(c)
 #endif
 
 #define HD6309_SER_TFM_SRC  (6)
@@ -59,16 +69,16 @@
 static const struct ser_struct ser_struct_hd6309[] = {
 	SER_ID_STRUCT_NEST(1,  &mc6809_ser_struct_data), // 1
 
-	SER_ID_STRUCT_ELEM(2,  ser_type_unsigned, struct HD6309, state),
-	SER_ID_STRUCT_ELEM(3,  ser_type_uint16,   struct HD6309, reg_w),
-	SER_ID_STRUCT_ELEM(4,  ser_type_uint8,    struct HD6309, reg_md),
-	SER_ID_STRUCT_ELEM(5,  ser_type_uint16,   struct HD6309, reg_v),
-	SER_ID_STRUCT_ELEM(8,  ser_type_uint8,    struct HD6309, reg_m),  // replaces tfm_data
+	SER_ID_STRUCT_ELEM(2,  struct HD6309, state),
+	SER_ID_STRUCT_ELEM(3,  struct HD6309, reg_w),
+	SER_ID_STRUCT_ELEM(4,  struct HD6309, reg_md),
+	SER_ID_STRUCT_ELEM(5,  struct HD6309, reg_v),
+	SER_ID_STRUCT_ELEM(8,  struct HD6309, reg_m),  // replaces tfm_data
 
 	SER_ID_STRUCT_UNHANDLED(HD6309_SER_TFM_SRC),
 	SER_ID_STRUCT_UNHANDLED(HD6309_SER_TFM_DEST),
-	SER_ID_STRUCT_ELEM(9,  ser_type_uint16,   struct HD6309, tfm_src_mod),
-	SER_ID_STRUCT_ELEM(10, ser_type_uint16,   struct HD6309, tfm_dest_mod),
+	SER_ID_STRUCT_ELEM(9,  struct HD6309, tfm_src_mod),
+	SER_ID_STRUCT_ELEM(10, struct HD6309, tfm_dest_mod),
 };
 
 static _Bool hd6309_read_elem(void *sptr, struct ser_handle *sh, int tag);
@@ -226,7 +236,7 @@ static const struct partdb_entry_funcs hd6309_funcs = {
         .is_a = hd6309_is_a,
 };
 
-const struct partdb_entry hd6309_part = { .name = "HD6309", .funcs = &hd6309_funcs };
+const struct partdb_entry hd6309_part = { .name = "HD6309", .description = "Hitachi | HD6309E CPU", .funcs = &hd6309_funcs };
 
 static struct part *hd6309_allocate(void) {
 	struct HD6309 *hcpu = part_new(sizeof(*hcpu));
@@ -242,6 +252,24 @@ static struct part *hd6309_allocate(void) {
 	cpu->run = hd6309_run;
 	cpu->mem_cycle = DELEGATE_DEFAULT2(void, bool, uint16);
 
+	// Tested: on power on, all registers have random-ish values, although
+	// it definitely seems to err towards bits being set in my environment.
+	// I've seen all bits of CC set on power up *except* E, but no bits set
+	// is valid, so I'll leave CC clear.
+	//
+	// CC has F and I set as part of reset.  DP is explicitly cleared on
+	// reset, but other registers are left untouched.
+	//
+	// Not tested yet: extra 6309 register initialisation.
+
+	REG_D = 0xffff;
+	REG_X = 0xffff;
+	REG_Y = 0xffff;
+	REG_U = 0xffff;
+	REG_S = 0xffff;
+	REG_W = 0xffff;
+	REG_V = 0xffff;
+
 #ifdef TRACE
 	// Tracing
 	hcpu->tracer = hd6309_trace_new(hcpu);
@@ -331,21 +359,15 @@ static void hd6309_run(struct MC6809 *cpu) {
 			cpu->firq_active = 0;
 			cpu->irq_active = 0;
 			hcpu->state = hd6309_state_reset_check_halt;
-#ifdef TRACE
-			if (logging.trace_cpu) {
-				hd6309_trace_irq(hcpu->tracer, MC6809_INT_VEC_RESET);
-			}
-#endif
+			HD6309_TRACE_VECTOR(hcpu);
 			// fall through
 
 		case hd6309_state_reset_check_halt:
-			if (cpu->halt) {
+			if (!cpu->halt) {
+				take_interrupt(cpu, 0, MC6809_INT_VEC_RESET);
+			} else {
 				NVMA_CYCLE;
-				continue;
 			}
-			REG_PC = fetch_word(cpu, MC6809_INT_VEC_RESET);
-			NVMA_CYCLE;
-			hcpu->state = hd6309_state_label_a;
 			continue;
 
 		// done_instruction case for backwards-compatibility
@@ -359,52 +381,45 @@ static void hd6309_run(struct MC6809 *cpu) {
 			// fall through
 
 		case hd6309_state_label_b:
-			if (cpu->nmi_active) {
+			if (UNLIKELY(cpu->nmi_active)) {
+				HD6309_TRACE_VECTOR(hcpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_irq_registers(cpu, 1);
 				hcpu->state = hd6309_state_dispatch_irq;
-				continue;
-			}
-			if (!(REG_CC & CC_F) && cpu->firq_active) {
+			} else if (UNLIKELY(!(REG_CC & CC_F) && cpu->firq_active)) {
+				HD6309_TRACE_VECTOR(hcpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_irq_registers(cpu, FIRQ_STACK_ALL);
 				hcpu->state = hd6309_state_dispatch_irq;
-				continue;
-			}
-			if (!(REG_CC & CC_I) && cpu->irq_active) {
+			} else if (UNLIKELY(!(REG_CC & CC_I) && cpu->irq_active)) {
+				HD6309_TRACE_VECTOR(hcpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_irq_registers(cpu, 1);
 				hcpu->state = hd6309_state_dispatch_irq;
-				continue;
+			} else {
+				HD6309_TRACE_INSTRUCTION(hcpu);
+				hcpu->state = hd6309_state_next_instruction;
+				cpu->page = 0;
+				// Instruction fetch hook called here so that machine
+				// can be stopped beforehand.
+				DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_hook);
 			}
-			hcpu->state = hd6309_state_next_instruction;
-			cpu->page = 0;
-			// Instruction fetch hook called here so that machine
-			// can be stopped beforehand.
-			DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_hook);
 			continue;
 
 		case hd6309_state_dispatch_irq:
 			if (cpu->nmi_active) {
 				cpu->nmi_active = cpu->nmi = cpu->nmi_latch = 0;
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_NMI);
-				hcpu->state = hd6309_state_label_a;
-				continue;
-			}
-			if (!(REG_CC & CC_F) && cpu->firq_active) {
+			} else if (!(REG_CC & CC_F) && cpu->firq_active) {
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_FIRQ);
-				hcpu->state = hd6309_state_label_a;
-				continue;
-			}
-			if (!(REG_CC & CC_I) && cpu->irq_active) {
+			} else if (!(REG_CC & CC_I) && cpu->irq_active) {
 				take_interrupt(cpu, CC_I, MC6809_INT_VEC_IRQ);
-				hcpu->state = hd6309_state_label_a;
-				continue;
+			} else {
+				hcpu->state = hd6309_state_cwai_check_halt;
 			}
-			hcpu->state = hd6309_state_cwai_check_halt;
 			continue;
 
 		case hd6309_state_cwai_check_halt:
@@ -412,10 +427,9 @@ static void hd6309_run(struct MC6809 *cpu) {
 			cpu->firq_active = cpu->firq_latch;
 			cpu->irq_active = cpu->irq_latch;
 			NVMA_CYCLE;
-			if (cpu->halt) {
-				continue;
+			if (!cpu->halt) {
+				hcpu->state = hd6309_state_dispatch_irq;
 			}
-			hcpu->state = hd6309_state_dispatch_irq;
 			continue;
 
 		case hd6309_state_sync:
@@ -429,8 +443,9 @@ static void hd6309_run(struct MC6809 *cpu) {
 			cpu->firq_active = cpu->firq_latch;
 			cpu->irq_active = cpu->irq_latch;
 			NVMA_CYCLE;
-			if (cpu->halt)
+			if (cpu->halt) {
 				hcpu->state = hd6309_state_sync_check_halt;
+			}
 			continue;
 
 		case hd6309_state_sync_check_halt:
@@ -441,7 +456,12 @@ static void hd6309_run(struct MC6809 *cpu) {
 			continue;
 
 		case hd6309_state_tfm:
-			// order is read, NVMA, write
+			// Order is read, NVMA, write.  XXX but is it?  Should
+			// check that and flag this as verified, because it
+			// would make more sense to do read, write, then NVMA
+			// (while it updates pointers which are all
+			// post-applied).  For sure though, the instruction is
+			// interruptable between the read and the write.
 			if (REG_W == 0) {
 				REG_PC += 3;
 				REG_CC |= CC_Z; // [hoglet67]
@@ -449,40 +469,36 @@ static void hd6309_run(struct MC6809 *cpu) {
 				break;
 			}
 			REG_M = fetch_byte_notrace(cpu, *hcpu->tfm_src);
-			NVMA_CYCLE;
 			hcpu->state = hd6309_state_tfm_write;
 			continue;
 
 		case hd6309_state_tfm_write:
 			if (cpu->nmi_active) {
 				hcpu->state = hd6309_state_label_b;
-				continue;
-			}
-			if (!(REG_CC & CC_F) && cpu->firq_active) {
+			} else if (!(REG_CC & CC_F) && cpu->firq_active) {
 				hcpu->state = hd6309_state_label_b;
-				continue;
-			}
-			if (!(REG_CC & CC_I) && cpu->irq_active) {
+			} else if (!(REG_CC & CC_I) && cpu->irq_active) {
 				hcpu->state = hd6309_state_label_b;
-				continue;
+			} else {
+				NVMA_CYCLE;
+				store_byte(cpu, *hcpu->tfm_dest, REG_M);
+				*hcpu->tfm_src += hcpu->tfm_src_mod;
+				*hcpu->tfm_dest += hcpu->tfm_dest_mod;
+				REG_W--;
+				cpu->nmi_active = cpu->nmi_latch;
+				cpu->firq_active = cpu->firq_latch;
+				cpu->irq_active = cpu->irq_latch;
+				hcpu->state = hd6309_state_tfm;
 			}
-			store_byte(cpu, *hcpu->tfm_dest, REG_M);
-			*hcpu->tfm_src += hcpu->tfm_src_mod;
-			*hcpu->tfm_dest += hcpu->tfm_dest_mod;
-			REG_W--;
-			cpu->nmi_active = cpu->nmi_latch;
-			cpu->firq_active = cpu->firq_latch;
-			cpu->irq_active = cpu->irq_latch;
-			hcpu->state = hd6309_state_tfm;
 			continue;
 
 		case hd6309_state_next_instruction:
 			{
 			unsigned op;
 			// Fetch op-code and process
-			hcpu->state = hd6309_state_label_a;
 			op = byte_immediate(cpu);
 			op |= cpu->page;
+			hcpu->state = hd6309_state_label_a;
 			switch (op) {
 
 			// 0x00 - 0x0f direct mode ops
@@ -713,7 +729,7 @@ static void hd6309_run(struct MC6809 *cpu) {
 
 			// 0x1d SEX inherent
 			case 0x1d:
-				REG_A = (REG_B & 0x80) ? 0xff : 0;
+				REG_D = sex8(REG_B);
 				CLR_NZ;
 				SET_NZ16(REG_D);
 				if (!NATIVE_MODE)
@@ -1030,7 +1046,6 @@ static void hd6309_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu, 1);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_SWI);
-				hcpu->state = hd6309_state_label_a;
 				continue;
 
 			// 0x80 - 0xbf A register arithmetic ops
@@ -1440,7 +1455,6 @@ static void hd6309_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu, 1);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_SWI2);
-				hcpu->state = hd6309_state_label_a;
 				continue;
 
 			// XXX to test: is there really no NEGW, ASRW or ASLW?
@@ -1595,11 +1609,10 @@ static void hd6309_run(struct MC6809 *cpu) {
 				ea = ea_direct(cpu);
 				mem_byte = fetch_byte_notrace(cpu, ea);
 				NVMA_CYCLE;
-				int dest_bit = postbyte & 7;
+				int dst_bit = postbyte & 7;
 				int src_bit = (postbyte >> 3) & 7;
-				int src_lsl = dest_bit - src_bit;
 				unsigned reg_code = (postbyte >> 6) & 3;
-				unsigned dst_mask = (1 << dest_bit);
+				unsigned dst_mask = (1 << dst_bit);
 				unsigned reg_val;
 				switch (reg_code) {
 				case 0: reg_val = REG_CC; break;
@@ -1613,46 +1626,47 @@ static void hd6309_run(struct MC6809 *cpu) {
 				unsigned out;
 				switch (op & 7) {
 				case 0: // BAND
-					out = ((mem_byte << src_lsl) & reg_val) & dst_mask;
+					out = (mem_byte >> src_bit) & (reg_val >> dst_bit);
 					break;
 				case 1: // BIAND
-					out = ((~mem_byte << src_lsl) & reg_val) & dst_mask;
+					out = (~mem_byte >> src_bit) & (reg_val >> dst_bit);
 					break;
 				case 2: // BOR
-					out = ((mem_byte << src_lsl) | reg_val) & dst_mask;
+					out = (mem_byte >> src_bit) | (reg_val >> dst_bit);
 					break;
 				case 3: // BIOR
-					out = ((~mem_byte << src_lsl) | reg_val) & dst_mask;
+					out = (~mem_byte >> src_bit) | (reg_val >> dst_bit);
 					break;
 				case 4: // BEOR
-					out = ((mem_byte << src_lsl) ^ reg_val) & dst_mask;
+					out = (mem_byte >> src_bit) ^ (reg_val >> dst_bit);
 					break;
 				case 5: // BIEOR
-					out = ((~mem_byte << src_lsl) ^ reg_val) & dst_mask;
+					out = (~mem_byte >> src_bit) ^ (reg_val >> dst_bit);
 					break;
 				case 6: // LDBT
-					out = (mem_byte << src_lsl) & dst_mask;
+					out = mem_byte >> src_bit;
 					break;
 				case 7: // STBT
-					out = (reg_val << src_lsl) & dst_mask;
+					out = reg_val >> src_bit;
 					break;
 				default: out = 0; break;
 				}
+				out &= 1;
 				if ((op & 7) == 7) {
 					// STBT
-					out |= (mem_byte & ~dst_mask);
-					store_byte(cpu, ea, out);
+					mem_byte = (mem_byte & ~dst_mask) | (out << dst_bit);
+					store_byte(cpu, ea, mem_byte);
 				} else {
 					switch (reg_code) {
 					default:
 					case 0:
-						REG_CC = (REG_CC & ~dst_mask) | (out & dst_mask);
+						REG_CC = (REG_CC & ~dst_mask) | (out << dst_bit);
 						break;
 					case 1:
-						REG_A = (REG_A & ~dst_mask) | (out & dst_mask);
+						REG_A = (REG_A & ~dst_mask) | (out << dst_bit);
 						break;
 					case 2:
-						REG_B = (REG_B & ~dst_mask) | (out & dst_mask);
+						REG_B = (REG_B & ~dst_mask) | (out << dst_bit);
 						break;
 					}
 				}
@@ -1683,7 +1697,6 @@ static void hd6309_run(struct MC6809 *cpu) {
 					stack_irq_registers(cpu, 1);
 					instruction_posthook(cpu);
 					take_interrupt(cpu, CC_F|CC_I, HD6309_INT_VEC_ILLEGAL);
-					hcpu->state = hd6309_state_label_a;
 					continue;
 				}
 				REG_PC -= 3;
@@ -1719,7 +1732,6 @@ static void hd6309_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu, 1);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_SWI3);
-				hcpu->state = hd6309_state_label_a;
 				continue;
 
 			// 0x1140 - 0x114f E register inherent ops
@@ -1792,12 +1804,9 @@ static void hd6309_run(struct MC6809 *cpu) {
 			} break;
 
 			// 0x118d, 0x119d, 0x11ad, 0x11bd DIVD
-			// TODO: "The cycle count of DIVD is more complex that
-			// has previously been documented" [hoglet67]
 			case 0x038d: case 0x039d: case 0x03ad: case 0x03bd: {
-				uint16_t tmp1;
+				uint16_t tmp1 = REG_D;
 				uint8_t tmp2;
-				tmp1 = REG_D;
 				switch ((op >> 4) & 3) {
 				case 0: tmp2 = byte_immediate(cpu); break;
 				case 1: tmp2 = byte_direct(cpu); break;
@@ -1805,51 +1814,86 @@ static void hd6309_run(struct MC6809 *cpu) {
 				case 3: tmp2 = byte_extended(cpu); break;
 				default: tmp2 = 0; break;
 				}
+				NVMA_CYCLE;
+				NVMA_CYCLE;
+				NVMA_CYCLE;
 				if (tmp2 == 0) {
 					REG_MD |= MD_D0;
+					CLR_NZV;
+					REG_CC |= CC_Z;  // [hoglet67]
 					stack_irq_registers(cpu, 1);
 					take_interrupt(cpu, CC_F|CC_I, HD6309_INT_VEC_ILLEGAL);
-					hcpu->state = hd6309_state_label_a;
 					continue;
 				}
-				int16_t stmp1 = *((int16_t *)&tmp1);
-				int8_t stmp2 = *((int8_t *)&tmp2);
-				int quotient = stmp1 / stmp2;
-				REG_M = (quotient < 0) ? 0xff : 0x00;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
+				_Bool nsign = 0;  // dividend sign
+				_Bool vsign = 0;  // divisor sign
+				if ((tmp1 >> 15) & 1) {
+					tmp1 = ~tmp1 + 1;
+					// Even if calculation is aborted, this
+					// negation is reflected in D:
+					REG_D = tmp1;
+					NVMA_CYCLE;  // [hoglet67]
+					nsign = 1;
+				}
+				if ((tmp2 >> 7) & 1) {
+					tmp2 = ~tmp2 + 1;
+					NVMA_CYCLE;  // [hoglet67]
+					vsign = 1;
+				}
+
+				for (int i = 6; i; i--)
+					NVMA_CYCLE;
+				uint16_t quotient = tmp1 / tmp2;
+				uint8_t remainder = tmp1 % tmp2;
+
 				CLR_NZVC;
-				if (quotient >= -256 && quotient < 256) {
-					REG_B = quotient;
-					REG_A = stmp1 - (quotient * stmp2);
-					REG_CC |= (REG_B & 1);
-					for (int i = 12; i; i--)
-						NVMA_CYCLE;
-					if ((quotient >= 0) == ((REG_B & 0x80) == 0)) {
-						SET_NZ8(REG_B);
-						NVMA_CYCLE;
-					} else {
-						REG_CC |= (CC_N | CC_V);
-					}
-				} else {
+				if ((quotient >> 8) != 0) {
+					// Range overflow
 					REG_CC |= CC_V;
+					if (nsign)
+						REG_CC |= CC_N;
+					break;
+				}
+
+				// According to [hoglet67] there is now a
+				// maximum of 13 cycles remaining, one fewer
+				// on 2's complement overflow.
+
+				if (nsign) {
+					remainder = ~remainder + 1;
+				}
+				if (nsign != vsign) {
+					uint16_t new_quotient = ~quotient + 1;
+					if ((new_quotient & 0x80) && !(quotient & 0x80)) {
+						quotient = new_quotient;
+					}
+					REG_M = 0xff;
+				} else {
+					REG_M = 0;
+				}
+
+				for (int i = 12; i; i--)
+					NVMA_CYCLE;
+				REG_A = remainder;
+				REG_B = quotient;
+				REG_CC |= (REG_B & 1);
+				if ((((quotient >> 15) ^ (REG_B >> 7)) & 1) == 0) {
+					// No overflow, take the extra cycle
+					SET_NZ8(REG_B);
+					NVMA_CYCLE;
+				} else {
+					// 2's complement overflow.  NOTE:
+					// [hoglet67] says N is clear here, but
+					// Tim Lindner's fuzzing suggests
+					// otherwise?
+					REG_CC |= (CC_N | CC_V);
 				}
 			} break;
 
 			// 0x118e, 0x119e, 0x11ae, 0x11be DIVQ
-			// TODO: "The cycle count of DIVQ is more complex that
-			// has previously been documented" [hoglet67]
 			case 0x038e: case 0x039e: case 0x03ae: case 0x03be: {
-				uint32_t tmp1;
+				uint32_t tmp1 = RREG_Q;
 				uint16_t tmp2;
-				tmp1 = RREG_Q;
 				switch ((op >> 4) & 3) {
 				case 0: tmp2 = word_immediate(cpu); break;
 				case 1: tmp2 = word_direct(cpu); break;
@@ -1857,40 +1901,79 @@ static void hd6309_run(struct MC6809 *cpu) {
 				case 3: tmp2 = word_extended(cpu); break;
 				default: tmp2 = 0; break;
 				}
+				NVMA_CYCLE;
+				NVMA_CYCLE;
+				NVMA_CYCLE;
 				if (tmp2 == 0) {
 					REG_MD |= MD_D0;
+					CLR_NZV;
+					REG_CC |= CC_Z;  // [hoglet67]
 					stack_irq_registers(cpu, 1);
 					take_interrupt(cpu, CC_F|CC_I, HD6309_INT_VEC_ILLEGAL);
-					hcpu->state = hd6309_state_label_a;
 					continue;
 				}
-				int32_t stmp1 = *((int32_t *)&tmp1);
-				int16_t stmp2 = *((int16_t *)&tmp2);
-				int quotient = stmp1 / stmp2;
-				REG_M = (quotient < 0) ? 0xff : 0x00;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
-				NVMA_CYCLE;
+				_Bool nsign = 0;  // dividend sign
+				_Bool vsign = 0;  // divisor sign
+				if ((tmp1 >> 31) & 1) {
+					tmp1 = ~tmp1 + 1;
+					// Even if calculation is aborted, this
+					// negation is reflected in Q:
+					REG_D = tmp1 >> 16;
+					REG_W = tmp1 & 0xffff;;
+					NVMA_CYCLE;  // [hoglet67]
+					nsign = 1;
+				}
+				if ((tmp2 >> 15) & 1) {
+					tmp2 = ~tmp2 + 1;
+					NVMA_CYCLE;  // [hoglet67]
+					vsign = 1;
+				}
+
+				for (int i = 6; i; i--)
+					NVMA_CYCLE;
+				uint32_t quotient = tmp1 / tmp2;
+				uint16_t remainder = tmp1 % tmp2;
+
 				CLR_NZVC;
-				if (quotient >= -65536 && quotient < 65536) {
-					REG_W = quotient;
-					REG_D = stmp1 - (quotient * stmp2);
-					REG_CC |= (REG_W & 1);
-					for (int i = 21; i; i--)
-						NVMA_CYCLE;
-					if ((quotient >= 0) == ((REG_E & 0x80) == 0)) {
-						SET_NZ16(REG_W);
-					} else {
-						REG_CC |= (CC_N | CC_V);
-					}
-				} else {
+				if ((quotient >> 16) != 0) {
+					// Range overflow
 					REG_CC |= CC_V;
+					if (nsign)
+						REG_CC |= CC_N;
+					break;
+				}
+
+				// According to [hoglet67] there are now 21
+				// cycles remaining.
+
+				if (nsign) {
+					remainder = ~remainder + 1;
+				}
+				if (nsign != vsign) {
+					uint32_t new_quotient = ~quotient + 1;
+					if ((new_quotient & 0x8000) && !(quotient & 0x8000)) {
+						quotient = new_quotient;
+					}
+					REG_M = 0xff;
+				} else {
+					REG_M = 0;
+				}
+
+				for (int i = 21; i; i--)
+					NVMA_CYCLE;
+				REG_D = remainder;
+				REG_W = quotient;
+				REG_CC |= (REG_W & 1);
+				if ((((quotient >> 31) ^ (REG_E >> 7)) & 1) == 0) {
+					// No overflow
+					SET_NZ16(REG_W);
+				} else {
+					// 2's complement overflow.  NOTE:
+					// [hoglet67] says N is clear here, but
+					// adopting results of Tim Lindner's
+					// fuzzing of DIVD (see above) until we
+					// know one way or another.
+					REG_CC |= (CC_N | CC_V);
 				}
 			} break;
 
@@ -1930,7 +2013,6 @@ static void hd6309_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu, 1);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, CC_F|CC_I, HD6309_INT_VEC_ILLEGAL);
-				hcpu->state = hd6309_state_label_a;
 				continue;
 			}
 
@@ -1957,6 +2039,7 @@ static void hd6309_run(struct MC6809 *cpu) {
 static void hd6309_set_pc(void *sptr, unsigned pc) {
 	struct HD6309 *hcpu = sptr;
 	struct MC6809 *cpu = &hcpu->mc6809;
+	HD6309_TRACE_INSTRUCTION(hcpu);
 	REG_PC = pc;
 	hcpu->state = hd6309_state_next_instruction;
 }
@@ -1973,7 +2056,7 @@ static uint8_t fetch_byte(struct MC6809 *cpu, uint16_t a) {
 	uint8_t v = fetch_byte_notrace(cpu, a);
 #ifdef TRACE
 	struct HD6309 *hcpu = (struct HD6309 *)cpu;
-	if (logging.trace_cpu) {
+	if (UNLIKELY(logging.trace_cpu)) {
 		hd6309_trace_byte(hcpu->tracer, v, a);
 	}
 #endif
@@ -1985,7 +2068,7 @@ static uint16_t fetch_word(struct MC6809 *cpu, uint16_t a) {
 	return fetch_word_notrace(cpu, a);
 #else
 	struct HD6309 *hcpu = (struct HD6309 *)cpu;
-	if (!logging.trace_cpu) {
+	if (LIKELY(!logging.trace_cpu)) {
 		return fetch_word_notrace(cpu, a);
 	}
 	unsigned v0 = fetch_byte_notrace(cpu, a);
@@ -2126,24 +2209,15 @@ static void stack_irq_registers(struct MC6809 *cpu, _Bool entire) {
 }
 
 static void take_interrupt(struct MC6809 *cpu, uint8_t mask, uint16_t vec) {
+	struct HD6309 *hcpu = (struct HD6309 *)cpu;
 	REG_CC |= mask;
 	NVMA_CYCLE;
-#ifdef TRACE
-	struct HD6309 *hcpu = (struct HD6309 *)cpu;
-	if (logging.trace_cpu) {
-		hd6309_trace_irq(hcpu->tracer, vec);
-	}
-#endif
+	hcpu->state = hd6309_state_irq_reset_vector;
 	REG_PC = fetch_word(cpu, vec);
+	hcpu->state = hd6309_state_label_a;
 	NVMA_CYCLE;
 }
 
 static void instruction_posthook(struct MC6809 *cpu) {
-#ifdef TRACE
-	struct HD6309 *hcpu = (struct HD6309 *)cpu;
-	if (logging.trace_cpu) {
-		hd6309_trace_print(hcpu->tracer);
-	}
-#endif
 	DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_posthook);
 }
