@@ -2,7 +2,7 @@
  *
  *  \brief Motorola MC6809 CPU.
  *
- *  \copyright Copyright 2003-2022 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -34,6 +34,9 @@
  *     http://tlindner.macmess.org/?p=945
  */
 
+// Additionally, I quite early on made reference to MAME source to understand
+// the overflow (V) flag.
+
 // TODO:
 //
 // - Many more instructions fall through to their unprefixed form after a
@@ -57,36 +60,43 @@
 #include "serialise.h"
 
 #ifdef TRACE
+// Tracing supported:
 #include "mc6809_trace.h"
+#define MC6809_TRACE_VECTOR(c) if (UNLIKELY(logging.trace_cpu)) { mc6809_trace_vector((c)->tracer); }
+#define MC6809_TRACE_INSTRUCTION(c) if (UNLIKELY(logging.trace_cpu)) { mc6809_trace_instruction((c)->tracer); }
+#else
+// Tracing not supported - no-op macros:
+#define MC6809_TRACE_VECTOR(c)
+#define MC6809_TRACE_INSTRUCTION(c)
 #endif
 
 static const struct ser_struct ser_struct_mc6809[] = {
-	SER_ID_STRUCT_ELEM(1, ser_type_bool, struct MC6809, halt),
-	SER_ID_STRUCT_ELEM(2, ser_type_bool, struct MC6809, nmi),
-	SER_ID_STRUCT_ELEM(3, ser_type_bool, struct MC6809, firq),
-	SER_ID_STRUCT_ELEM(4, ser_type_bool, struct MC6809, irq),
-	SER_ID_STRUCT_ELEM(5, ser_type_uint8, struct MC6809, D),
+	SER_ID_STRUCT_ELEM(1, struct MC6809, halt),
+	SER_ID_STRUCT_ELEM(2, struct MC6809, nmi),
+	SER_ID_STRUCT_ELEM(3, struct MC6809, firq),
+	SER_ID_STRUCT_ELEM(4, struct MC6809, irq),
+	SER_ID_STRUCT_ELEM(5, struct MC6809, D),
 
-	SER_ID_STRUCT_ELEM(6, ser_type_unsigned, struct MC6809, state),
-	SER_ID_STRUCT_ELEM(7, ser_type_bool, struct MC6809, running),
-	SER_ID_STRUCT_ELEM(8, ser_type_uint16, struct MC6809, page),
+	SER_ID_STRUCT_ELEM(6, struct MC6809, state),
+	SER_ID_STRUCT_ELEM(7, struct MC6809, running),
+	SER_ID_STRUCT_ELEM(8, struct MC6809, page),
 
-	SER_ID_STRUCT_ELEM(9, ser_type_uint8, struct MC6809, reg_cc),
-	SER_ID_STRUCT_ELEM(10, ser_type_uint16, struct MC6809, reg_d),
-	SER_ID_STRUCT_ELEM(11, ser_type_uint8, struct MC6809, reg_dp),
-	SER_ID_STRUCT_ELEM(12, ser_type_uint16, struct MC6809, reg_x),
-	SER_ID_STRUCT_ELEM(13, ser_type_uint16, struct MC6809, reg_y),
-	SER_ID_STRUCT_ELEM(14, ser_type_uint16, struct MC6809, reg_u),
-	SER_ID_STRUCT_ELEM(15, ser_type_uint16, struct MC6809, reg_s),
-	SER_ID_STRUCT_ELEM(16, ser_type_uint16, struct MC6809, reg_pc),
+	SER_ID_STRUCT_ELEM(9, struct MC6809, reg_cc),
+	SER_ID_STRUCT_ELEM(10, struct MC6809, reg_d),
+	SER_ID_STRUCT_ELEM(11, struct MC6809, reg_dp),
+	SER_ID_STRUCT_ELEM(12, struct MC6809, reg_x),
+	SER_ID_STRUCT_ELEM(13, struct MC6809, reg_y),
+	SER_ID_STRUCT_ELEM(14, struct MC6809, reg_u),
+	SER_ID_STRUCT_ELEM(15, struct MC6809, reg_s),
+	SER_ID_STRUCT_ELEM(16, struct MC6809, reg_pc),
 
-	SER_ID_STRUCT_ELEM(17, ser_type_bool, struct MC6809, nmi_armed),
-	SER_ID_STRUCT_ELEM(18, ser_type_bool, struct MC6809, nmi_latch),
-	SER_ID_STRUCT_ELEM(19, ser_type_bool, struct MC6809, firq_latch),
-	SER_ID_STRUCT_ELEM(20, ser_type_bool, struct MC6809, irq_latch),
-	SER_ID_STRUCT_ELEM(21, ser_type_bool, struct MC6809, nmi_active),
-	SER_ID_STRUCT_ELEM(22, ser_type_bool, struct MC6809, firq_active),
-	SER_ID_STRUCT_ELEM(23, ser_type_bool, struct MC6809, irq_active),
+	SER_ID_STRUCT_ELEM(17, struct MC6809, nmi_armed),
+	SER_ID_STRUCT_ELEM(18, struct MC6809, nmi_latch),
+	SER_ID_STRUCT_ELEM(19, struct MC6809, firq_latch),
+	SER_ID_STRUCT_ELEM(20, struct MC6809, irq_latch),
+	SER_ID_STRUCT_ELEM(21, struct MC6809, nmi_active),
+	SER_ID_STRUCT_ELEM(22, struct MC6809, firq_active),
+	SER_ID_STRUCT_ELEM(23, struct MC6809, irq_active),
 };
 
 // External, as struct data is also used by 6809-compatible parts:
@@ -204,7 +214,7 @@ static const struct partdb_entry_funcs mc6809_funcs = {
 	.is_a = mc6809_is_a,
 };
 
-const struct partdb_entry mc6809_part = { .name = "MC6809", .funcs = &mc6809_funcs };
+const struct partdb_entry mc6809_part = { .name = "MC6809", .description = "Motorola | MC6809E CPU", .funcs = &mc6809_funcs };
 
 static struct part *mc6809_allocate(void) {
 	struct MC6809 *cpu = part_new(sizeof(*cpu));
@@ -219,6 +229,11 @@ static struct part *mc6809_allocate(void) {
 	cpu->run = mc6809_run;
 	cpu->mem_cycle = DELEGATE_DEFAULT2(void, bool, uint16);
 
+	// Tested: (almost?) always, all registers are zeroed on power on.
+	//
+	// CC has F and I set as part of reset.  DP is explicitly cleared on
+	// reset, but other registers are left untouched.
+
 #ifdef TRACE
 	// Tracing
 	cpu->tracer = mc6809_trace_new(cpu);
@@ -281,21 +296,15 @@ static void mc6809_run(struct MC6809 *cpu) {
 			cpu->firq_active = 0;
 			cpu->irq_active = 0;
 			cpu->state = mc6809_state_reset_check_halt;
-#ifdef TRACE
-			if (logging.trace_cpu) {
-				mc6809_trace_irq(cpu->tracer, MC6809_INT_VEC_RESET);
-			}
-#endif
+			MC6809_TRACE_VECTOR(cpu);
 			// fall through
 
 		case mc6809_state_reset_check_halt:
-			if (cpu->halt) {
+			if (!cpu->halt) {
+				take_interrupt(cpu, 0, MC6809_INT_VEC_RESET);
+			} else {
 				NVMA_CYCLE;
-				continue;
 			}
-			REG_PC = fetch_word(cpu, MC6809_INT_VEC_RESET);
-			NVMA_CYCLE;
-			cpu->state = mc6809_state_label_a;
 			continue;
 
 		// done_instruction case for backwards-compatibility
@@ -309,52 +318,45 @@ static void mc6809_run(struct MC6809 *cpu) {
 			// fall through
 
 		case mc6809_state_label_b:
-			if (cpu->nmi_active) {
+			if (UNLIKELY(cpu->nmi_active)) {
+				MC6809_TRACE_VECTOR(cpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_irq_registers(cpu);
 				cpu->state = mc6809_state_dispatch_irq;
-				continue;
-			}
-			if (!(REG_CC & CC_F) && cpu->firq_active) {
+			} else if (UNLIKELY(!(REG_CC & CC_F) && cpu->firq_active)) {
+				MC6809_TRACE_VECTOR(cpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_firq_registers(cpu);
 				cpu->state = mc6809_state_dispatch_irq;
-				continue;
-			}
-			if (!(REG_CC & CC_I) && cpu->irq_active) {
+			} else if (UNLIKELY(!(REG_CC & CC_I) && cpu->irq_active)) {
+				MC6809_TRACE_VECTOR(cpu);
 				peek_byte(cpu, REG_PC);
 				peek_byte(cpu, REG_PC);
 				stack_irq_registers(cpu);
 				cpu->state = mc6809_state_dispatch_irq;
-				continue;
+			} else {
+				MC6809_TRACE_INSTRUCTION(cpu);
+				cpu->state = mc6809_state_next_instruction;
+				cpu->page = 0;
+				// Instruction fetch hook called here so that machine
+				// can be stopped beforehand.
+				DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_hook);
 			}
-			cpu->state = mc6809_state_next_instruction;
-			cpu->page = 0;
-			// Instruction fetch hook called here so that machine
-			// can be stopped beforehand.
-			DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_hook);
 			continue;
 
 		case mc6809_state_dispatch_irq:
 			if (cpu->nmi_active) {
 				cpu->nmi_active = cpu->nmi = cpu->nmi_latch = 0;
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_NMI);
-				cpu->state = mc6809_state_label_a;
-				continue;
-			}
-			if (!(REG_CC & CC_F) && cpu->firq_active) {
+			} else if (!(REG_CC & CC_F) && cpu->firq_active) {
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_FIRQ);
-				cpu->state = mc6809_state_label_a;
-				continue;
-			}
-			if (!(REG_CC & CC_I) && cpu->irq_active) {
+			} else if (!(REG_CC & CC_I) && cpu->irq_active) {
 				take_interrupt(cpu, CC_I, MC6809_INT_VEC_IRQ);
-				cpu->state = mc6809_state_label_a;
-				continue;
+			} else {
+				cpu->state = mc6809_state_cwai_check_halt;
 			}
-			cpu->state = mc6809_state_cwai_check_halt;
 			continue;
 
 		case mc6809_state_cwai_check_halt:
@@ -362,10 +364,9 @@ static void mc6809_run(struct MC6809 *cpu) {
 			cpu->firq_active = cpu->firq_latch;
 			cpu->irq_active = cpu->irq_latch;
 			NVMA_CYCLE;
-			if (cpu->halt) {
-				continue;
+			if (!cpu->halt) {
+				cpu->state = mc6809_state_dispatch_irq;
 			}
-			cpu->state = mc6809_state_dispatch_irq;
 			continue;
 
 		case mc6809_state_sync:
@@ -379,8 +380,9 @@ static void mc6809_run(struct MC6809 *cpu) {
 			cpu->firq_active = cpu->firq_latch;
 			cpu->irq_active = cpu->irq_latch;
 			NVMA_CYCLE;
-			if (cpu->halt)
+			if (cpu->halt) {
 				cpu->state = mc6809_state_sync_check_halt;
+			}
 			continue;
 
 		case mc6809_state_sync_check_halt:
@@ -392,10 +394,10 @@ static void mc6809_run(struct MC6809 *cpu) {
 
 		case mc6809_state_next_instruction: {
 			unsigned op;
-			cpu->state = mc6809_state_label_a;
 			// Fetch op-code and process
 			op = byte_immediate(cpu);
 			op |= cpu->page;
+			cpu->state = mc6809_state_label_a;
 			switch (op) {
 
 			// 0x00 - 0x0f direct mode ops
@@ -591,10 +593,11 @@ static void mc6809_run(struct MC6809 *cpu) {
 				break;
 
 			// 0x16 LBRA relative
-			// 0x10xx, 0x11xx - page 2/3 fallthrough, illegal
+			// 0x1016, 0x1116 - page 2/3 fallthrough, illegal
+			// 0x108d, 0x118d - LBRA observed by darrena on discord
 			case 0x16:
-			case 0x0216:
-			case 0x0316: {
+			case 0x0216: case 0x028d:
+			case 0x0316: case 0x038d: {
 				uint16_t ea = long_relative(cpu);
 				REG_PC += ea;
 				NVMA_CYCLE;
@@ -662,7 +665,7 @@ static void mc6809_run(struct MC6809 *cpu) {
 			case 0x1d:
 			case 0x021d:
 			case 0x031d:
-				REG_A = (REG_B & 0x80) ? 0xff : 0;
+				REG_D = sex8(REG_B);
 				CLR_NZ;
 				SET_NZ16(REG_D);
 				peek_byte(cpu, REG_PC);
@@ -992,7 +995,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				push_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_RESET);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// 0x3f SWI inherent
@@ -1001,7 +1003,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, CC_F|CC_I, MC6809_INT_VEC_SWI);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// 0x80 - 0xbf A register arithmetic ops
@@ -1198,8 +1199,8 @@ static void mc6809_run(struct MC6809 *cpu) {
 			// 0x9d, 0xad, 0xbd JSR
 			// 0x10xx, 0x11xx - page 2/3 fallthrough, illegal
 			case 0x8d: case 0x9d: case 0xad: case 0xbd:
-			case 0x028d: case 0x029d: case 0x02ad: case 0x02bd:
-			case 0x038d: case 0x039d: case 0x03ad: case 0x03bd: {
+			case 0x029d: case 0x02ad: case 0x02bd:
+			case 0x039d: case 0x03ad: case 0x03bd: {
 				unsigned ea;
 				switch ((op >> 4) & 3) {
 				case 0: ea = short_relative(cpu); ea += REG_PC; NVMA_CYCLE; NVMA_CYCLE; NVMA_CYCLE; break;
@@ -1331,7 +1332,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				push_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_SWI2);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// 0x103f SWI2 inherent
@@ -1340,7 +1340,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_SWI2);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// 0x113e FIRQ inherent, illegal [hoglet67]
@@ -1349,7 +1348,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				push_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_FIRQ);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// 0x113f SWI3 inherent
@@ -1358,7 +1356,6 @@ static void mc6809_run(struct MC6809 *cpu) {
 				stack_irq_registers(cpu);
 				instruction_posthook(cpu);
 				take_interrupt(cpu, 0, MC6809_INT_VEC_SWI3);
-				cpu->state = mc6809_state_label_a;
 				continue;
 
 			// Illegal instruction
@@ -1393,6 +1390,7 @@ static void mc6809_run(struct MC6809 *cpu) {
 
 static void mc6809_set_pc(void *sptr, unsigned pc) {
 	struct MC6809 *cpu = sptr;
+	MC6809_TRACE_INSTRUCTION(cpu);
 	REG_PC = pc;
 	cpu->state = mc6809_state_next_instruction;
 }
@@ -1408,7 +1406,7 @@ static void mc6809_set_pc(void *sptr, unsigned pc) {
 static uint8_t fetch_byte(struct MC6809 *cpu, uint16_t a) {
 	uint8_t v = fetch_byte_notrace(cpu, a);
 #ifdef TRACE
-	if (logging.trace_cpu) {
+	if (UNLIKELY(logging.trace_cpu)) {
 		mc6809_trace_byte(cpu->tracer, v, a);
 	}
 #endif
@@ -1419,7 +1417,7 @@ static uint16_t fetch_word(struct MC6809 *cpu, uint16_t a) {
 #ifndef TRACE
 	return fetch_word_notrace(cpu, a);
 #else
-	if (!logging.trace_cpu) {
+	if (LIKELY(!logging.trace_cpu)) {
 		return fetch_word_notrace(cpu, a);
 	}
 	unsigned v0 = fetch_byte_notrace(cpu, a);
@@ -1530,21 +1528,13 @@ static void stack_firq_registers(struct MC6809 *cpu) {
 static void take_interrupt(struct MC6809 *cpu, uint8_t mask, uint16_t vec) {
 	REG_CC |= mask;
 	NVMA_CYCLE;
-#ifdef TRACE
-	if (logging.trace_cpu) {
-		mc6809_trace_irq(cpu->tracer, vec);
-	}
-#endif
+	cpu->state = mc6809_state_irq_reset_vector;
 	REG_PC = fetch_word(cpu, vec);
+	cpu->state = mc6809_state_label_a;
 	NVMA_CYCLE;
 }
 
 static void instruction_posthook(struct MC6809 *cpu) {
-#ifdef TRACE
-	if (logging.trace_cpu) {
-		mc6809_trace_print(cpu->tracer);
-	}
-#endif
 	DELEGATE_SAFE_CALL(cpu->debug_cpu.instruction_posthook);
 }
 
