@@ -2,7 +2,7 @@
  *
  *  \brief Motorola MC6809 CPU tracing.
  *
- *  \copyright Copyright 2005-2017 Ciaran Anscomb
+ *  \copyright Copyright 2005-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -25,6 +25,8 @@
 
 #include "xalloc.h"
 
+#include "events.h"
+#include "logging.h"
 #include "mc6809.h"
 #include "mc6809_trace.h"
 
@@ -190,7 +192,7 @@ static struct {
 		{ "ANDA", IMMEDIATE },
 		{ "BITA", IMMEDIATE },
 		{ "LDA", IMMEDIATE },
-		{ "DISCRD*", IMMEDIATE },
+		{ "DSCA*", IMMEDIATE },
 		{ "EORA", IMMEDIATE },
 		{ "ADCA", IMMEDIATE },
 		{ "ORA", IMMEDIATE },
@@ -198,7 +200,7 @@ static struct {
 		{ "CMPX", WORD_IMMEDIATE },
 		{ "BSR", RELATIVE },
 		{ "LDX", WORD_IMMEDIATE },
-		{ "STXIMM*", IMMEDIATE },
+		{ "STXI*", IMMEDIATE },
 		// 0x90 - 0x9F
 		{ "SUBA", DIRECT },
 		{ "CMPA", DIRECT },
@@ -258,7 +260,7 @@ static struct {
 		{ "ANDB", IMMEDIATE },
 		{ "BITB", IMMEDIATE },
 		{ "LDB", IMMEDIATE },
-		{ "DISCRD*", IMMEDIATE },
+		{ "DSCB*", IMMEDIATE },
 		{ "EORB", IMMEDIATE },
 		{ "ADCB", IMMEDIATE },
 		{ "ORB", IMMEDIATE },
@@ -266,7 +268,7 @@ static struct {
 		{ "LDD", WORD_IMMEDIATE },
 		{ "HCF*", INHERENT },
 		{ "LDU", WORD_IMMEDIATE },
-		{ "STUIMM*", IMMEDIATE },
+		{ "STUI*", IMMEDIATE },
 		// 0xD0 - 0xDF
 		{ "SUBB", DIRECT },
 		{ "CMPB", DIRECT },
@@ -870,22 +872,22 @@ static struct {
 	}
 };
 
-/* The next byte is expected to be one of these, with special exceptions:
- * WANT_PRINT - expecting trace_print to be called
- * WANT_NOTHING - expecting a byte that is to be ignored */
+/* The next byte is expected to be one of these, with the special exceptions of
+ * WANT_PRINT, which indicates a trace line is complete. */
 
 enum {
 	WANT_INSTRUCTION,
+	WANT_INSTRUCTION2,
 	WANT_IRQ_VECTOR,
+	WANT_IRQ_VECTOR2,  // second byte
 	WANT_IDX_POSTBYTE,
 	WANT_VALUE,
-	WANT_PRINT,
-	WANT_NOTHING
+	WANT_PRINT
 };
 
 /* Sequences of expected bytes */
 
-static int const state_list_irq[] = { WANT_VALUE, WANT_PRINT };
+static int const state_list_irq[] = { WANT_IRQ_VECTOR2, WANT_PRINT };
 static int const state_list_inherent[] = { WANT_PRINT };
 static int const state_list_idx[] = { WANT_IDX_POSTBYTE };
 static int const state_list_imm8[] = { WANT_VALUE, WANT_PRINT };
@@ -918,10 +920,10 @@ static char const * const idx_fmts[17] = {
 	"%s$%04x,%s%s",
 	"%s*%s",
 	"%sD,%s%s",
-	"%s%s$%02x,PCR%s",
-	"%s$%04x,PCR%s",
+	"%s<$%04x,PCR%s",
+	"%s>$%04x,PCR%s",
 	"%s*%s",
-	"%s$%04X%s",
+	"%s$%04x%s",
 	"%s%d,%s%s",
 };
 
@@ -966,16 +968,20 @@ static char const * const irq_names[8] = {
 
 /* Current state */
 
-#define BYTES_BUF_SIZE 5
+// Maximum number of bytes to hold in buffer for printing.  Note that illegal
+// runs of 0x10 or 0x11 could overflow this, but the instruction decode should
+// still be ok.
+#define MAX_NBYTES 5
 
 struct mc6809_trace {
 	struct MC6809 *cpu;
 
 	int state;
+	event_ticks start_tick;
 	int page;
 	uint16_t instr_pc;
-	int bytes_count;
-	uint8_t bytes_buf[BYTES_BUF_SIZE];
+	unsigned nbytes;
+	uint8_t bytes[MAX_NBYTES];
 
 	const char *mnemonic;
 	char operand_text[19];
@@ -988,12 +994,12 @@ struct mc6809_trace {
 	_Bool idx_indirect;
 };
 
-static void reset_state(struct mc6809_trace *tracer);
-static void trace_print_short(struct mc6809_trace *tracer);
+static void reset_state(struct mc6809_trace *);
+static void print_record(struct mc6809_trace *);
 
-#define STACK_PRINT(t,r) do { \
-		if (not_first) { strcat((t)->operand_text, "," r); } \
-		else { strcat((t)->operand_text, r); not_first = 1; } \
+#define STACK_PRINT(t,r,c) do { \
+		if (c) { strcat((t)->operand_text, r ","); } \
+		else { strcat((t)->operand_text, r); } \
 	} while (0)
 
 #define sex5(v) ((int)((v) & 0x0f) - (int)((v) & 0x10))
@@ -1005,6 +1011,7 @@ struct mc6809_trace *mc6809_trace_new(struct MC6809 *cpu) {
 	struct mc6809_trace *tracer = xmalloc(sizeof(*tracer));
 	*tracer = (struct mc6809_trace){0};
 	tracer->cpu = cpu;
+	tracer->start_tick = event_current_tick;
 	reset_state(tracer);
 	return tracer;
 }
@@ -1015,52 +1022,58 @@ void mc6809_trace_free(struct mc6809_trace *tracer) {
 
 void mc6809_trace_reset(struct mc6809_trace *tracer) {
 	reset_state(tracer);
-	mc6809_trace_irq(tracer, 0xfffe);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-static void reset_state(struct mc6809_trace *tracer) {
-	tracer->state = WANT_INSTRUCTION;
-	tracer->page = PAGE0;
-	tracer->instr_pc = 0;
-	tracer->bytes_count = 0;
-	tracer->mnemonic = "*";
-	strcpy(tracer->operand_text, "*");
+// Called at each timing checkpoint
 
-	tracer->ins_type = PAGE0;
-	tracer->state_list = NULL;
-	tracer->idx_mode = 0;
-	tracer->idx_reg = "";
-	tracer->idx_indirect = 0;
+void mc6809_trace_vector(struct mc6809_trace *tracer) {
+	print_record(tracer);
+	tracer->state = WANT_IRQ_VECTOR;
+}
+
+void mc6809_trace_instruction(struct mc6809_trace *tracer) {
+	print_record(tracer);
+	tracer->state = WANT_INSTRUCTION;
 }
 
 /* Called for each memory read */
 
 void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
+	struct MC6809 *cpu = tracer->cpu;
 
-	// Record PC of instruction
-	if (tracer->bytes_count == 0) {
+	if (cpu->state == mc6809_state_next_instruction && tracer->state == WANT_INSTRUCTION) {
+		tracer->instr_pc = pc;
+	} else if (cpu->state == mc6809_state_irq_reset_vector && tracer->state == WANT_IRQ_VECTOR) {
+		tracer->mnemonic = irq_names[(pc & 15) >> 1];
 		tracer->instr_pc = pc;
 	}
 
-	// Record byte if considered part of instruction
-	if (tracer->bytes_count < BYTES_BUF_SIZE && tracer->state != WANT_PRINT && tracer->state != WANT_NOTHING) {
-		tracer->bytes_buf[tracer->bytes_count++] = byte;
+	if (tracer->state == WANT_PRINT) {
+		return;
+	}
+
+	// Record byte if it follows current PC.  Note: nbytes not
+	// incremented here: individual states will bump it if necessary.
+	if (pc == (tracer->instr_pc + tracer->nbytes)) {
+		if (tracer->nbytes < MAX_NBYTES) {
+			tracer->bytes[tracer->nbytes] = byte;
+		}
 	}
 
 	switch (tracer->state) {
 
 		// Instruction fetch
-		default:
 		case WANT_INSTRUCTION:
+		case WANT_INSTRUCTION2:
 			tracer->value = 0;
-			tracer->state_list = NULL;
 			tracer->mnemonic = instructions[tracer->page][byte].mnemonic;
 			tracer->ins_type = instructions[tracer->page][byte].type;
 			switch (tracer->ins_type) {
-				// Change page, stay in WANT_INSTRUCTION state
+				// Change page
 				case PAGE2: case PAGE3:
+					tracer->state = WANT_INSTRUCTION2;
 					tracer->page = tracer->ins_type;
 					break;
 				// Otherwise use an appropriate state list:
@@ -1079,6 +1092,7 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 					tracer->state_list = state_list_imm16;
 					break;
 			}
+			tracer->nbytes++;
 			break;
 
 		// First byte of an IRQ vector
@@ -1086,11 +1100,14 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 			tracer->value = byte;
 			tracer->ins_type = IRQVECTOR;
 			tracer->state_list = state_list_irq;
+			tracer->nbytes++;
 			break;
 
 		// Building a value byte by byte
 		case WANT_VALUE:
+		case WANT_IRQ_VECTOR2:
 			tracer->value = (tracer->value << 8) | byte;
+			tracer->nbytes++;
 			break;
 
 		// Indexed postbyte - record relevant details
@@ -1116,24 +1133,21 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 				tracer->idx_mode = IDX_PD2;
 			}
 			tracer->state_list = idx_state_lists[tracer->idx_mode];
+			tracer->nbytes++;
 			break;
 
-		// Expecting CPU code to call trace_print
-		case WANT_PRINT:
-			tracer->state_list = NULL;
-			return;
-
-		// This byte is to be ignored (used following IRQ vector fetch)
-		case WANT_NOTHING:
+		default:
 			break;
 	}
 
 	// Get next state from state list
-	if (tracer->state_list)
+	if (tracer->state_list) {
 		tracer->state = *(tracer->state_list++);
+	}
 
-	if (tracer->state != WANT_PRINT)
+	if (tracer->state != WANT_PRINT) {
 		return;
+	}
 
 	// If the next state is WANT_PRINT, we're done with the instruction, so
 	// prep the operand text for printing.
@@ -1162,27 +1176,27 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 			break;
 
 		case STACKS: {
-			_Bool not_first = 0;
-			if (tracer->value & 0x01) { STACK_PRINT(tracer, "CC"); }
-			if (tracer->value & 0x02) { STACK_PRINT(tracer, "A"); }
-			if (tracer->value & 0x04) { STACK_PRINT(tracer, "B"); }
-			if (tracer->value & 0x08) { STACK_PRINT(tracer, "DP"); }
-			if (tracer->value & 0x10) { STACK_PRINT(tracer, "X"); }
-			if (tracer->value & 0x20) { STACK_PRINT(tracer, "Y"); }
-			if (tracer->value & 0x40) { STACK_PRINT(tracer, "U"); }
-			if (tracer->value & 0x80) { STACK_PRINT(tracer, "PC"); }
+			unsigned postbyte = tracer->value;
+			if (postbyte & 0x01) { STACK_PRINT(tracer, "CC", postbyte & 0xfe); }
+			if (postbyte & 0x02) { STACK_PRINT(tracer, "A",  postbyte & 0xfc); }
+			if (postbyte & 0x04) { STACK_PRINT(tracer, "B",  postbyte & 0xf8); }
+			if (postbyte & 0x08) { STACK_PRINT(tracer, "DP", postbyte & 0xf0); }
+			if (postbyte & 0x10) { STACK_PRINT(tracer, "X",  postbyte & 0xe0); }
+			if (postbyte & 0x20) { STACK_PRINT(tracer, "Y",  postbyte & 0xc0); }
+			if (postbyte & 0x40) { STACK_PRINT(tracer, "U",  postbyte & 0x80); }
+			if (postbyte & 0x80) { STACK_PRINT(tracer, "PC", 0); }
 		} break;
 
 		case STACKU: {
-			_Bool not_first = 0;
-			if (tracer->value & 0x01) { STACK_PRINT(tracer, "CC"); }
-			if (tracer->value & 0x02) { STACK_PRINT(tracer, "A"); }
-			if (tracer->value & 0x04) { STACK_PRINT(tracer, "B"); }
-			if (tracer->value & 0x08) { STACK_PRINT(tracer, "DP"); }
-			if (tracer->value & 0x10) { STACK_PRINT(tracer, "X"); }
-			if (tracer->value & 0x20) { STACK_PRINT(tracer, "Y"); }
-			if (tracer->value & 0x40) { STACK_PRINT(tracer, "S"); }
-			if (tracer->value & 0x80) { STACK_PRINT(tracer, "PC"); }
+			unsigned postbyte = tracer->value;
+			if (postbyte & 0x01) { STACK_PRINT(tracer, "CC", postbyte & 0xfe); }
+			if (postbyte & 0x02) { STACK_PRINT(tracer, "A",  postbyte & 0xfc); }
+			if (postbyte & 0x04) { STACK_PRINT(tracer, "B",  postbyte & 0xf8); }
+			if (postbyte & 0x08) { STACK_PRINT(tracer, "DP", postbyte & 0xf0); }
+			if (postbyte & 0x10) { STACK_PRINT(tracer, "X",  postbyte & 0xe0); }
+			if (postbyte & 0x20) { STACK_PRINT(tracer, "Y",  postbyte & 0xc0); }
+			if (postbyte & 0x40) { STACK_PRINT(tracer, "S",  postbyte & 0x80); }
+			if (postbyte & 0x80) { STACK_PRINT(tracer, "PC", 0); }
 		} break;
 
 		case REGISTER:
@@ -1215,9 +1229,12 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 				snprintf(tracer->operand_text, sizeof(tracer->operand_text), idx_fmts[tracer->idx_mode], pre, tracer->value, tracer->idx_reg, post);
 				break;
 			case IDX_PCR8:
-				snprintf(tracer->operand_text, sizeof(tracer->operand_text), idx_fmts[tracer->idx_mode], pre, (value8<0)?"-":"", (value8<0)?-value8:value8, post);
+				snprintf(tracer->operand_text, sizeof(tracer->operand_text), idx_fmts[tracer->idx_mode], pre, (uint16_t)(tracer->instr_pc + tracer->nbytes + value8), post);
 				break;
-			case IDX_PCR16: case IDX_EXT16:
+			case IDX_PCR16:
+				snprintf(tracer->operand_text, sizeof(tracer->operand_text), idx_fmts[tracer->idx_mode], pre, (uint16_t)(tracer->instr_pc + tracer->nbytes + tracer->value), post);
+				break;
+			case IDX_EXT16:
 				snprintf(tracer->operand_text, sizeof(tracer->operand_text), idx_fmts[tracer->idx_mode], pre, tracer->value, post);
 				break;
 			case IDX_ILL_A: case IDX_ILL_E:
@@ -1237,51 +1254,66 @@ void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc) {
 			snprintf(tracer->operand_text, sizeof(tracer->operand_text), "$%04x", pc);
 			break;
 
-		// CPU code will not call trace_print after IRQ vector fetch
-		// and before the next instruction, therefore the state list
-		// for IRQ vectors skips an expected dummy byte, and this
-		// prints the trace line early.
-
 		case IRQVECTOR:
-			trace_print_short(tracer);
-			printf("\n");
-			fflush(stdout);
 			break;
 
 		default:
 			break;
 	}
+
 }
 
-/* Called just before an IRQ vector fetch */
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-void mc6809_trace_irq(struct mc6809_trace *tracer, int vector) {
-	reset_state(tracer);
-	tracer->state = WANT_IRQ_VECTOR;
-	tracer->bytes_count = 0;
-	tracer->mnemonic = irq_names[(vector & 15) >> 1];
+static void reset_state(struct mc6809_trace *tracer) {
+	tracer->state = WANT_PRINT;
+	tracer->page = PAGE0;
+	tracer->instr_pc = 0;
+	tracer->nbytes = 0;
+	tracer->mnemonic = "*";
+	strcpy(tracer->operand_text, "*");
+
+	tracer->ins_type = PAGE0;
+	tracer->state_list = NULL;
+	tracer->idx_mode = 0;
+	tracer->idx_reg = "";
+	tracer->idx_indirect = 0;
 }
 
-/* Called after each instruction */
-
-void mc6809_trace_print(struct mc6809_trace *tracer) {
-	if (tracer->state != WANT_PRINT) return;
-	trace_print_short(tracer);
+static void print_record(struct mc6809_trace *tracer) {
 	struct MC6809 *cpu = tracer->cpu;
-	printf("cc=%02x a=%02x b=%02x dp=%02x "
-	       "x=%04x y=%04x u=%04x s=%04x\n",
-	       cpu->reg_cc, MC6809_REG_A(cpu), MC6809_REG_B(cpu), cpu->reg_dp,
-	       cpu->reg_x, cpu->reg_y, cpu->reg_u, cpu->reg_s);
+
+	if (tracer->nbytes == 0) {
+		// Nothing to print.
+		return;
+	}
+
+	// XXX currently no way to reset start_tick when turning trace mode
+	// on/off, so first instruction after switching on will have crazy dt.
+
+	int dt = event_tick_delta(event_current_tick, tracer->start_tick);
+	tracer->start_tick = event_current_tick;
+
+	char bytes_string[(MAX_NBYTES*2)+1];
+	for (unsigned i = 0; i < tracer->nbytes; i++) {
+		snprintf(bytes_string + i*2, 3, "%02x", tracer->bytes[i]);
+	}
+
+	if (tracer->ins_type == IRQVECTOR) {
+		printf("%04x| %-12s%-24s", tracer->instr_pc, bytes_string, tracer->mnemonic);
+	} else {
+		printf("%04x| %-12s%-6s%-18s", tracer->instr_pc, bytes_string, tracer->mnemonic, tracer->operand_text);
+		printf("  cc=%02x a=%02x b=%02x dp=%02x "
+		       "x=%04x y=%04x u=%04x s=%04x",
+		       cpu->reg_cc, MC6809_REG_A(cpu), MC6809_REG_B(cpu), cpu->reg_dp,
+		       cpu->reg_x, cpu->reg_y, cpu->reg_u, cpu->reg_s);
+	}
+
+	if (logging.trace_cpu_timing) {
+		printf("  dt=%d", dt);
+	}
+
+	printf("\n");
 	fflush(stdout);
 	reset_state(tracer);
 }
-
-static void trace_print_short(struct mc6809_trace *tracer) {
-	char bytes_string[(BYTES_BUF_SIZE*2)+1];
-	if (tracer->bytes_count == 0) return;
-	for (int i = 0; i < tracer->bytes_count; i++) {
-		snprintf(bytes_string + i*2, 3, "%02x", tracer->bytes_buf[i]);
-	}
-	printf("%04x| %-12s%-8s%-20s", tracer->instr_pc, bytes_string, tracer->mnemonic, tracer->operand_text);
-	reset_state(tracer);
-}
