@@ -2,7 +2,7 @@
  *
  *  \brief Motorola MC6809 CPU tracing.
  *
- *  \copyright Copyright 2005-2017 Ciaran Anscomb
+ *  \copyright Copyright 2005-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -24,11 +24,10 @@
 struct mc6809_trace;
 
 struct mc6809_trace *mc6809_trace_new(struct MC6809 *cpu);
-void mc6809_trace_free(struct mc6809_trace *tracer);
+void mc6809_trace_free(struct mc6809_trace *);
 
-void mc6809_trace_reset(struct mc6809_trace *tracer);
-void mc6809_trace_byte(struct mc6809_trace *tracer, uint8_t byte, uint16_t pc);
-void mc6809_trace_irq(struct mc6809_trace *tracer, int vector);
-void mc6809_trace_print(struct mc6809_trace *tracer);
+void mc6809_trace_vector(struct mc6809_trace *);
+void mc6809_trace_instruction(struct mc6809_trace *);
+void mc6809_trace_byte(struct mc6809_trace *, uint8_t byte, uint16_t pc);
 
 #endif
