@@ -97,6 +97,11 @@ struct MC6847_private {
 	unsigned render_mode;
 	unsigned pal_padding;
 
+	// Set when pixel_data[] contains only the current border colour - PAL
+	// padding lines (which occur during vertical border) will invalidate
+	// this.
+	_Bool have_border_only;
+
 	/* Unsafe warning: pixel_data[] needs to be 8 elements longer than a
 	 * full scanline, for the mid-scanline 32 -> 16 byte mode switch case
 	 * where many extra pixels are emitted.  8 is the maximum number of
@@ -124,55 +129,55 @@ struct MC6847_private {
 #define MC6847_SER_VRAM (34)
 
 static struct ser_struct ser_struct_mc6847[] = {
-	SER_ID_STRUCT_ELEM(1, ser_type_unsigned, struct MC6847, row),
-	SER_ID_STRUCT_ELEM(2, ser_type_bool, struct MC6847, is_dragon64),
-	SER_ID_STRUCT_ELEM(3, ser_type_bool, struct MC6847, is_dragon32),
-	SER_ID_STRUCT_ELEM(4, ser_type_bool, struct MC6847, is_coco),
-	SER_ID_STRUCT_ELEM(5, ser_type_bool, struct MC6847, is_pal),
+	SER_ID_STRUCT_ELEM(1, struct MC6847, row),
+	SER_ID_STRUCT_ELEM(2, struct MC6847, is_dragon64),
+	SER_ID_STRUCT_ELEM(3, struct MC6847, is_dragon32),
+	SER_ID_STRUCT_ELEM(4, struct MC6847, is_coco),
+	SER_ID_STRUCT_ELEM(5, struct MC6847, is_pal),
 
-	SER_ID_STRUCT_ELEM(6, ser_type_unsigned, struct MC6847_private, GM),
-	SER_ID_STRUCT_ELEM(7, ser_type_bool, struct MC6847_private, nA_S),
-	SER_ID_STRUCT_ELEM(8, ser_type_bool, struct MC6847_private, nA_G),
-	SER_ID_STRUCT_ELEM(9, ser_type_bool, struct MC6847_private, EXT),
-	SER_ID_STRUCT_ELEM(10, ser_type_bool, struct MC6847_private, CSS),
-	SER_ID_STRUCT_ELEM(11, ser_type_bool, struct MC6847_private, CSSa),
-	SER_ID_STRUCT_ELEM(12, ser_type_bool, struct MC6847_private, CSSb),
-	SER_ID_STRUCT_ELEM(13, ser_type_bool, struct MC6847_private, inverted_text),
+	SER_ID_STRUCT_ELEM(6, struct MC6847_private, GM),
+	SER_ID_STRUCT_ELEM(7, struct MC6847_private, nA_S),
+	SER_ID_STRUCT_ELEM(8, struct MC6847_private, nA_G),
+	SER_ID_STRUCT_ELEM(9, struct MC6847_private, EXT),
+	SER_ID_STRUCT_ELEM(10, struct MC6847_private, CSS),
+	SER_ID_STRUCT_ELEM(11, struct MC6847_private, CSSa),
+	SER_ID_STRUCT_ELEM(12, struct MC6847_private, CSSb),
+	SER_ID_STRUCT_ELEM(13, struct MC6847_private, inverted_text),
 
-	SER_ID_STRUCT_ELEM(14, ser_type_event, struct MC6847_private, hs_fall_event),
-	SER_ID_STRUCT_ELEM(15, ser_type_event, struct MC6847_private, hs_rise_event),
-	SER_ID_STRUCT_ELEM(16, ser_type_tick, struct MC6847_private, scanline_start),
-	SER_ID_STRUCT_ELEM(17, ser_type_unsigned, struct MC6847_private, beam_pos),
-	SER_ID_STRUCT_ELEM(18, ser_type_unsigned, struct MC6847_private, scanline),
+	SER_ID_STRUCT_TYPE(14, ser_type_event, struct MC6847_private, hs_fall_event),
+	SER_ID_STRUCT_TYPE(15, ser_type_event, struct MC6847_private, hs_rise_event),
+	SER_ID_STRUCT_TYPE(16, ser_type_tick, struct MC6847_private, scanline_start),
+	SER_ID_STRUCT_ELEM(17, struct MC6847_private, beam_pos),
+	SER_ID_STRUCT_ELEM(18, struct MC6847_private, scanline),
 
-	SER_ID_STRUCT_ELEM(19, ser_type_uint8, struct MC6847_private, vram_g_data),
-	SER_ID_STRUCT_ELEM(20, ser_type_uint8, struct MC6847_private, vram_sg_data),
+	SER_ID_STRUCT_ELEM(19, struct MC6847_private, vram_g_data),
+	SER_ID_STRUCT_ELEM(20, struct MC6847_private, vram_sg_data),
 
-	SER_ID_STRUCT_ELEM(21, ser_type_bool, struct MC6847_private, is_32byte),
-	SER_ID_STRUCT_ELEM(22, ser_type_bool, struct MC6847_private, GM0),
-	SER_ID_STRUCT_ELEM(23, ser_type_uint8, struct MC6847_private, s_fg_colour),
-	SER_ID_STRUCT_ELEM(24, ser_type_uint8, struct MC6847_private, s_bg_colour),
-	SER_ID_STRUCT_ELEM(25, ser_type_uint8, struct MC6847_private, fg_colour),
-	SER_ID_STRUCT_ELEM(26, ser_type_uint8, struct MC6847_private, bg_colour),
-	SER_ID_STRUCT_ELEM(27, ser_type_uint8, struct MC6847_private, cg_colours),
-	SER_ID_STRUCT_ELEM(28, ser_type_uint8, struct MC6847_private, border_colour),
-	SER_ID_STRUCT_ELEM(29, ser_type_uint8, struct MC6847_private, bright_orange),
-	SER_ID_STRUCT_ELEM(30, ser_type_int, struct MC6847_private, vram_bit),
-	SER_ID_STRUCT_ELEM(31, ser_type_unsigned, struct MC6847_private, render_mode),
-	SER_ID_STRUCT_ELEM(32, ser_type_unsigned, struct MC6847_private, pal_padding),
+	SER_ID_STRUCT_ELEM(21, struct MC6847_private, is_32byte),
+	SER_ID_STRUCT_ELEM(22, struct MC6847_private, GM0),
+	SER_ID_STRUCT_ELEM(23, struct MC6847_private, s_fg_colour),
+	SER_ID_STRUCT_ELEM(24, struct MC6847_private, s_bg_colour),
+	SER_ID_STRUCT_ELEM(25, struct MC6847_private, fg_colour),
+	SER_ID_STRUCT_ELEM(26, struct MC6847_private, bg_colour),
+	SER_ID_STRUCT_ELEM(27, struct MC6847_private, cg_colours),
+	SER_ID_STRUCT_ELEM(28, struct MC6847_private, border_colour),
+	SER_ID_STRUCT_ELEM(29, struct MC6847_private, bright_orange),
+	SER_ID_STRUCT_ELEM(30, struct MC6847_private, vram_bit),
+	SER_ID_STRUCT_ELEM(31, struct MC6847_private, render_mode),
+	SER_ID_STRUCT_ELEM(32, struct MC6847_private, pal_padding),
 
-	SER_ID_STRUCT_ELEM(33, ser_type_unsigned, struct MC6847_private, burst),
+	SER_ID_STRUCT_ELEM(33, struct MC6847_private, burst),
 
 	SER_ID_STRUCT_UNHANDLED(MC6847_SER_VRAM),
-	SER_ID_STRUCT_ELEM(35, ser_type_unsigned, struct MC6847_private, vram_index),
-	SER_ID_STRUCT_ELEM(36, ser_type_unsigned, struct MC6847_private, vram_nbytes),
+	SER_ID_STRUCT_ELEM(35, struct MC6847_private, vram_index),
+	SER_ID_STRUCT_ELEM(36, struct MC6847_private, vram_nbytes),
 
-	SER_ID_STRUCT_ELEM(37, ser_type_unsigned, struct MC6847_private, lborder_remaining),
-	SER_ID_STRUCT_ELEM(38, ser_type_unsigned, struct MC6847_private, vram_remaining),
-	SER_ID_STRUCT_ELEM(39, ser_type_unsigned, struct MC6847_private, rborder_remaining),
+	SER_ID_STRUCT_ELEM(37, struct MC6847_private, lborder_remaining),
+	SER_ID_STRUCT_ELEM(38, struct MC6847_private, vram_remaining),
+	SER_ID_STRUCT_ELEM(39, struct MC6847_private, rborder_remaining),
 
-	SER_ID_STRUCT_ELEM(40, ser_type_unsigned, struct MC6847_private, nLPR),
-	SER_ID_STRUCT_ELEM(41, ser_type_bool, struct MC6847_private, is_t1),
+	SER_ID_STRUCT_ELEM(40, struct MC6847_private, nLPR),
+	SER_ID_STRUCT_ELEM(41, struct MC6847_private, is_t1),
 };
 
 static _Bool mc6847_read_elem(void *sptr, struct ser_handle *sh, int tag);
@@ -214,8 +219,8 @@ static const struct partdb_entry_funcs mc6847_funcs = {
 	.ser_struct_data = &mc6847_ser_struct_data,
 };
 
-const struct partdb_entry mc6847_part = { .name = "MC6847", .funcs = &mc6847_funcs };
-const struct partdb_entry mc6847t1_part = { .name = "MC6847T1", .funcs = &mc6847_funcs };
+const struct partdb_entry mc6847_part = { .name = "MC6847", .description = "Motorola | MC6847 VDG", .funcs = &mc6847_funcs };
+const struct partdb_entry mc6847t1_part = { .name = "MC6847T1", .description = "Motorola | MC6847T1 VDG", .funcs = &mc6847_funcs };
 
 static struct part *mc6847_allocate(void) {
 	struct MC6847_private *vdg = part_new(sizeof(*vdg));
@@ -300,13 +305,12 @@ static void do_hs_fall(void *data) {
 	// Finish rendering previous scanline
 	if (vdg->frame == 0) {
 		if (vdg->scanline < VDG_ACTIVE_AREA_START) {
-			if (vdg->scanline == 0) {
-				uint8_t *v = vdg->pixel_data + VDG_LEFT_BORDER_START;
-				for (unsigned j = VDG_tAVB; j > 0; j--) {
-					*(v++) = vdg->border_colour;
-				}
+			if (!vdg->have_border_only) {
+				memset(vdg->pixel_data + VDG_LEFT_BORDER_START, vdg->border_colour, VDG_tAVB);
+				vdg->have_border_only = 1;
 			}
 		} else if (vdg->scanline >= VDG_ACTIVE_AREA_START && vdg->scanline < VDG_ACTIVE_AREA_END) {
+			vdg->have_border_only = 0;
 			render_scanline(vdg);
 			vdg->public.row++;
 			if (vdg->public.row > 11)
@@ -315,11 +319,9 @@ static void do_hs_fall(void *data) {
 				vdg->A += vdg->is_32byte ? 32 : 16;
 			vdg->beam_pos = VDG_LEFT_BORDER_START;
 		} else if (vdg->scanline >= VDG_ACTIVE_AREA_END) {
-			if (vdg->scanline == VDG_ACTIVE_AREA_END) {
-				uint8_t *v = vdg->pixel_data + VDG_LEFT_BORDER_START;
-				for (unsigned j = VDG_tAVB; j > 0; j--) {
-					*(v++) = vdg->border_colour;
-				}
+			if (!vdg->have_border_only) {
+				memset(vdg->pixel_data + VDG_LEFT_BORDER_START, vdg->border_colour, VDG_tAVB);
+				vdg->have_border_only = 1;
 			}
 		}
 	}
@@ -344,14 +346,20 @@ static void do_hs_fall(void *data) {
 		if (!vdg->public.is_coco) {
 			if (vdg->scanline == SCANLINE(VDG_ACTIVE_AREA_END + 24)
 			    || vdg->scanline == SCANLINE(VDG_ACTIVE_AREA_END + 32)) {
+				memset(vdg->pixel_data + VDG_LEFT_BORDER_START, VDG_BLACK, VDG_tAVB);
+				vdg->have_border_only = 0;
 				vdg->pal_padding = 25;
 				vdg->hs_fall_event.delegate.func = do_hs_fall_pal;
 			}
 		} else {
 			if (vdg->scanline == SCANLINE(VDG_ACTIVE_AREA_END + 26)) {
+				memset(vdg->pixel_data + VDG_LEFT_BORDER_START, VDG_BLACK, VDG_tAVB);
+				vdg->have_border_only = 0;
 				vdg->pal_padding = 26;
 				vdg->hs_fall_event.delegate.func = do_hs_fall_pal;
 			} else if (vdg->scanline == SCANLINE(VDG_ACTIVE_AREA_END + 48)) {
+				memset(vdg->pixel_data + VDG_LEFT_BORDER_START, VDG_BLACK, VDG_tAVB);
+				vdg->have_border_only = 0;
 				vdg->pal_padding = 24;
 				vdg->hs_fall_event.delegate.func = do_hs_fall_pal;
 			}
@@ -617,7 +625,7 @@ static void render_scanline(struct MC6847_private *vdg) {
 
 void mc6847_reset(struct MC6847 *vdgp) {
 	struct MC6847_private *vdg = (struct MC6847_private *)vdgp;
-	memset(vdg->pixel_data, VDG_BLACK, sizeof(vdg->pixel_data));
+	memset(vdg->pixel_data, 0, sizeof(vdg->pixel_data));
 	vdg->beam_pos = VDG_LEFT_BORDER_START;
 	vdg->scanline = 0;
 	vdg->public.row = 0;
