@@ -2,7 +2,7 @@
  *
  *  \brief Motorola SN74LS783/MC6883 Synchronous Address Multiplexer.
  *
- *  \copyright Copyright 2003-2022 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -28,16 +28,22 @@
 struct MC6883 {
 	struct part part;
 
-	unsigned S;
-	unsigned Z;
-	unsigned V;
-	_Bool RAS;
+	uint8_t S;
+	unsigned Zrow;
+	unsigned Zcol;
+	unsigned Vrow;
+	unsigned Vcol;
+
+	_Bool RAS0;
+	_Bool RAS1;
+
 	DELEGATE_T3(void, int, bool, uint16) cpu_cycle;
 	DELEGATE_T0(void) vdg_update;
 };
 
 void mc6883_reset(struct MC6883 *);
 void mc6883_mem_cycle(void *, _Bool RnW, uint16_t A);
+unsigned mc6883_decode(struct MC6883 *, _Bool RnW, uint16_t A);
 void mc6883_vdg_hsync(struct MC6883 *, _Bool level);
 void mc6883_vdg_fsync(struct MC6883 *, _Bool level);
 int mc6883_vdg_bytes(struct MC6883 *, int nbytes);
