@@ -2,7 +2,7 @@
  *
  *  \brief Serialisation and deserialisation helpers.
  *
- *  \copyright Copyright 2015-2022 Ciaran Anscomb
+ *  \copyright Copyright 2015-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -16,12 +16,29 @@
  *  \endlicenseblock
  */
 
+// If using ser_write_struct_data()/ser_read_struct_data() and there's need to
+// change the type in a data structure, there are certain changes that are safe
+// with no other modification other than the type in the ser_struct entry,
+// because they are actually serialised in the same way:
+//
+// Signed types: ser_type_int, ser_type_int8, ser_type_int16, ser_type_int32
+//
+// Unsigned types: ser_type_bool, ser_type_unsigned, ser_type_uint8,
+// ser_type_uint16, ser_type_uint32
+//
+// Most other changes will require a per-element handler for compatibility
+// (being flagged in the ser_struct entry as "unhandled" by
+// ser_*_struct_data()).
+
 #include "top-config.h"
 
 // Comment this out for debugging
 #define SER_DEBUG(...)
 
+#define _POSIX_C_SOURCE 200112L
+
 #include <assert.h>
+#include <errno.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,12 +54,16 @@
 #include "serialise.h"
 
 #ifndef SER_DEBUG
-#define SER_DEBUG(...) LOG_PRINT(__VA_ARGS__)
+#define SER_DEBUG(...) LOG_DEBUG(__VA_ARGS__)
+#define SER_MOD_DEBUG(l,...) LOG_MOD_DEBUG(l, "serialise", __VA_ARGS__)
+#else
+#define SER_MOD_DEBUG(l,...)
 #endif
 
 struct ser_handle {
 	FILE *fd;
 	int error;
+	off_t error_pos;
 
 	// After reading a (TAG,LENGTH), this will contain LENGTH.  Attempts to
 	// read more than this many bytes as data will cause an error.  Any
@@ -58,13 +79,11 @@ struct ser_handle {
 
 static void s_write_uint8(struct ser_handle *sh, int v);
 static void s_write_uint16(struct ser_handle *sh, int v);
-static void s_write_uint32(struct ser_handle *sh, uint32_t v);
 static void s_write_vuint32(struct ser_handle *sh, uint32_t v);
 static void s_write_vint32(struct ser_handle *sh, int32_t v);
 static void s_write(struct ser_handle *sh, const void *ptr, size_t size);
 static int s_read_uint8(struct ser_handle *sh);
 static int s_read_uint16(struct ser_handle *sh);
-static uint32_t s_read_uint32(struct ser_handle *sh);
 static uint32_t s_read_vuint32(struct ser_handle *sh);
 static void s_read(struct ser_handle *sh, void *ptr, size_t size);
 static void *s_read_new(struct ser_handle *sh, size_t size);
@@ -72,12 +91,17 @@ static void *s_read_new(struct ser_handle *sh, size_t size);
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 struct ser_handle *ser_open(const char *filename, enum ser_mode mode) {
-	FILE *fd;
+	if (!filename) {
+		errno = EINVAL;
+		return NULL;
+	}
+	FILE *fd = NULL;
 	if (mode == ser_mode_read) {
 		fd = fopen(filename, "rb");
 	} else if (mode == ser_mode_write) {
 		fd = fopen(filename, "wb");
-	} else {
+	}
+	if (!fd) {
 		return NULL;
 	}
 	struct ser_handle *sh = xmalloc(sizeof(*sh));
@@ -103,7 +127,7 @@ void ser_write_tag(struct ser_handle *sh, int tag, size_t length) {
 		ser_set_error(sh, ser_error_bad_tag);
 		return;
 	}
-	SER_DEBUG("ser_write_tag(%d, %zd)\n", tag, length);
+	SER_MOD_DEBUG(2, "ser_write_tag(%d, %zd)\n", tag, length);
 	s_write_vuint32(sh, tag);
 	s_write_vuint32(sh, length);
 	sh->length = length;
@@ -114,7 +138,7 @@ void ser_write_close_tag(struct ser_handle *sh) {
 		return;
 	// XXX handle this case more gracefully (e.g. write padding bytes)
 	assert(sh->length == 0);
-	SER_DEBUG("ser_write_close_tag()\n");
+	SER_MOD_DEBUG(2, "ser_write_close_tag()\n");
 	s_write_vint32(sh, 0);
 }
 
@@ -124,7 +148,7 @@ int ser_read_tag(struct ser_handle *sh) {
 
 	// Skip any data remaining from previous read
 	if (sh->length) {
-		SER_DEBUG("ser_read_tag(): skipping %zd bytes\n", sh->length);
+		SER_MOD_DEBUG(2, "ser_read_tag(): skipping %zd bytes\n", sh->length);
 		if (fseek(sh->fd, sh->length, SEEK_CUR) < 0) {
 			ser_set_error(sh, ser_error_file_io);
 			return -1;
@@ -133,7 +157,7 @@ int ser_read_tag(struct ser_handle *sh) {
 	}
 
 	int tag = s_read_vuint32(sh);
-	SER_DEBUG("ser_read_tag(): %d\n", tag);
+	SER_MOD_DEBUG(2, "ser_read_tag(): %d\n", tag);
 
 	if (tag == 0) {
 		// Closing tag (special value zero).
@@ -179,9 +203,11 @@ int ser_error(struct ser_handle *sh) {
 void ser_set_error(struct ser_handle *sh, int error) {
 	assert(sh != NULL);
 	// Don't hide earlier errors:
-	if (sh->error)
+	if (sh->error) {
 		return;
+	}
 	sh->error = error;
+	sh->error_pos = ftello(sh->fd);
 }
 
 const char *ser_errstr(struct ser_handle *sh) {
@@ -198,6 +224,12 @@ const char *ser_errstr(struct ser_handle *sh) {
 	}
 }
 
+ssize_t ser_errpos(struct ser_handle *sh) {
+	if (!sh) {
+		return -1;
+	}
+	return sh->error_pos;
+}
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
@@ -218,11 +250,6 @@ static void s_write_uint16(struct ser_handle *sh, int v) {
 		ser_set_error(sh, ser_error_file_io);
 }
 
-static void s_write_uint32(struct ser_handle *sh, uint32_t v) {
-	s_write_uint16(sh, v >> 16);
-	s_write_uint16(sh, v & 0xffff);
-}
-
 static void s_write_vuint32(struct ser_handle *sh, uint32_t v) {
 	if (sh->error)
 		return;
@@ -262,11 +289,6 @@ static int s_read_uint16(struct ser_handle *sh) {
 	return r;
 }
 
-static uint32_t s_read_uint32(struct ser_handle *sh) {
-	uint32_t r = s_read_uint16(sh) << 16;
-	return r | s_read_uint16(sh);
-}
-
 static uint32_t s_read_vuint32(struct ser_handle *sh) {
 	if (sh->error)
 		return 0;
@@ -345,8 +367,12 @@ void ser_write_vuint32(struct ser_handle *sh, int tag, uint32_t v) {
 }
 
 void ser_write_string(struct ser_handle *sh, int tag, const char *s) {
-	size_t length = s ? strlen(s) : 0;
-	ser_write(sh, tag, s, length);
+	if (!s) {
+		char dummy = '\0';
+		ser_write(sh, tag, &dummy, 1);
+	} else {
+		ser_write(sh, tag, s, strlen(s));
+	}
 }
 
 void ser_write_sds(struct ser_handle *sh, int tag, const sds s) {
@@ -354,6 +380,20 @@ void ser_write_sds(struct ser_handle *sh, int tag, const sds s) {
 	ser_write(sh, tag, s, length);
 }
 
+void ser_write_array_uint8(struct ser_handle *sh, int tag, uint8_t *src, size_t nelems) {
+	return ser_write(sh, tag, src, nelems);
+}
+
+void ser_write_array_uint16(struct ser_handle *sh, int tag, uint16_t *src, size_t nelems) {
+	if (!sh)
+		return;
+	ser_write_tag(sh, tag, nelems * 2);
+	for (size_t i = 0; i < nelems; i++) {
+		ser_write_uint16_untagged(sh, src[i]);
+	}
+	ser_write_close_tag(sh);
+}
+
 void ser_write(struct ser_handle *sh, int tag, const void *ptr, size_t size) {
 	if (!sh)
 		return;
@@ -507,6 +547,32 @@ void ser_read(struct ser_handle *sh, void *ptr, size_t size) {
 	s_read(sh, ptr, size);
 }
 
+size_t ser_read_array_uint8(struct ser_handle *sh, uint8_t **dst, size_t nelems) {
+	if (!sh || sh->error)
+		return 0;
+	assert(dst != NULL);
+	if (!nelems || nelems > sh->length)
+		nelems = sh->length;
+	if (!(*dst))
+		*dst = xmalloc(nelems);
+	ser_read(sh, *dst, nelems);
+	return nelems;
+}
+
+size_t ser_read_array_uint16(struct ser_handle *sh, uint16_t **dst, size_t nelems) {
+	if (!sh || sh->error)
+		return 0;
+	assert(dst != NULL);
+	if (!nelems || nelems > (sh->length / 2))
+		nelems = sh->length / 2;
+	if (!(*dst))
+		*dst = xmalloc(nelems * 2);
+	uint16_t *data = *dst;
+	for (size_t i = 0; i < nelems; i++)
+		data[i] = s_read_uint16(sh);
+	return nelems;
+}
+
 // These allocate their own storage:
 
 char *ser_read_string(struct ser_handle *sh) {
@@ -515,6 +581,10 @@ char *ser_read_string(struct ser_handle *sh) {
 	char *s = xmalloc(sh->length+1);
 	s_read(sh, s, sh->length);
 	s[sh->length] = 0;
+	if (sh->length == 1 && !*s) {
+		free(s);
+		s = NULL;
+	}
 	sh->length = 0;
 	return s;
 }
@@ -544,42 +614,11 @@ void *ser_read_new(struct ser_handle *sh, size_t size) {
 
 void ser_write_struct_data(struct ser_handle *sh, const struct ser_struct_data *ssd, void *s) {
 	const struct ser_struct *ss = ssd->elems;
-	for (int i = 0; i < ssd->num_elems && !sh->error; i++) {
+	for (unsigned i = 0; i < ssd->num_elems && !sh->error; ++i) {
 		int tag = ss[i].tag;
 		enum ser_type type = ss[i].type;
-		SER_DEBUG("ser_write_struct(): tag=%d type=%d alength=%d\n", tag, type, ss[i].alength);
-		void *ptr = s + ss[i].offset;
-
-		if (ss[i].alength > 0) {
-			// write array
-			switch (type) {
-			case ser_type_uint8:
-				ser_write_tag(sh, tag, ss[i].alength);
-				for (int j = 0; !sh->error && j < ss[i].alength; j++) {
-					s_write_uint8(sh, ((uint8_t *)ptr)[j]);
-				}
-				break;
-			case ser_type_uint16:
-				ser_write_tag(sh, tag, ss[i].alength * 2);
-				for (int j = 0; !sh->error && j < ss[i].alength; j++) {
-					s_write_uint16(sh, ((uint16_t *)ptr)[j]);
-				}
-				break;
-			case ser_type_uint32:
-				ser_write_tag(sh, tag, ss[i].alength * 4);
-				for (int j = 0; !sh->error && j < ss[i].alength; j++) {
-					s_write_uint32(sh, ((uint32_t *)ptr)[j]);
-				}
-				break;
-			default:
-				ser_set_error(sh, ser_error_type);
-				break;
-			}
-			if (ser_error(sh)) {
-				SER_DEBUG("serialisation ERROR!\n");
-			}
-			continue;
-		}
+		SER_MOD_DEBUG(2, "ser_write_struct(): tag=%d type=%d\n", tag, type);
+		void *ptr = (char *)s + ss[i].offset;
 
 		switch (type) {
 
@@ -673,7 +712,7 @@ void ser_write_struct_data(struct ser_handle *sh, const struct ser_struct_data *
 			assert(ssd->write_elem != NULL);
 			if (!ssd->write_elem(s, sh, tag)) {
 				ser_set_error(sh, ser_error_bad_tag);
-				SER_DEBUG("ser_write_struct_data(%p) FAILED\n", ss);
+				SER_MOD_DEBUG(2, "ser_write_struct_data(%p) FAILED\n", ss);
 			}
 			break;
 
@@ -682,50 +721,30 @@ void ser_write_struct_data(struct ser_handle *sh, const struct ser_struct_data *
 			break;
 		}
 		if (ser_error(sh)) {
-			SER_DEBUG("serialisation ERROR!\n");
+			SER_MOD_DEBUG(2, "serialisation ERROR!\n");
 		}
 	}
 	ser_write_close_tag(sh);
-	SER_DEBUG("ser_write_struct_data(%p) finished\n", ss);
+	SER_MOD_DEBUG(2, "ser_write_struct_data(%p) finished\n", ss);
 }
 
 void ser_read_struct_data(struct ser_handle *sh, const struct ser_struct_data *ssd, void *s) {
 	const struct ser_struct *ss = ssd->elems;
 	int tag;
 	while (!sh->error && (tag = ser_read_tag(sh)) > 0) {
-		int i;
-		for (i = 0; i < ssd->num_elems; i++) {
-			if (ss[i].tag == tag)
+		unsigned i;
+		for (i = 0; i < ssd->num_elems; ++i) {
+			if (ss[i].tag == tag) {
 				break;
+			}
 		}
 		if (i >= ssd->num_elems) {
 			continue;
 		}
 
 		enum ser_type type = ss[i].type;
-		void *ptr = s + ss[i].offset;
-		SER_DEBUG("ser_read_struct(): tag=%d type=%d alength=%d\n", tag, type, ss[i].alength);
-
-		if (ss[i].alength > 0) {
-			// read array
-			for (int j = 0; !sh->error && j < ss[j].alength; j++) {
-				switch (type) {
-				case ser_type_uint8:
-					((uint8_t *)ptr)[j] = s_read_uint8(sh);
-					break;
-				case ser_type_uint16:
-					((uint16_t *)ptr)[j] = s_read_uint16(sh);
-					break;
-				case ser_type_uint32:
-					((uint32_t *)ptr)[j] = s_read_uint32(sh);
-					break;
-				default:
-					ser_set_error(sh, ser_error_type);
-					break;
-				}
-			}
-			continue;
-		}
+		void *ptr = (char *)s + ss[i].offset;
+		SER_MOD_DEBUG(2, "ser_read_struct(): tag=%d type=%d\n", tag, type);
 
 		switch (type) {
 
@@ -814,7 +833,7 @@ void ser_read_struct_data(struct ser_handle *sh, const struct ser_struct_data *s
 		case ser_type_unhandled:
 			if (!ssd->read_elem || !ssd->read_elem(s, sh, tag)) {
 				ser_set_error(sh, ser_error_bad_tag);
-				SER_DEBUG("ser_read_struct_data(%p) FAILED\n", ss);
+				SER_MOD_DEBUG(2, "ser_read_struct_data(%p) FAILED\n", ss);
 			}
 			break;
 
