@@ -2,7 +2,7 @@
  *
  *  \brief Snapshotting of emulated system.
  *
- *  \copyright Copyright 2003-2022 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -19,6 +19,7 @@
 #include "top-config.h"
 
 #include <assert.h>
+#include <errno.h>
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
@@ -27,6 +28,7 @@
 #include "c-strcase.h"
 #include "xalloc.h"
 
+#include "auto_kbd.h"
 #include "cart.h"
 #include "fs.h"
 #include "keyboard.h"
@@ -38,6 +40,7 @@
 #include "mc6847/mc6847.h"
 #include "mc6883.h"
 #include "part.h"
+#include "ram.h"
 #include "serialise.h"
 #include "snapshot.h"
 #include "tape.h"
@@ -73,6 +76,10 @@ static int read_v2_snapshot(const char *filename);
 #ifdef WANT_V1_SNAPSHOTS
 const char *snapv1_header = "XRoar snapshot.\012\000";
 static int read_v1_snapshot(const char *filename);
+#else
+#ifndef HAVE_WASM
+static int read_v1_snapshot(const char *f) { (void)f; return -1; }
+#endif
 #endif
 
 // For WebAssembly, read the snapshot in, but if that resulted in any pending
@@ -93,22 +100,21 @@ int read_snapshot(const char *filename) {
 #ifndef HAVE_WASM
 	if (read_v2_snapshot(filename) < 0 &&
 	    read_v1_snapshot(filename) < 0) {
-		LOG_WARN("Snapshot: read failed\n");
+		LOG_MOD_WARN("snapshot", "%s: read failed\n", filename);
 		return -1;
 	}
 #else
-	wasm_retry_open = 1;
-	if (read_v2_snapshot(filename) < 0) {
-		wasm_retry_open = 0;
-		LOG_WARN("Snapshot: read failed\n");
-		return -1;
-	}
-	wasm_retry_open = 0;
-	if (wasm_waiting_files) {
-		char *fncopy = xstrdup(filename);
-		event_queue_auto(&UI_EVENT_LIST, DELEGATE_AS0(void, do_retry_read_snapshot, fncopy), 1);
-		return -1;
+	read_v2_snapshot(filename);
+	if (xroar.machine) {
+		wasm_ui_prepare_machine(xroar.machine->config);
+		struct cart *c = xroar.machine->get_interface(xroar.machine, "cart");
+		if (c) {
+			wasm_ui_prepare_cartridge(c->config);
+		}
 	}
+	char *fncopy = xstrdup(filename);
+	event_queue_auto(&UI_EVENT_LIST, DELEGATE_AS0(void, do_retry_read_snapshot, fncopy), 1);
+	return -1;
 #endif
 	return 0;
 }
@@ -118,29 +124,33 @@ int read_snapshot(const char *filename) {
 int write_snapshot(const char *filename) {
 	if (xroar_filetype_by_ext(filename) == FILETYPE_RAM) {
 		// simple ram dump
-		if (!xroar_machine->dump_ram) {
-			LOG_WARN("Running machine does not support RAM dump.\n");
+		if (!xroar.machine->dump_ram) {
+			LOG_MOD_WARN("snapshot/ram", "machine does not support RAM dump\n");
 			return -1;
 		}
 		FILE *fd = fopen(filename, "wb");
-		if (!fd)
+		if (!fd) {
+			LOG_MOD_WARN("snapshot/ram", "%s: %s\n", filename, strerror(errno));
 			return -1;
-		xroar_machine->dump_ram(xroar_machine, fd);
+		}
+		xroar.machine->dump_ram(xroar.machine, fd);
 		fclose(fd);
 		return 0;
 	}
 
 	struct ser_handle *sh = ser_open(filename, ser_mode_write);
-	if (!sh)
+	if (!sh) {
+		LOG_MOD_WARN("snapshot/v2", "%s: %s\n", filename, strerror(errno));
 		return -1;
+	}
 
 	ser_write_tag(sh, 0x23, strlen(snapv2_header));
 	ser_write_untagged(sh, snapv2_header, strlen(snapv2_header));
 
 	ser_write_open_string(sh, SNAPSHOT_SER_MACHINE, "machine");
-	part_serialise((struct part *)xroar_machine, sh);
+	part_serialise((struct part *)xroar.machine, sh);
 
-	vdrive_interface_serialise(xroar_vdrive_interface, sh, SNAPSHOT_SER_VDRIVE_INTF);
+	vdrive_interface_serialise(xroar.vdrive_interface, sh, SNAPSHOT_SER_VDRIVE_INTF);
 
 	ser_write_close_tag(sh);
 	ser_close(sh);
@@ -151,8 +161,9 @@ int write_snapshot(const char *filename) {
 
 static int read_v2_snapshot(const char *filename) {
 	struct ser_handle *sh = ser_open(filename, ser_mode_read);
-	if (!sh)
+	if (!sh) {
 		return -1;
+	}
 
 	int tag = ser_read_tag(sh);
 	if (tag != 0x23) {
@@ -160,7 +171,7 @@ static int read_v2_snapshot(const char *filename) {
 		return -1;
 	}
 	char *id = ser_read_string(sh);
-	if (strcmp(id, snapv2_header) != 0) {
+	if (!id || strcmp(id, snapv2_header) != 0) {
 		ser_close(sh);
 		return -1;
 	}
@@ -175,7 +186,7 @@ static int read_v2_snapshot(const char *filename) {
 			// Deserialises new machine.
 			m = (struct machine *)part_deserialise(sh);
 			if (m && !part_is_a((struct part *)m, "machine")) {
-				LOG_WARN("Snapshot v2 read: not a machine\n");
+				LOG_MOD_WARN("snapshot/v2", "read: not a machine\n");
 				part_free((struct part *)m);
 				m = NULL;
 				ser_set_error(sh, ser_error_format);
@@ -191,11 +202,11 @@ static int read_v2_snapshot(const char *filename) {
 				ser_set_error(sh, ser_error_format);
 				break;
 			}
-			vdrive_interface_deserialise(xroar_vdrive_interface, sh);
+			vdrive_interface_deserialise(xroar.vdrive_interface, sh);
 			break;
 
 		default:
-			LOG_WARN("Snapshot v2 read: unknown tag '%d'\n", tag);
+			LOG_MOD_WARN("snapshot/v2", "read: unknown tag '%d'\n", tag);
 			break;
 		}
 		if (ser_error(sh))
@@ -203,7 +214,7 @@ static int read_v2_snapshot(const char *filename) {
 	}
 
 	if (ser_error(sh)) {
-		LOG_WARN("Snapshot v2 read: %s\n", ser_errstr(sh));
+		LOG_MOD_WARN("snapshot/v2", "read: %s at byte %zd\n", ser_errstr(sh), ser_errpos(sh));
 	}
 
 	ser_close(sh);
@@ -218,13 +229,16 @@ static int read_v2_snapshot(const char *filename) {
 	// TODO: verify that any deserialised carts had their configs included
 	// earlier in the snapshot.
 
-	if (xroar_machine) {
-		part_free((struct part *)xroar_machine);
+	if (xroar.auto_kbd) {
+		auto_kbd_free(xroar.auto_kbd);
+		xroar.auto_kbd = NULL;
 	}
-	xroar_machine_config = m->config;
-	xroar_machine = m;
+	if (xroar.machine) {
+		part_free((struct part *)xroar.machine);
+	}
+	xroar.machine_config = m->config;
+	xroar.machine = m;
 	xroar_connect_machine();
-	xroar_connect_cart();
 
 	return 0;
 }
@@ -351,16 +365,20 @@ static int read_v1_snapshot(const char *filename) {
 	// Default to Dragon 64 for old snapshots
 	struct machine_config *mc = machine_config_by_arch(1);  // 1 == old ARCH_DRAGON64
 	xroar_configure_machine(mc);
-	xroar_machine->reset(xroar_machine, RESET_HARD);
+	xroar.machine->reset(xroar.machine, RESET_HARD);
 	// If old snapshot, buffer contains register dump
 	if (buffer[0] != 'X') {
-		old_set_registers(xroar_machine, buffer + 3);
+		old_set_registers(xroar.machine, buffer + 3);
 	}
+
+	// Cache values to set once done
+	int keymap = -1;
+
 	struct cart_config *cart_config = NULL;
 	while ((section = fs_read_uint8(fd)) >= 0) {
 		unsigned size = fs_read_uint16(fd);
 		if (size == 0) size = 0x10000;
-		LOG_DEBUG(3, "Snapshot v1 read: chunk type %d, size %u\n", section, size);
+		LOG_MOD_DEBUG(3, "snapshot/v1", "read: chunk type %d, size %u\n", section, size);
 		switch (section) {
 			case ID_ARCHITECTURE:
 				// Deprecated: Machine architecture
@@ -371,21 +389,20 @@ static int read_v1_snapshot(const char *filename) {
 					free(mc->architecture);
 				mc->architecture = xstrdup(old_arch_mapping[tmp]);
 				xroar_configure_machine(mc);
-				xroar_machine->reset(xroar_machine, RESET_HARD);
+				xroar.machine->reset(xroar.machine, RESET_HARD);
 				size--;
 				break;
 			case ID_KEYBOARD_MAP:
 				// Deprecated: Keyboard map
 				if (size < 1) break;
-				tmp = fs_read_uint8(fd);
-				xroar_set_keyboard_type(1, tmp);
+				keymap = fs_read_uint8(fd);
 				size--;
 				break;
 			case ID_REGISTER_DUMP:
 				// Deprecated
 				if (size < 14) break;
 				size -= fread(buffer, 1, 14, fd);
-				old_set_registers(xroar_machine, buffer);
+				old_set_registers(xroar.machine, buffer);
 				break;
 
 			case ID_MC6809_STATE:
@@ -393,12 +410,12 @@ static int read_v1_snapshot(const char *filename) {
 					// MC6809 state
 					if (size < 20) break;
 					if (mc->cpu != CPU_MC6809) {
-						LOG_WARN("Snapshot v1 read: CPU mismatch - skipping MC6809 chunk\n");
+						LOG_MOD_WARN("snapshot/v1", "read: CPU mismatch - skipping MC6809 chunk\n");
 						break;
 					}
-					struct MC6809 *cpu = (struct MC6809 *)part_component_by_id_is_a(&xroar_machine->part, "CPU", "MC6809");
+					struct MC6809 *cpu = (struct MC6809 *)part_component_by_id_is_a(&xroar.machine->part, "CPU", "MC6809");
 					if (!cpu) {
-						LOG_WARN("Snapshot v1 read: CPU not found - skipping MC6809 chunk\n");
+						LOG_MOD_WARN("snapshot/v1", "read: CPU not found - skipping MC6809 chunk\n");
 						break;
 					}
 					cpu->reg_cc = fs_read_uint8(fd);
@@ -457,12 +474,12 @@ static int read_v1_snapshot(const char *filename) {
 					// HD6309 state
 					if (size < 27) break;
 					if (mc->cpu != CPU_HD6309) {
-						LOG_WARN("Snapshot v1 read: CPU mismatch - skipping HD6309 chunk\n");
+						LOG_MOD_WARN("snapshot/v1", "read: CPU mismatch - skipping HD6309 chunk\n");
 						break;
 					}
-					struct HD6309 *hcpu = (struct HD6309 *)part_component_by_id_is_a(&xroar_machine->part, "CPU", "HD6309");
+					struct HD6309 *hcpu = (struct HD6309 *)part_component_by_id_is_a(&xroar.machine->part, "CPU", "HD6309");
 					if (!hcpu) {
-						LOG_WARN("Snapshot v1 read: CPU not found - skipping HD6309 chunk\n");
+						LOG_MOD_WARN("snapshot/v1", "read: CPU not found - skipping HD6309 chunk\n");
 						break;
 					}
 					struct MC6809 *cpu = &hcpu->mc6809;
@@ -527,13 +544,13 @@ static int read_v1_snapshot(const char *filename) {
 					// v1.8 adds a separate cart chunk
 					switch (tmp) {
 					case 1: // DragonDOS
-						xroar_set_cart(1, "dragondos");
+						cart_config = cart_config_by_name("dragondos");
 						break;
 					case 2: // RSDOS
-						xroar_set_cart(1, "rsdos");
+						cart_config = cart_config_by_name("rsdos");
 						break;
 					case 3: // Delta
-						xroar_set_cart(1, "delta");
+						cart_config = cart_config_by_name("delta");
 						break;
 					default: break;
 					}
@@ -544,16 +561,16 @@ static int read_v1_snapshot(const char *filename) {
 					size--;
 				}
 				xroar_configure_machine(mc);
-				xroar_machine->reset(xroar_machine, RESET_HARD);
+				xroar.machine->reset(xroar.machine, RESET_HARD);
 				break;
 
 			case ID_PIA_REGISTERS:
-				for (int i = 0; i < 2; i++) {
+				for (unsigned i = 0; i < 2; ++i) {
 					char id[5];
-					snprintf(id, sizeof(id), "PIA%d", i);
-					struct MC6821 *pia = (struct MC6821 *)part_component_by_id_is_a(&xroar_machine->part, id, "MC6821");
+					snprintf(id, sizeof(id), "PIA%u", i);
+					struct MC6821 *pia = (struct MC6821 *)part_component_by_id_is_a(&xroar.machine->part, id, "MC6821");
 					if (!pia) {
-						LOG_WARN("Snapshot v1 read: %s not found - skipping PIA\n", id);
+						LOG_MOD_WARN("snapshot/v1", "read: %s not found - skipping PIA\n", id);
 						break;
 					}
 					if (size < 3) break;
@@ -562,29 +579,37 @@ static int read_v1_snapshot(const char *filename) {
 					pia->a.control_register = fs_read_uint8(fd);
 					size -= 3;
 					mc6821_update_a_state(pia);
+					mc6821_write(pia, 1, pia->a.control_register);
 					if (size < 3) break;
 					pia->b.direction_register = fs_read_uint8(fd);
 					pia->b.output_register = fs_read_uint8(fd);
 					pia->b.control_register = fs_read_uint8(fd);
 					size -= 3;
 					mc6821_update_b_state(pia);
+					mc6821_write(pia, 3, pia->b.control_register);
 				}
 				break;
 
 			case ID_RAM_PAGE0:
 				{
-					struct machine_memory *ram0 = xroar_machine->get_component(xroar_machine, "RAM0");
-					assert(ram0 != NULL);
-					ram0->size = (size < ram0->max_size) ? size : ram0->max_size;
-					size -= fread(ram0->data, 1, ram0->size, fd);
+					struct part *p = &xroar.machine->part;
+					struct ram *ram = (struct ram *)part_component_by_id_is_a(p, "RAM", "ram");
+					if (ram && ram->d && ram->d[0]) {
+						size_t psize = (ram->bank_nelems < size) ? ram->bank_nelems : size;
+						uint8_t *ram_dst = ram->d[0];
+						size -= fread(ram_dst, 1, psize, fd);
+					}
 				}
 				break;
 			case ID_RAM_PAGE1:
 				{
-					struct machine_memory *ram1 = xroar_machine->get_component(xroar_machine, "RAM1");
-					assert(ram1 != NULL);
-					ram1->size = (size < ram1->max_size) ? size : ram1->max_size;
-					size -= fread(ram1->data, 1, ram1->size, fd);
+					struct part *p = &xroar.machine->part;
+					struct ram *ram = (struct ram *)part_component_by_id_is_a(p, "RAM", "ram");
+					if (ram && ram->d && ram->d[0]) {
+						size_t psize = (ram->bank_nelems < (size + 0x8000)) ? ram->bank_nelems - 0x8000: size;
+						uint8_t *ram_dst = (uint8_t *)ram->d[0] + 0x8000;
+						size -= fread(ram_dst, 1, psize, fd);
+					}
 				}
 				break;
 			case ID_SAM_REGISTERS:
@@ -593,9 +618,9 @@ static int read_v1_snapshot(const char *filename) {
 				tmp = fs_read_uint16(fd);
 				size -= 2;
 				{
-					struct MC6883 *sam = (struct MC6883 *)part_component_by_id_is_a(&xroar_machine->part, "SAM", "SN74LS783");
+					struct MC6883 *sam = (struct MC6883 *)part_component_by_id_is_a(&xroar.machine->part, "SAM", "SN74LS783");
 					if (!sam) {
-						LOG_WARN("Snapshot v1 read: SAM not found - skipping SAM registers chunk\n");
+						LOG_MOD_WARN("snapshot/v1", "read: SAM not found - skipping SAM registers chunk\n");
 						break;
 					}
 					mc6883_set_register(sam, tmp);
@@ -611,7 +636,7 @@ static int read_v1_snapshot(const char *filename) {
 				size -= 3;
 				if (version_major != SNAPSHOT_VERSION_MAJOR
 				    || version_minor > SNAPSHOT_VERSION_MINOR) {
-					LOG_WARN("Snapshot v1 read: version %d.%d not supported.\n", version_major, version_minor);
+					LOG_MOD_WARN("snapshot/v1", "read: version %d.%d not supported.\n", version_major, version_minor);
 					fclose(fd);
 					return -1;
 				}
@@ -623,13 +648,15 @@ static int read_v1_snapshot(const char *filename) {
 					int drive;
 					size--;
 					drive = fs_read_uint8(fd);
-					vdrive_eject_disk(xroar_vdrive_interface, drive);
+					vdrive_eject_disk(xroar.vdrive_interface, drive);
 					if (size > 0) {
-						char *name = malloc(size);
-						if (name != NULL) {
-							size -= fread(name, 1, size, fd);
-							vdrive_insert_disk(xroar_vdrive_interface, drive, vdisk_load(name));
-						}
+						char *name = xmalloc(size);
+						size -= fread(name, 1, size, fd);
+						// the NUL should be in the snapshot, but
+						// terminate explicitly just in case:
+						name[size-1] = 0;
+						vdrive_insert_disk(xroar.vdrive_interface, drive, vdisk_load(name));
+						free(name);
 					}
 				}
 				break;
@@ -675,21 +702,21 @@ static int read_v1_snapshot(const char *filename) {
 
 			default:
 				// Unknown chunk
-				LOG_WARN("Snapshot v1 read: unknown chunk type %d in snaphot.\n", section);
+				LOG_MOD_WARN("snapshot/v1", "read: unknown chunk type %d in snaphot.\n", section);
 				break;
 		}
 		if (size > 0) {
-			LOG_WARN("Snapshot v1 read: skipping extra bytes in snapshot chunk type %d\n", (int)section);
+			LOG_MOD_WARN("snapshot/v1", "read: skipping extra bytes in snapshot chunk type %d\n", (int)section);
 			for (; size; size--)
 				(void)fs_read_uint8(fd);
 		}
 	}
 	fclose(fd);
 	if (cart_config) {
-		// XXX really we need something to update the UI here, the
-		// embedded cart config may have changed description.  more
-		// importantly, the UI won't know about the id.
-		xroar_set_cart(1, cart_config->name);
+		ui_update_state(-1, ui_tag_cartridge, cart_config->id, NULL);
+	}
+	if (keymap >= 0) {
+		ui_update_state(-1, ui_tag_keymap, keymap, NULL);
 	}
 	return 0;
 }
