@@ -2,7 +2,7 @@
  *
  *  \brief Dragon sound interface.
  *
- *  \copyright Copyright 2003-2021 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -21,6 +21,7 @@
 
 #include "top-config.h"
 
+#include <assert.h>
 #include <math.h>
 #include <stdint.h>
 #include <stdlib.h>
@@ -32,6 +33,7 @@
 
 #include "events.h"
 #include "logging.h"
+#include "messenger.h"
 #include "module.h"
 #include "sound.h"
 #include "tape.h"
@@ -43,12 +45,16 @@ struct sound_interface_private {
 
 	struct sound_interface public;
 
+	// Messenger client id
+	int msgr_client_id;
+
 	// Describes the mix & output buffers:
 	unsigned buffer_nframes;
 	float *mix_buffer;  // mix buffer
 	int output_nchannels;
 	enum sound_fmt output_fmt;
 	void *output_buffer;  // final output may not be floats
+	_Bool output_buffer_is_silent;
 
 	// Current index into the buffer
 	unsigned buffer_frame;
@@ -80,11 +86,12 @@ struct sound_interface_private {
 
 	// Inputs to audio mux.  Suitable for mixing to the output buffer, so
 	// may have been filtered.
-	float *mux_input[4];
+	float *mux_input[5];
+	float *non_muxed_output;
 
 	// The unfiltered "current" values of those inputs.  Suitable for
 	// feeding back to the single bit sound pin as an input.
-	float mux_input_raw[4];
+	float mux_input_raw[5];
 
 	// Scale & offset to apply to the audio mux.  Intended to simulate the
 	// changing characteristics when enabling or disabling single-bit
@@ -105,6 +112,7 @@ enum sound_source {
 	SOURCE_DAC,
 	SOURCE_TAPE,
 	SOURCE_CART,
+	SOURCE_AY,  // Dragon Professional only
 	SOURCE_NONE,
 	SOURCE_SINGLE_BIT,
 	NUM_SOURCES
@@ -131,6 +139,7 @@ static const float source_gain_v[NUM_SOURCES][3] = {
 	{ 4.50/MAX_V, 2.84/MAX_V, 3.40/MAX_V },  // DAC
 	{ 0.50/MAX_V, 0.40/MAX_V, 0.50/MAX_V },  // Tape
 	{ 4.70/MAX_V, 2.84/MAX_V, 3.40/MAX_V },  // Cart
+	{ 4.70/MAX_V, 2.84/MAX_V, 3.40/MAX_V },  // AY
 	{ 0.00/MAX_V, 0.00/MAX_V, 0.00/MAX_V },  // None
 	{ 0.00/MAX_V, 0.00/MAX_V, 0.00/MAX_V }   // Single-bit
 };
@@ -140,22 +149,30 @@ static const float source_offset_v[NUM_SOURCES][3] = {
 	{ 0.20/MAX_V, 0.18/MAX_V, 1.30/MAX_V },  // DAC
 	{ 2.05/MAX_V, 1.60/MAX_V, 2.35/MAX_V },  // Tape
 	{ 0.00/MAX_V, 0.18/MAX_V, 1.30/MAX_V },  // Cart
+	{ 0.00/MAX_V, 0.18/MAX_V, 1.30/MAX_V },  // AY
 	{ 0.00/MAX_V, 0.00/MAX_V, 0.00/MAX_V },  // None
 	{ 0.00/MAX_V, 0.00/MAX_V, 3.90/MAX_V }   // Single-bit
 };
 
+static void sound_ui_set_gain(void *, int tag, void *smsg);
+
 struct sound_interface *sound_interface_new(void *buf, enum sound_fmt fmt, unsigned rate,
 					    unsigned nchannels, unsigned nframes) {
 	struct sound_interface_private *snd = xmalloc(sizeof(*snd));
 	*snd = (struct sound_interface_private){0};
 	struct sound_interface *sndp = &snd->public;
 
-	sound_set_gain(sndp, -3.0);
+	// Register with messenger
+	snd->msgr_client_id = messenger_client_register();
+
+	ui_messenger_preempt_group(snd->msgr_client_id, ui_tag_gain, MESSENGER_NOTIFY_DELEGATE(sound_ui_set_gain, snd));
+
+	snd->gain = 0.7;  // -3dBFS
 
 	_Bool fmt_big_endian = 1;
 
 	if (nchannels < 1 || nchannels > 2) {
-		LOG_WARN("Invalid number of audio channels: disabling sound.");
+		LOG_MOD_WARN("sound", "invalid number of audio channels");
 		free(snd);
 		return NULL;
 	}
@@ -194,7 +211,7 @@ struct sound_interface *sound_interface_new(void *buf, enum sound_fmt fmt, unsig
 		LOG_DEBUG(1, "16-bit signed %s-endian, ", fmt_big_endian ? "big" : "little" );
 		break;
 	case SOUND_FMT_FLOAT:
-		LOG_DEBUG(1, "Floating point, ");
+		LOG_DEBUG(1, "floating point, ");
 		break;
 	case SOUND_FMT_NULL:
 	default:
@@ -226,12 +243,16 @@ struct sound_interface *sound_interface_new(void *buf, enum sound_fmt fmt, unsig
 
 	snd->current.mux_source = 0;
 	snd->next.mux_source = 0;
-	for (unsigned i = 0; i < 4; i++) {
+	for (unsigned i = 0; i < 5; i++) {
 		snd->mux_input[i] = xmalloc((nframes + 1) * sizeof(float));
 		for (unsigned j = 0; j < (nframes + 1); j++) {
 			snd->mux_input[i][j] = 0.0;
 		}
 	}
+	snd->non_muxed_output = xmalloc((nframes + 1) * sizeof(float));
+	for (unsigned j = 0; j < (nframes + 1); j++) {
+		snd->non_muxed_output[j] = 0.0;
+	}
 
 	snd->last_cycle = event_current_tick;
 
@@ -245,11 +266,13 @@ struct sound_interface *sound_interface_new(void *buf, enum sound_fmt fmt, unsig
 
 void sound_interface_free(struct sound_interface *sndp) {
 	struct sound_interface_private *snd = (struct sound_interface_private *)sndp;
+	messenger_client_unregister(snd->msgr_client_id);
 	event_dequeue(&snd->flush_event);
 	if (snd->output_fmt != SOUND_FMT_FLOAT) {
 		free(snd->mix_buffer);
 	}
-	for (unsigned i = 0; i < 4; i++) {
+	free(snd->non_muxed_output);
+	for (unsigned i = 0; i < 5; i++) {
 		free(snd->mux_input[i]);
 	}
 	free(snd);
@@ -302,6 +325,53 @@ static void send_buffer(struct sound_interface_private *snd) {
 	snd->buffer_frame = 0;
 }
 
+// Send "silence" if audio module needs to be kept up to date with queued audio
+
+void sound_send_silence(struct sound_interface *sndp) {
+	struct sound_interface_private *snd = (struct sound_interface_private *)sndp;
+	// Only if the audio module needs it:
+	if (!DELEGATE_DEFINED(snd->public.write_silence))
+		return;
+
+	if (snd->output_buffer && !snd->output_buffer_is_silent) {
+		// Fill the output buffer with copies of its last sample
+		int nsamples = snd->output_nchannels * snd->buffer_nframes;
+		switch (snd->output_fmt) {
+		case SOUND_FMT_U8:
+		case SOUND_FMT_S8: {
+			uint8_t *output = snd->output_buffer;
+			uint8_t v = output[nsamples - 1];
+			for (int i = nsamples - 1; i; i--)
+				*(output++) = v;
+		} break;
+		case SOUND_FMT_S16_HE:
+		case SOUND_FMT_S16_SE: {
+			uint16_t *output = snd->output_buffer;
+			uint16_t v = output[nsamples - 1];
+			for (int i = nsamples - 1; i; i--)
+				*(output++) = v;
+		} break;
+		case SOUND_FMT_FLOAT: {
+			float *output = snd->output_buffer;
+			float v = output[nsamples - 1];
+			for (int i = nsamples; i; i--)
+				*(output++) = v;
+		} break;
+		default:
+			break;
+		}
+		// Flag so we don't have to do this every time
+		snd->output_buffer_is_silent = 1;
+	}
+
+	snd->output_buffer = DELEGATE_CALL(snd->public.write_silence, snd->output_buffer);
+	if (snd->output_fmt == SOUND_FMT_FLOAT) {
+		// No need to convert floats, point mix buffer at output buffer.
+		snd->mix_buffer = snd->output_buffer;
+	}
+	snd->buffer_frame = 0;
+}
+
 // Fill sound buffer to current point in time, sending to audio module when full.
 
 void sound_update(struct sound_interface *sndp) {
@@ -360,6 +430,25 @@ void sound_update(struct sound_interface *sndp) {
 		snd->mux_input_raw[SOURCE_CART] = 0.0;
 	}
 
+	if (DELEGATE_DEFINED(sndp->get_ay_audio)) {
+		if (mux_source == SOURCE_AY && sndp->ratelimit) {
+			snd->mux_input_raw[SOURCE_AY] = DELEGATE_CALL(sndp->get_ay_audio, event_current_tick, nframes, snd->mux_input[SOURCE_AY]);
+		} else {
+			snd->mux_input_raw[SOURCE_AY] = DELEGATE_CALL(sndp->get_ay_audio, event_current_tick, nframes, NULL);
+			if (mux_source == SOURCE_AY) {
+				mux_source = SOURCE_NONE;
+			}
+		}
+	} else {
+		snd->mux_input_raw[SOURCE_AY] = 0.0;
+	}
+
+	float *non_muxed_output = NULL;
+	if (DELEGATE_DEFINED(sndp->get_non_muxed_audio)) {
+		non_muxed_output = snd->non_muxed_output;
+		DELEGATE_CALL(sndp->get_non_muxed_audio, event_current_tick, nframes, non_muxed_output);
+	}
+
 	// Only fill DAC buffer if it's selected
 	if (mux_source == SOURCE_DAC) {
 		for (unsigned i = 0; i < nframes; i++) {
@@ -383,6 +472,9 @@ void sound_update(struct sound_interface *sndp) {
 			float *ptr = (float *)snd->mix_buffer + snd->buffer_frame * snd->output_nchannels;
 			for (int i = 0; i < count; i++) {
 				float mix_sample = (*(mux_output++) * snd->mux_gain) + snd->bus_offset;
+				if (non_muxed_output) {
+					mix_sample += *(non_muxed_output++);
+				}
 				for (int j = 0; j < snd->output_nchannels; j++) {
 					*(ptr++) = (mix_sample + snd->current.external[j]) * snd->gain;
 				}
@@ -392,6 +484,7 @@ void sound_update(struct sound_interface *sndp) {
 		if (snd->buffer_frame >= snd->buffer_nframes) {
 			send_buffer(snd);
 		}
+		snd->output_buffer_is_silent = 0;
 	}
 
 	// Now that audio has been dealt with up to the current point in time,
@@ -442,26 +535,43 @@ void sound_set_ratelimit(struct sound_interface *sndp, _Bool ratelimit) {
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-// Sets the overall gain applied after mixing sources.  Generally a negative
-// number relative to 0dBFS.
-void sound_set_gain(struct sound_interface *sndp, double db) {
-	struct sound_interface_private *snd = (struct sound_interface_private *)sndp;
-	float v = powf(10., db / 20.);
-	snd->gain = v;
-	if (xroar_ui_interface) {
-		DELEGATE_CALL(xroar_ui_interface->update_state, ui_tag_gain, (int)(v * 100.), &db);
-	}
-}
-
-// Old way to set the gain by specifying a percentage of full scale (linear).
-void sound_set_volume(struct sound_interface *sndp, int v) {
-	struct sound_interface_private *snd = (struct sound_interface_private *)sndp;
-	if (v < 0) v = 0;
-	if (v > 200) v = 200;
-	snd->gain = (float)v / 100.;
-	float db = log10f(snd->gain) * 20.;
-	if (xroar_ui_interface) {
-		DELEGATE_CALL(xroar_ui_interface->update_state, ui_tag_gain, v, &db);
+static void sound_ui_set_gain(void *sptr, int tag, void *smsg) {
+	struct sound_interface_private *snd = sptr;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_gain);
+	if (uimsg->data) {
+		float db = *(float *)uimsg->data;
+		float v;
+		int vi;
+		if (db < -49.9) {
+			v = 0.;
+			vi = 0;
+			db = -50.;
+		} else {
+			v = powf(10., db / 20.);
+			vi = (int)(v * 100.);
+		}
+		snd->gain = v;
+		uimsg->value = vi;
+		*(float *)uimsg->data = db;
+	} else {
+		int vi = uimsg->value;
+		float v;
+		static float db;
+		if (vi <= 0) {
+			vi = 0;
+			v = 0.;
+			db = -50.;
+		} else {
+			if (vi > 200) {
+				vi = 200;
+			}
+			v = (float)vi / 100.;
+			db = log10f(v) * 20.;
+		}
+		snd->gain = v;
+		uimsg->value = vi;
+		uimsg->data = &db;
 	}
 }
 
