@@ -2,7 +2,7 @@
  *
  *  \brief CAS format tape images.
  *
- *  \copyright Copyright 2003-2022 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -166,7 +166,6 @@ static void commit_data(struct tape *t);
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 static struct cue_entry *cue_entry_new(int type);
-static void cue_entry_free(struct cue_entry *entry);
 
 static void cue_set_timing(struct tape_cas *cas, int bit0_cw, int bit1_cw);
 static void cue_add_silence(struct tape_cas *cas, int ms);
@@ -211,12 +210,14 @@ static void init_filename_block(uint8_t *block, const char *filename, int type,
 static void cue_read_ascii(struct tape_cas *cas, const char *filename);
 static void cue_read_k7(struct tape_cas *cas);
 
-static struct tape *do_tape_cas_open(struct tape_interface *ti, const char *filename,
+static struct tape *do_tape_cas_open(const char *filename,
 				     const char *mode, _Bool is_ascii, _Bool is_k7) {
 	struct tape *t;
 	struct tape_cas *cas;
 	FILE *fd;
 
+	const char *mod = is_ascii ? "tape/asc" : (is_k7 ? "tape/k7" : "tape/cas");
+
 	_Bool writing = mode[0] == 'w';
 	_Bool new_file = 0;
 	// if writing, open in read/write mode so that existing cue data can be
@@ -224,21 +225,21 @@ static struct tape *do_tape_cas_open(struct tape_interface *ti, const char *file
 	if (writing) {
 		if (!(fd = fopen(filename, "r+b"))) {
 			if (!(fd = fopen(filename, "wb"))) {
-				LOG_WARN("TAPE/CAS: cannot open '%s': %s\n", filename, strerror(errno));
+				LOG_MOD_WARN(mod, "%s: %s\n", filename, strerror(errno));
 				return NULL;
 			}
 			new_file = 1;
 		}
 	} else {
 		if (!(fd = fopen(filename, "rb"))) {
-			LOG_WARN("TAPE/CAS: cannot open '%s': %s\n", filename, strerror(errno));
+			LOG_MOD_WARN(mod, "%s: %s\n", filename, strerror(errno));
 			return NULL;
 		}
 	}
 
 	off_t filesize = fs_file_size(fd);
 	if (filesize == -1) {
-		LOG_WARN("TAPE/CAS: cannot stat '%s': %s\n", filename, strerror(errno));
+		LOG_MOD_WARN(mod, "%s: %s\n", filename, strerror(errno));
 		fclose(fd);
 		return NULL;
 	}
@@ -249,7 +250,7 @@ static struct tape *do_tape_cas_open(struct tape_interface *ti, const char *file
 		is_k7 = 0;
 	}
 
-	t = tape_new(ti);
+	t = tape_new();
 	t->module = &tape_cas_module;
 	cas = xmalloc(sizeof(*cas));
 	*cas = (struct tape_cas){0};
@@ -294,6 +295,7 @@ static struct tape *do_tape_cas_open(struct tape_interface *ti, const char *file
 		// we won't know what to do with this tape, so bail
 		if (!writing && !cas->cue.list) {
 			cas_close(t);
+			LOG_MOD_WARN(mod, "%s: failed to read CUE data\n", filename);
 			return NULL;
 		}
 	}
@@ -419,11 +421,11 @@ static void cue_read_k7(struct tape_cas *cas) {
 	goto last_block;
 
 malformed:
-	LOG_WARN("TAPE/CAS: malformed K7 file\n");
+	LOG_MOD_WARN("tape/k7", "malformed K7 file\n");
 	goto last_block;
 
 short_read:
-	LOG_WARN("TAPE/CAS: short read processing K7 file\n");
+	LOG_MOD_WARN("tape/k7", "short read processing K7 file\n");
 	goto last_block;
 
 last_block:
@@ -456,16 +458,16 @@ static void cue_read_ascii(struct tape_cas *cas, const char *filename) {
 	cue_add_block(cas, 0xff, 0, 0, offset, NULL);
 }
 
-struct tape *tape_cas_open(struct tape_interface *ti, const char *filename, const char *mode) {
-	return do_tape_cas_open(ti, filename, mode, 0, 0);
+struct tape *tape_cas_open(const char *filename, const char *mode) {
+	return do_tape_cas_open(filename, mode, 0, 0);
 }
 
-struct tape *tape_k7_open(struct tape_interface *ti, const char *filename, const char *mode) {
-	return do_tape_cas_open(ti, filename, mode, 0, 1);
+struct tape *tape_k7_open(const char *filename, const char *mode) {
+	return do_tape_cas_open(filename, mode, 0, 1);
 }
 
-struct tape *tape_asc_open(struct tape_interface *ti, const char *filename, const char *mode) {
-	return do_tape_cas_open(ti, filename, mode, 1, 0);
+struct tape *tape_asc_open(const char *filename, const char *mode) {
+	return do_tape_cas_open(filename, mode, 1, 0);
 }
 
 static void cas_close(struct tape *t) {
@@ -706,6 +708,16 @@ static int cas_pulse_in(struct tape *t, int *pulse_width) {
 
 /* Writing */
 
+// CAS writing takes an incoming stream of samples and measures the time
+// between zero crossings as "pulses".  If no average pulse widths have yet
+// been calculated, a buffer of pulses is built up - this should be coming from
+// leader bytes, with alternating '0' and '1' bits.
+//
+// Once the pulse buffer is filled, it is processed to calculate nominal
+// average pulse widths for cycles forming '0' and '1' bits.  The buffer and
+// any subsequently pulses are then converted to bits and written until silence
+// or motor control desyncs the writer, and the process begins again.
+
 static void process_pulse_buffer(struct tape *t);
 
 static void rewind_tape(struct tape *t) {
@@ -713,7 +725,7 @@ static void rewind_tape(struct tape *t) {
 
 	cas->rewind = 0;
 	if (fs_truncate(cas->fd, 0) < 0) {
-		LOG_WARN("TAPE/CAS: rewind failed: %s\n", strerror(errno));
+		LOG_MOD_WARN("tape/cas", "rewind failed: %s\n", strerror(errno));
 		return;
 	}
 	cas->cue.builder->bit_av_pw = 0;
@@ -886,6 +898,8 @@ static void buffer_pulse(struct tape *t, int ticks) {
 	cas->output.pulse_buffer[cas->output.num_buffered_pulses++] = ticks;
 	if (cas->output.num_buffered_pulses < PULSE_BUFFER_SIZE)
 		return;
+	// Pulse buffer full, process it to determine average pulse widths for
+	// '0' and '1' bits, and write the buffered pulses.
 	process_pulse_buffer(t);
 }
 
@@ -900,13 +914,13 @@ static void pulse_out(struct tape *t, int ticks) {
 		return;
 	}
 
-	// haven't determined cycle widths yet
 	if (cas->cue.builder->bit_av_pw == 0) {
+		// Haven't determined pulse widths yet, so buffer this pulse.
 		buffer_pulse(t, ticks);
 		return;
 	}
 
-	// otherwise add the pulse
+	// Otherwise add the pulse.
 	add_pulse(t, ticks);
 }
 
@@ -919,6 +933,8 @@ static int cas_sample_out(struct tape *t, uint8_t sample, int ticks) {
 		cas->output.sense = sense;
 	}
 	if (sense != cas->output.sense) {
+		// Sense changed; this is a zero crossing.  Emit as a "pulse"
+		// with pulse_out().
 		pulse_out(t, cas->output.pulse_ticks);
 		cas->output.pulse_ticks = ticks;
 		cas->output.sense = sense;
@@ -945,7 +961,8 @@ static struct cue_entry *cue_entry_new(int type) {
 	return new;
 }
 
-static void cue_entry_free(struct cue_entry *entry) {
+static void cue_entry_free(void *sptr) {
+	struct cue_entry *entry = sptr;
 	if (entry->section.data)
 		free(entry->section.data);
 	free(entry);
@@ -1129,16 +1146,16 @@ static _Bool read_cue_data(struct tape_cas *cas) {
 			off_t end = fs_read_uint31(fd);
 			if (start < 0 || end < 0)
 				goto read_failed;
-			int size = end - start;
 			elength -= 8;
 			if (end > start) {
+				int size = end - start;
 				cue_add_raw_section(cas, size, start, NULL);
 			}
 		}
 			break;
 
 		default:
-			LOG_WARN("TAPE/CAS/read_cue_data(): unexpected entry type: %d (%d bytes)\n", type, elength);
+			LOG_MOD_WARN("tape/cas", "read_cue_data(): unexpected entry type: %d (%d bytes)\n", type, elength);
 			if (fseeko(fd, elength, SEEK_CUR) < 0) {
 				cue_list_free(cas);
 				return 0;
@@ -1147,10 +1164,11 @@ static _Bool read_cue_data(struct tape_cas *cas) {
 		}
 
 		if (elength != 0) {
-			if (elength > 0)
-				LOG_WARN("TAPE/CAS/read_cue_data(): read underrun: %d bytes\n", elength);
-			else
-				LOG_WARN("TAPE/CAS/read_cue_data(): read overrun: %d bytes\n", -elength);
+			if (elength > 0) {
+				LOG_MOD_WARN("tape/cas", "read_cue_data(): read underrun: %d bytes\n", elength);
+			} else {
+				LOG_MOD_WARN("tape/cas", "read_cue_data(): read overrun: %d bytes\n", -elength);
+			}
 			cue_list_free(cas);
 			return 0;
 		}
@@ -1159,15 +1177,15 @@ static _Bool read_cue_data(struct tape_cas *cas) {
 
 read_failed:
 	if (feof(fd)) {
-		LOG_WARN("TAPE/CAS/read_cue_data(): EOF reading CUE data\n");
+		LOG_MOD_WARN("tape/cas", "read_cue_data(): EOF reading CUE data\n");
 	} else if (ferror(fd)) {
-		LOG_WARN("TAPE/CAS/read_cue_data(): error reading CUE data\n");
+		LOG_MOD_WARN("tape/cas", "read_cue_data(): error reading CUE data\n");
 	}
 	cue_list_free(cas);
 	return 0;
 
 seek_failed:
-	LOG_WARN("TAPE/CAS/read_cue_data(): seek failed: %s\n", strerror(errno));
+	LOG_MOD_WARN("tape/cas", "read_cue_data(): seek failed: %s\n", strerror(errno));
 	cue_list_free(cas);
 	return 0;
 }
@@ -1250,7 +1268,7 @@ static void write_cue_data(struct tape_cas *cas) {
 			break;
 
 		default:
-			LOG_ERROR("TAPE/CAS/write_cue_data(): unexpected entry type: %d\n", entry->type);
+			LOG_MOD_ERROR("tape/cas", "write_cue_data(): unexpected entry type: %d\n", entry->type);
 			break;
 		}
 	}
