@@ -2,7 +2,7 @@
  *
  *  \brief Virtual floppy drives.
  *
- *  \copyright Copyright 2003-2022 Ciaran Anscomb
+ *  \copyright Copyright 2003-2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -28,7 +28,9 @@
 
 #include "events.h"
 #include "logging.h"
+#include "messenger.h"
 #include "serialise.h"
+#include "ui.h"
 #include "vdisk.h"
 #include "vdrive.h"
 #include "xroar.h"
@@ -46,6 +48,9 @@ struct drive_data {
 struct vdrive_interface_private {
 	struct vdrive_interface public;
 
+	// Messenger client id
+	int msgr_client_id;
+
 	_Bool ready_state;
 	_Bool tr00_state;
 	_Bool index_state;
@@ -71,24 +76,24 @@ struct vdrive_interface_private {
 #define VDRIVE_SER_DRIVE (5)
 
 static const struct ser_struct ser_struct_vdrive[] = {
-	SER_ID_STRUCT_ELEM(1, ser_type_bool,      struct vdrive_interface_private, ready_state),
-	SER_ID_STRUCT_ELEM(2, ser_type_bool,      struct vdrive_interface_private, tr00_state),
-	SER_ID_STRUCT_ELEM(3, ser_type_bool,      struct vdrive_interface_private, index_state),
-	SER_ID_STRUCT_ELEM(4, ser_type_bool,      struct vdrive_interface_private, write_protect_state),
+	SER_ID_STRUCT_ELEM(1, struct vdrive_interface_private, ready_state),
+	SER_ID_STRUCT_ELEM(2, struct vdrive_interface_private, tr00_state),
+	SER_ID_STRUCT_ELEM(3, struct vdrive_interface_private, index_state),
+	SER_ID_STRUCT_ELEM(4, struct vdrive_interface_private, write_protect_state),
 
 	SER_ID_STRUCT_UNHANDLED(VDRIVE_SER_DRIVE),
-	SER_ID_STRUCT_ELEM(5, ser_type_unhandled, struct vdrive_interface_private, drives),
-	SER_ID_STRUCT_ELEM(6, ser_type_int,       struct vdrive_interface_private, cur_direction),
-	SER_ID_STRUCT_ELEM(7, ser_type_unsigned,  struct vdrive_interface_private, cur_drive_number),
-	SER_ID_STRUCT_ELEM(8, ser_type_unsigned,  struct vdrive_interface_private, cur_head),
-	SER_ID_STRUCT_ELEM(9, ser_type_unsigned,  struct vdrive_interface_private, cur_density),
-	SER_ID_STRUCT_ELEM(10, ser_type_unsigned, struct vdrive_interface_private, head_incr),
-	SER_ID_STRUCT_ELEM(11, ser_type_unsigned, struct vdrive_interface_private, head_pos),
+	SER_ID_STRUCT_TYPE(5, ser_type_unhandled, struct vdrive_interface_private, drives),
+	SER_ID_STRUCT_ELEM(6, struct vdrive_interface_private, cur_direction),
+	SER_ID_STRUCT_ELEM(7, struct vdrive_interface_private, cur_drive_number),
+	SER_ID_STRUCT_ELEM(8, struct vdrive_interface_private, cur_head),
+	SER_ID_STRUCT_ELEM(9, struct vdrive_interface_private, cur_density),
+	SER_ID_STRUCT_ELEM(10, struct vdrive_interface_private, head_incr),
+	SER_ID_STRUCT_ELEM(11, struct vdrive_interface_private, head_pos),
 
-	SER_ID_STRUCT_ELEM(12, ser_type_tick,     struct vdrive_interface_private, last_update_cycle),
-	SER_ID_STRUCT_ELEM(13, ser_type_tick,     struct vdrive_interface_private, track_start_cycle),
-	SER_ID_STRUCT_ELEM(14, ser_type_event,    struct vdrive_interface_private, index_pulse_event),
-	SER_ID_STRUCT_ELEM(15, ser_type_event,    struct vdrive_interface_private, reset_index_pulse_event),
+	SER_ID_STRUCT_TYPE(12, ser_type_tick,     struct vdrive_interface_private, last_update_cycle),
+	SER_ID_STRUCT_TYPE(13, ser_type_tick,     struct vdrive_interface_private, track_start_cycle),
+	SER_ID_STRUCT_TYPE(14, ser_type_event,    struct vdrive_interface_private, index_pulse_event),
+	SER_ID_STRUCT_TYPE(15, ser_type_event,    struct vdrive_interface_private, reset_index_pulse_event),
 
 };
 
@@ -105,6 +110,11 @@ static const struct ser_struct_data vdrive_ser_struct_data = {
 #define VDRIVE_SER_DRIVE_CYL (1)
 #define VDRIVE_SER_DRIVE_FILENAME (2)
 
+// UI interaction
+
+void vdrive_ui_set_write_enable(void *, int tag, void *smsg);
+void vdrive_ui_set_write_back(void *, int tag, void *smsg);
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 /* Public methods */
@@ -149,6 +159,12 @@ struct vdrive_interface *vdrive_interface_new(void) {
 	*vip = (struct vdrive_interface_private){0};
 	struct vdrive_interface *vi = &vip->public;
 
+	// Register with messenger
+	vip->msgr_client_id = messenger_client_register();
+
+	ui_messenger_preempt_group(vip->msgr_client_id, ui_tag_disk_write_enable, MESSENGER_NOTIFY_DELEGATE(vdrive_ui_set_write_enable, vip));
+	ui_messenger_preempt_group(vip->msgr_client_id, ui_tag_disk_write_back, MESSENGER_NOTIFY_DELEGATE(vdrive_ui_set_write_back, vip));
+
 	vip->tr00_state = 1;
 	vip->current_drive = &vip->drives[0];
 	vip->cur_direction = 1;
@@ -183,6 +199,7 @@ void vdrive_interface_free(struct vdrive_interface *vi) {
 	if (!vi)
 		return;
 	struct vdrive_interface_private *vip = (struct vdrive_interface_private *)vi;
+	messenger_client_unregister(vip->msgr_client_id);
 	event_dequeue(&vip->index_pulse_event);
 	event_dequeue(&vip->reset_index_pulse_event);
 	for (unsigned i = 0; i < MAX_DRIVES; i++) {
@@ -235,7 +252,7 @@ static _Bool vdrive_read_elem(void *sptr, struct ser_handle *sh, int tag) {
 }
 
 static _Bool vdrive_write_elem(void *sptr, struct ser_handle *sh, int tag) {
-	struct vdrive_interface_private *vip = sptr;
+	const struct vdrive_interface_private *vip = sptr;
 	switch (tag) {
 	case VDRIVE_SER_DRIVE:
 		for (unsigned i = 0; i < MAX_DRIVES; i++) {
@@ -267,6 +284,13 @@ void vdrive_interface_deserialise(struct vdrive_interface *vi, struct ser_handle
 
 	ser_read_struct_data(sh, &vdrive_ser_struct_data, vip);
 
+	for (unsigned i = 0; i < MAX_DRIVES; ++i) {
+		struct vdisk *disk = vip->drives[i].disk;
+		ui_update_state(-1, ui_tag_disk_data, i, disk);
+		ui_update_state(-1, ui_tag_disk_write_enable, disk ? !disk->write_protect : 0, (void *)(intptr_t)i);
+		ui_update_state(-1, ui_tag_disk_write_back, disk ? disk->write_back : 0, (void *)(intptr_t)i);
+	}
+
 	vip->current_drive = &vip->drives[vip->cur_drive_number];
 	if (vip->current_drive->disk) {
 		vip->idamptr = vdisk_extend_disk(vip->current_drive->disk, vip->current_drive->current_cyl, vip->cur_head);
@@ -341,6 +365,47 @@ void vdrive_flush(struct vdrive_interface *vi) {
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
+// UI interaction
+
+void vdrive_ui_set_write_enable(void *sptr, int tag, void *smsg) {
+	struct vdrive_interface_private *vip = sptr;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_disk_write_enable);
+
+	int drive = (intptr_t)uimsg->data;
+	if (drive < 0 || drive >= MAX_DRIVES) {
+		return;
+	}
+	struct vdisk *vd = vip->drives[drive].disk;
+	if (!vd) {
+		uimsg->value = 0;
+		return;
+	}
+	// Note supplied current and return values are inverted (protect = !enable)
+	vd->write_protect = !ui_msg_adjust_value_range(uimsg, !vd->write_protect, 0, 0, 1,
+						       UI_ADJUST_FLAG_CYCLE);
+}
+
+void vdrive_ui_set_write_back(void *sptr, int tag, void *smsg) {
+	struct vdrive_interface_private *vip = sptr;
+	struct ui_state_message *uimsg = smsg;
+	assert(tag == ui_tag_disk_write_back);
+
+	int drive = (intptr_t)uimsg->data;
+	if (drive < 0 || drive >= MAX_DRIVES) {
+		return;
+	}
+	struct vdisk *vd = vip->drives[drive].disk;
+	if (!vd) {
+		uimsg->value = 0;
+		return;
+	}
+	vd->write_back = ui_msg_adjust_value_range(uimsg, vd->write_back, 0, 0, 1,
+						   UI_ADJUST_FLAG_CYCLE);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
 /* Signals to all drives */
 
 static void vdrive_set_dirc(void *sptr, _Bool dirc) {
@@ -391,6 +456,7 @@ void vdrive_write(void *sptr, uint8_t data) {
 	if (!vip->track_base) {
 		vip->idamptr = vdisk_extend_disk(vip->current_drive->disk, vip->current_drive->current_cyl, vip->cur_head);
 		vip->track_base = (uint8_t *)vip->idamptr;
+		ui_update_state(-1, ui_tag_disk_data, vip->cur_drive_number, vip->current_drive->disk);
 	}
 	for (unsigned i = vip->head_incr; i; i--) {
 		if (vip->track_base && vip->head_pos < vip->current_drive->disk->track_length) {
@@ -438,6 +504,7 @@ void vdrive_write_idam(void *sptr) {
 	if (!vip->track_base) {
 		vip->idamptr = vdisk_extend_disk(vip->current_drive->disk, vip->current_drive->current_cyl, vip->cur_head);
 		vip->track_base = (uint8_t *)vip->idamptr;
+		ui_update_state(-1, ui_tag_disk_data, vip->cur_drive_number, vip->current_drive->disk);
 	}
 	if (vip->track_base && (vip->head_pos+vip->head_incr) < vip->current_drive->disk->track_length) {
 		// Write 0xfe
@@ -464,11 +531,11 @@ void vdrive_write_idam(void *sptr) {
 }
 
 unsigned vdrive_time_to_next_byte(void *sptr) {
-	struct vdrive_interface_private *vip = sptr;
+	const struct vdrive_interface_private *vip = sptr;
 	event_ticks next_cycle = vip->track_start_cycle + (vip->head_pos - 128) * BYTE_TIME;
 	int to_time = event_tick_delta(next_cycle, event_current_tick);
 	if (to_time < 0) {
-		LOG_DEBUG(3, "Negative time to next byte!\n");
+		LOG_MOD_DEBUG(3, "vdrive", "negative time to next byte!\n");
 		return 1;
 	}
 	return (unsigned)to_time;
@@ -500,7 +567,7 @@ unsigned vdrive_time_to_next_idam(void *sptr) {
 	next_cycle = vip->track_start_cycle + (next_head_pos - 128) * BYTE_TIME;
 	int to_time = event_tick_delta(next_cycle, event_current_tick);
 	if (to_time < 0) {
-		LOG_DEBUG(3, "Negative time to next IDAM!\n");
+		LOG_MOD_DEBUG(3, "vdrive", "negative time to next IDAM!\n");
 		return 1;
 	}
 	return to_time;
@@ -578,10 +645,14 @@ static void set_write_protect_state(struct vdrive_interface_private *vip, _Bool
 }
 
 static void update_signals(struct vdrive_interface_private *vip) {
-	struct vdrive_interface *vi = &vip->public;
 	set_ready_state(vip, vip->current_drive->disk != NULL);
 	set_tr00_state(vip, vip->current_drive->current_cyl == 0);
-	DELEGATE_SAFE_CALL(vi->update_drive_cyl_head, vip->cur_drive_number, vip->current_drive->current_cyl, vip->cur_head);
+	struct vdrive_info vdrive_info = {
+		.drive = vip->cur_drive_number,
+		.cylinder = vip->current_drive->current_cyl,
+		.head = vip->cur_head,
+	};
+	ui_update_state(vip->msgr_client_id, ui_tag_disk_drive_info, 0, &vdrive_info);
 	if (!vip->ready_state) {
 		set_write_protect_state(vip, 0);
 		vip->track_base = NULL;
