@@ -2,7 +2,7 @@
  *
  *  \brief Windows tape control window.
  *
- *  \copyright Copyright 2023 Ciaran Anscomb
+ *  \copyright Copyright 2024 Ciaran Anscomb
  *
  *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
  *
@@ -19,14 +19,9 @@
 #ifndef XROAR_WINDOWS_TAPECONTROL_H_
 #define XROAR_WINDOWS_TAPECONTROL_H_
 
-struct ui_sdl2_interface;
+struct ui_windows32_interface;
+struct uiw32_dialog;
 
-void windows32_tc_create_window(struct ui_sdl2_interface *uisdl2);
-void windows32_tc_show_window(struct ui_sdl2_interface *uisdl2);
-
-void windows32_tc_update_tape_state(struct ui_sdl2_interface *uisdl2, int flags);
-void windows32_tc_update_input_filename(struct ui_sdl2_interface *uisdl2, const char *filename);
-void windows32_tc_update_output_filename(struct ui_sdl2_interface *uisdl2, const char *filename);
-void windows32_tc_update_tape_playing(struct ui_sdl2_interface *uisdl2, int playing);
+struct uiw32_dialog *uiw32_tc_dialog_new(struct ui_windows32_interface *);
 
 #endif
